package applet;

import java.awt.BorderLayout;
import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;

import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.border.TitledBorder;
import javax.swing.text.BadLocationException;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;

@SuppressWarnings("serial")
public class ConsoleView extends TuneTab {

	public class ConsoleOutput extends ByteArrayOutputStream {

		private final boolean fIsStdOut;

		public ConsoleOutput(boolean isStdOut) {
			fIsStdOut = isStdOut;
		}

		@Override
		public synchronized void write(byte[] b, int off, int len) {
			super.write(b, off, len);
			updateConsole(fIsStdOut, new String(b, off, len));
		}

		@Override
		public synchronized void write(int b) {
			super.write(b);
			updateConsole(fIsStdOut, "" + (char) b);
		}

	}

	private final ConsoleOutput fOut;
	private final ConsoleOutput fErr;
	private final JTextPane fConsoleArea;
	private final StyledDocument fDocument;

	private static final String STYLE_STDOUT = "STYLE_STDOUT";
	private static final String STYLE_STDERR = "STYLE_STDERR";
	private final JScrollPane fScrollPane;

	public ConsoleView() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		JPanel consoleGroup = new JPanel();
		consoleGroup.setBorder(new TitledBorder("Console"));
		consoleGroup.setLayout(new BorderLayout());
		add(consoleGroup);

		fConsoleArea = new JTextPane();
		fConsoleArea.setEditable(false);
		fDocument = fConsoleArea.getStyledDocument();
		fScrollPane = new JScrollPane(fConsoleArea);
		consoleGroup.add(fScrollPane, BorderLayout.CENTER);

		// BEGIN set style
		Style def = StyleContext.getDefaultStyleContext().getStyle(
				StyleContext.DEFAULT_STYLE);
		Style regular = fDocument.addStyle("regular", def);
		StyleConstants.setFontFamily(regular, "Courier");
		StyleConstants.setFontSize(regular, 12);

		Style s = fDocument.addStyle(STYLE_STDOUT, regular);
		StyleConstants.setForeground(s, Color.blue);

		s = fDocument.addStyle(STYLE_STDERR, regular);
		StyleConstants.setBold(s, true);
		StyleConstants.setForeground(s, Color.red);
		// END set style

		fOut = new ConsoleOutput(true);
		System.setOut(new PrintStream(fOut));
		fErr = new ConsoleOutput(false);
		System.setErr(new PrintStream(fErr));
	}

	public synchronized void updateConsole(boolean isStdOut, String str) {
		try {
			fDocument.insertString(fDocument.getLength(), str, fDocument
					.getStyle(isStdOut ? STYLE_STDOUT : STYLE_STDERR));
			fConsoleArea.setCaretPosition(fDocument.getLength());
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
	}

	@Override
	public void setTune(Player m_engine, SidTune m_tune) {
		// TODO Auto-generated method stub
	}

	public void propertyChange(PropertyChangeEvent evt) {
		// TODO Auto-generated method stub
	}

}
