package applet;

import static sidplay.ConsolePlayer.playerExit;
import static sidplay.ConsolePlayer.playerFast;
import static sidplay.ConsolePlayer.playerRestart;

import java.applet.Applet;
import java.applet.AudioClip;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.FilteredImageSource;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.CharArrayWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.SocketAddress;
import java.net.URL;
import java.util.ArrayList;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.GrayFilter;
import javax.swing.ImageIcon;
import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JProgressBar;
import javax.swing.JTabbedPane;
import javax.swing.UIManager;

import libsidplay.common.SIDEmu;
import libsidplay.components.c1530.Datasette;
import libsidplay.components.c1541.C1541;
import libsidplay.sidtune.SidTune;
import libsidutils.PRG2TAP;
import libsidutils.STIL;
import libsidutils.STIL.STILEntry;
import resid_builder.FilterConfig;
import resid_builder.ReSID;
import resid_builder.resid.SID;
import resid_builder.resid.ISIDDefs.ChipModel;
import resid_builder.resid.ISIDDefs.SamplingMethod;
import sidplay.ConsolePlayer;
import sidplay.IniConfig;
import sidplay.ConsolePlayer.OUTPUTS;
import sidplay.ConsolePlayer.SIDEMUS;
import sidplay.IniConfig.favorites_section;
import sidplay.IniConfig.sidplay2_section;
import sidplay.audio.CompareDriver;
import applet.disassembler.MemoryView;
import applet.events.IMuteVoice;
import applet.events.ITuneStateChanged;
import applet.events.favorites.IAddFavoritesTab;
import applet.events.favorites.IChangeFavoritesTab;
import applet.events.favorites.IFavoriteTabNames;
import applet.events.favorites.IGetFavorites;
import applet.events.favorites.IRemoveFavoritesTab;
import applet.favorites.FavoritesView;
import applet.favorites.IFavorites;
import applet.fileviewer.CGSCView;
import applet.fileviewer.HVSCView;
import applet.general.GeneralCanvas;
import applet.graphical.GraphicalView;
import applet.hardsid.HardSIDView;
import applet.joystick.JoystickTab;
import applet.settings.SettingsPanel;
import applet.siddump.SIDDumpView;
import applet.sidreg.SIDRegView;
import applet.stil.STILView;
import applet.video.D64FileFilter;
import applet.video.TapeFileFilter;
import applet.video.VideoView;

/**
 * @author Ken Hndel
 * @author Joakim Eriksson
 * 
 *         SID Player GUI
 */
public class JSIDPlay2 extends JApplet implements ActionListener, PropertyChangeListener {

	/** Build date, set by ant */
	public static final String DATE = "24.09.2010";

	/** Application icon */
	private static final ImageIcon WINDOW_ICON = new ImageIcon(JSIDPlay2.class.getResource("icons/8580_3390.jpg"));
	
	/** Main window title */
	public static final String FRAME_TITLE = "Java SIDPlay2 - Music Player & C64 SID Chip Emulator";

	/** Icon for indicating the playback source */
	private static final ImageIcon PLAY_ICON = new ImageIcon(JSIDPlay2.class.getResource("icons/fav_play.png"));

	public static final String PROP_TIME = "time";

	public static final String PROP_HVSC = "hvsc";

	public static final String PROP_CGSC = "cgsc";

	public static final String PROP_FILTER = "filter";

	public static final String PROP_CHIP_MODEL = "chip_model";

	public static final String PROP_PLAY_ORIGINAL = "play_original";

	//
	// command Keys
	//

	private static final int LOAD_KEY = KeyEvent.VK_L;
	private static final int DISK_KEY = KeyEvent.VK_D;
	private static final int TAPE_KEY = KeyEvent.VK_T;
	private static final int CART_KEY = KeyEvent.VK_C;
	private static final int PAUSE_KEY = KeyEvent.VK_P;
	private static final int PREVIOUS_KEY = KeyEvent.VK_E;
	private static final int NEXT_KEY = KeyEvent.VK_N;
	private static final int RESET_KEY = KeyEvent.VK_R;
	private static final int NORMAL_KEY = KeyEvent.VK_O;
	private static final int FASTER_KEY = KeyEvent.VK_F;
	private static final int ADD_FAV_KEY = KeyEvent.VK_A;
	private static final int FREEZE_KEY = KeyEvent.VK_Z;

	private static final ImageIcon stilIcon = new ImageIcon(JSIDPlay2.class.getResource("icons/stil.png"));

	private static final ImageIcon stilNoIcon = new ImageIcon(JSIDPlay2.class.getResource("icons/stil_no.png"));

	/**
	 * If java's automatically generated disabled icon is too dark,
	 * use this method to make a lighter disabled icon.
	 * 
	 * @param jbutton the button to modify
	 */
	public static void makeDisabledIcon(JButton jbutton) {
		ImageIcon ii = (ImageIcon) jbutton.getIcon();
		FilteredImageSource fis = new FilteredImageSource(ii.getImage().getSource(), new GrayFilter(false, -192));
		jbutton.setDisabledIcon(new ImageIcon(jbutton.createImage(fis)));
	}

	/**
	 * If java's automatically generated disabled icon is too dark,
	 * use this method to make a lighter disabled icon.
	 * 
	 * @param jlabel the label to modify
	 */
	public static void makeDisabledIcon(JLabel jlabel) {
		ImageIcon ii = (ImageIcon) jlabel.getIcon();
		FilteredImageSource fis = new FilteredImageSource(ii.getImage().getSource(), new GrayFilter(false, -192));
		jlabel.setDisabledIcon(new ImageIcon(jlabel.createImage(fis)));
	}

	/**
	 * Console player
	 */
	public final ConsolePlayer cp;

	/**
	 * Console player thread
	 */
	private Thread fPlayerThread;

	// redirect out/err streams first!
	private final ConsoleView consoleView = new ConsoleView();

	private final STIL stil = new STIL();

	// GUI
	private JButton fLoadButton;

	private JButton fPreviousButton;

	private JButton fNextButton;

	private JButton fPauseButton;

	private String fSongLength;

	protected ArrayList<TuneTab> tabs = new ArrayList<TuneTab>();

	private JTabbedPane tabbedPane;

	public PropertyChangeSupport support = new PropertyChangeSupport(this);

	private static AudioClip trackSound;
	private static AudioClip motorSound;

	/**
	 * Applet constructor
	 */
	public JSIDPlay2() {
		this(new String[0]);
	}

	/**
	 * Application constructor
	 */
	public JSIDPlay2(final String[] args) {
		support.addPropertyChangeListener(this);
		
		cp = new ConsolePlayer();
		if (args.length != 0) {
			cp.args(args);
		}

		// Load C1541 drive sound
		final URL trackUrl = getClass().getResource(
		"/applet/drive/sounds/track.wav");
		trackSound = Applet.newAudioClip(trackUrl);
		final URL motorUrl = getClass().getResource(
		"/applet/drive/sounds/motor.wav");
		motorSound = Applet.newAudioClip(motorUrl);
	}

	//
	// Threads
	//

	private final Runnable playerRunnable = new Runnable() {
		public void run() {
			while (true) {
				if (!cp.open()) {
					return;
				}

				setUI();

				int oldSeconds = -1;
				while (true) {
					if (cp.state() == ConsolePlayer.playerPaused) {
						try {
							Thread.sleep(250);
						}
						catch (InterruptedException e) {
							break;
						}
					}
					
					if (!cp.play()) {
						break;
					}

					final int seconds = cp.m_engine.time();
					support.firePropertyChange(PROP_TIME, oldSeconds, seconds);
					oldSeconds = seconds;
				}

				// save settings and filter definitions,
				// only if dirty
				cp.m_iniCfg.write();

				if ((cp.state() & ~playerFast) == playerRestart) {
					continue;
				}

				break;
			}
			
			support.firePropertyChange(ITuneStateChanged.class.getName(), null, new ITuneStateChanged() {
				public String getTune() {
					return getPlayingFile();
				}

				public boolean naturalFinished() {
					return cp.state() == playerExit;
				}

			});
		}
	};

	private JLabel fStatusLine;

	private JPanel fButtonsPanel;

	protected int mouseX;

	protected int mouseY;

	protected JPopupMenu fContextPopup;

	private JButton fAddFavButton;

	/**
	 * Is Floppy Disk Drive sound enabled?
	 */
	private boolean fDriveSound;
	
	/**
	 * Last state of the Floppy motor.
	 */
	private boolean oldMotor;
	
	/**
	 * Last state of the Floppy track.
	 */
	private int oldTrack;

	private VideoView fVideoView;

	private JProgressBar fProgressBar;

	//
	// Applet methods
	//
	@Override
	public void init() {
		createGUI();
	}

	@Override
	public void start() {
		setChipModel(cp.m_iniCfg.emulation().getDefaultSidModel());
		
		final Thread gUIThread = new Thread(new Runnable() {
			public void run() {
				while (true) {
					try {
						Thread.sleep(50);
					} catch (final InterruptedException e) {
						break;
					}
					for (final TuneTab tab : tabs) {
						if (tab.isVisible()) {
							tab.updateTab();
						}
					}
					setStatusLine();
				}
			}
		});
		gUIThread.setPriority(Thread.MIN_PRIORITY);
		gUIThread.start();

		startC64();
	}

	@Override
	public void stop() {
	}
	
	public void stopC64() {
		try {
			while (fPlayerThread.isAlive()) {
				cp.quit();
				fPlayerThread.join(100);
			}
		} catch (final InterruptedException e) {
		}
		
		cp.close();
	}

	//
	// Getter and Setter
	//

	public void setChipModel(final ChipModel model) {
		SID sid = getSID(0);
		if (sid != null) {
			sid.setChipModel(model);
		}
		support.firePropertyChange(PROP_CHIP_MODEL, model == ChipModel.MOS6581, model == ChipModel.MOS8580);
	}

	public void setDigiBoost(final boolean selected) {
		// set settings
		cp.m_iniCfg.emulation().setDigiBoosted8580(selected);

		final int input = selected ? -0x10000 : 0;
		SID sid = getSID(0);
		if (sid != null) {
			sid.input(input);
		}
		sid = getSID(1);
		if (sid != null) {
			sid.input(input);
		}
	}

	public boolean getDigiBoost8580() {
		return cp.m_iniCfg.emulation().isDigiBoosted8580();
	}

	public ChipModel getLockModel() {
		return cp.m_iniCfg.emulation().getUserSidModel();
	}

	public void setLockModel(final ChipModel userModel) {
		cp.m_iniCfg.emulation().setUserSidModel(userModel);
		cp.updateSidEmulation();
	}

	public void setOutputDevice(final OUTPUTS device, final SIDEMUS emu) {
		cp.m_driver.output = device;
		cp.m_driver.sid = emu;

		for (int i = 0; i < tabbedPane.getComponentCount(); i++) {
			if (tabbedPane.getComponent(i) instanceof SettingsPanel
					|| tabbedPane.getComponent(i) instanceof GraphicalView) {
				if (emu == SIDEMUS.EMU_HARDSID) {
					tabbedPane.setEnabledAt(i, false);
				} else {
					tabbedPane.setEnabledAt(i, true);
				}
			}
		}
	}

	public OUTPUTS getOutputDevice() {
		return cp.m_driver.output;
	}

	public SIDEMUS getOutputDeviceEmu() {
		return cp.m_driver.sid;
	}

	public void setOutputDir(final String dir) {
		cp.m_outfile = dir;
	}

	public String getOutputDir() {
		return cp.m_outfile;
	}

	public void setPlayOriginal(final boolean playOriginal) {
		cp.m_iniCfg.audio().setPlayOriginal(playOriginal);
		if (cp.m_driver.device instanceof CompareDriver) {
			((CompareDriver) cp.m_driver.device).setPlayOriginal(playOriginal);
		}
		support.firePropertyChange(PROP_PLAY_ORIGINAL, !playOriginal, playOriginal);
	}

	public void setHardSIDDevices(final int hardsid6581, final int hardsid8580) {
		cp.m_iniCfg.emulation().setHardsid6581(hardsid6581);
		cp.m_iniCfg.emulation().setHardsid8580(hardsid8580);
		// A restart is necessary to close/re-open the HardSID card
		cp.restart();
	}

	public void setEnableSLDb(final boolean enableSLDb) {
		cp.m_iniCfg.sidplay2().setEnableDatabase(enableSLDb);
		cp.setSLDb(enableSLDb);
	}

	public boolean getEnableSLDb() {
		return cp.m_iniCfg.sidplay2().isEnableDatabase();
	}

	public void setMute(final int sidNum, final int voiceNo, final boolean mute) {
		cp.m_engine.mute(sidNum, voiceNo, mute);
		support.firePropertyChange(IMuteVoice.class.getName(), null, new IMuteVoice() {

			public int getSidNum() {
				return sidNum;
			}

			public int getVoice() {
				return voiceNo;
			}

			public boolean isMute() {
				return mute;
			}

		});
	}

	public void setFilter(final String filterName) {
		if (filterName == null) {
			return;
		}

		boolean enable;
		if ("".equals(filterName)) {
			enable = false;
		} else {
			enable = true;
			saveCurrentFilter(filterName);
		}
		cp.m_iniCfg.emulation().setFilter(enable);

		final FilterConfig f6581 = cp.m_iniCfg.getFilter(cp.m_iniCfg.emulation().getFilter6581());
		final FilterConfig f8580 = cp.m_iniCfg.getFilter(cp.m_iniCfg.emulation().getFilter8580());

		for (int i = 0; i < 2; i ++) {
			final SIDEmu resid = getSIDEmu(i);
			if (resid != null) {
				resid.filter(enable);
				resid.filter(f6581, f8580);
			}
		}

		support.firePropertyChange(PROP_FILTER, null, filterName);
	}

	private void saveCurrentFilter(final String filterName) {
		final SID sid = getSID(0);
		if (sid != null) {
			final ChipModel model = sid.getChipModel();
			if (model == ChipModel.MOS6581) {
				cp.m_iniCfg.emulation().setFilter6581(filterName);
			} else {
				cp.m_iniCfg.emulation().setFilter8580(filterName);
			}
		}
	}

	public void setDistortion(final String filterName, final float attenuation, final float nonlinearity, final float resonanceFactor) {
		if (filterName != null) {
			FilterConfig filter = cp.m_iniCfg.getFilter(filterName);
			if (filter != null) {
				// save settings
				filter.setAttenuation(attenuation);
				filter.setNonlinearity(nonlinearity);
				filter.setResonanceFactor(resonanceFactor);
			}
		}

		for (int i = 0; i < 2; i ++) {
			final SID sid = getSID(i);
			if (sid != null) {
				sid.getFilter6581().setDistortionProperties(attenuation, nonlinearity, resonanceFactor);
				sid.getFilter8580().setDistortionProperties(attenuation, nonlinearity, resonanceFactor);
			}
		}
	}

	public float[] getDistortion() {
		final SID sid = getSID(0);
		if (sid == null) {
			return null;
		}
		return sid.getFilter().getDistortionProperties();
	}

	public void setCGSC(final String absolutePath) {
		// save settings
		final String oldCgsc = cp.m_iniCfg.sidplay2().getCgsc();
		cp.m_iniCfg.sidplay2().setCgsc(absolutePath);

		if (!absolutePath.equals(oldCgsc)) {
			support.firePropertyChange(PROP_CGSC, oldCgsc, absolutePath);
		}
	}

	public String getCGSCName(final String filename) {
		try {
			if (cp.m_iniCfg.sidplay2().getCgsc() == null
					|| cp.m_iniCfg.sidplay2().getCgsc().length() == 0) {
				return null;
			}
			final String canonicalPath = new File(filename).getCanonicalPath();
			final String cgscCanonicalPath = new File(
					cp.m_iniCfg.sidplay2().getCgsc()).getCanonicalPath();
			if (canonicalPath.startsWith(cgscCanonicalPath)) {
				final String name = canonicalPath.substring(cgscCanonicalPath.length()).replace('\\', '/');
				if (name.startsWith("/")) {
					return name;
				}
			}
		} catch (final IOException e) {
			e.printStackTrace();
		}
		return null;
	}

	public int getSongLength(final SidTune sidTuneMod) {
		final boolean databaseNameDefined = cp.m_iniCfg.sidplay2().getHvsc() != null && !"".equals(cp.m_iniCfg.sidplay2().getHvsc());
		if (!cp.m_database.isOpened() && databaseNameDefined) {
			cp.m_database.open(cp.m_iniCfg.sidplay2().getHvsc() + File.separator + "DOCUMENTS" + File.separator + "Songlength.txt");
		}
		final int length = cp.m_database.isOpened() ? cp.m_database.length(sidTuneMod) : -1;
		if (databaseNameDefined && length == -1) {
			System.err.println(cp.m_database.error());
		}
		return length;
	}

	public boolean getForceStereoTune() {
		return cp.m_engine.getConfig().forceStereoTune;
	}

	public void setForceStereoSid(final boolean forceForceStereoTune) {
		cp.m_iniCfg.emulation().setForceStereoTune(forceForceStereoTune);
		cp.m_engine.getConfig().forceStereoTune = forceForceStereoTune;
		restartC64();
	}

	public ChipModel getStereoSidModel() {
		return cp.m_iniCfg.emulation().getStereoSidModel();
	}

	public void setStereoSidModel(final ChipModel stereoSidModel) {
		cp.m_iniCfg.emulation().setStereoSidModel(stereoSidModel);
		cp.updateSidEmulation();
	}

	public String getDualSidBase() {
		return String.format("0x%4x", cp.m_engine.getConfig().dualSidBase);
	}

	public void setDualSidBase(final String hexBase) {
		cp.m_iniCfg.emulation().setDualSidBase(IniConfig.toInt(hexBase));
		cp.m_engine.getConfig().dualSidBase = IniConfig.toInt(hexBase);
	}

	public int getFrequency() {
		return cp.m_iniCfg.audio().getFrequency();
	}

	public void setFrequency(final int frequency) {
		cp.m_iniCfg.audio().setFrequency(frequency);
		cp.m_engine.getConfig().frequency = frequency;
	}

	public SamplingMethod getSamplingMethod() {
		return cp.m_iniCfg.audio().getSampling();
	}

	public void setSamplingMethod(final SamplingMethod sampling) {
		cp.m_iniCfg.audio().setSampling(sampling);
		cp.updateSidEmulation();
	}

	//
	// Helper methods
	//

	public int getFullSongLength(final SidTune currentTune) {
		final String md5 = currentTune.getMD5Digest();
		int length = 0;
		for (int i = 1; i <= currentTune.getInfo().songs; i++) {
			if (!cp.m_database.isOpened() && cp.m_iniCfg.sidplay2().getHvsc() != null) {
				cp.m_database.open(cp.m_iniCfg.sidplay2().getHvsc() + File.separator + "DOCUMENTS" + File.separator + "Songlengths.txt");
			}
			int curr_length = cp.m_database.length(md5, i);
			if (curr_length == -1) {
				System.err.println(cp.m_database.error());
				curr_length = 0;
			}
			length += curr_length;
		}
		return length;
	}

	public String getHVSCName(final String filename) {
		try {
			if (cp.m_iniCfg.sidplay2().getHvsc() == null
					|| cp.m_iniCfg.sidplay2().getHvsc().length() == 0) {
				return null;
			}
			final String canonicalPath = new File(filename).getCanonicalPath();
			final String hvscCanonicalPath = new File(cp.m_iniCfg.sidplay2().getHvsc()).getCanonicalPath();
			if (canonicalPath.startsWith(hvscCanonicalPath)) {
				final String name = canonicalPath.substring(hvscCanonicalPath.length()).replace('\\', '/');
				if (name.startsWith("/")) {
					return name;
				}
			}
		} catch (final IOException e) {
			e.printStackTrace();
		}
		return null;
	}

	public void setHVSC(final String dir) {
		final String oldHVSCDir = cp.m_iniCfg.sidplay2().getHvsc();
		cp.m_iniCfg.sidplay2().setHvsc(dir);
		support.firePropertyChange(PROP_HVSC, oldHVSCDir, dir);
	}

	public String getHVSCDir() {
		return cp.m_iniCfg.sidplay2().getHvsc();
	}

	public STILEntry getSTIL(final String filename) {
		final String name = getHVSCName(filename);
		if (null != name) {
			if (!stil.isOpened() && cp.m_iniCfg.sidplay2().getHvsc() != null) {
				stil.open(cp.m_iniCfg.sidplay2().getHvsc() + File.separator + "DOCUMENTS" + File.separator + "STIL.txt");
			}
			return stil.getSTIL(name);
		}
		return null;
	}

	public String getCGSCDir() {
		return cp.m_iniCfg.sidplay2().getCgsc();
	}

	public int getDefaultPlayLength() {
		return cp.m_iniCfg.sidplay2().getPlayLength();
	}

	public void setDefaultPlayLength(final int seconds) {
		cp.m_timer.length = seconds;
		cp.m_iniCfg.sidplay2().setPlayLength(seconds);
	}

	public boolean getSingleTrack() {
		return cp.m_iniCfg.sidplay2().isSingle();
	}

	public void setSingleTrack(final boolean single) {
		// save settings
		cp.m_iniCfg.sidplay2().setSingle(single);
		cp.setCurrentSingle(single);
	}

	public void setProxy(final boolean proxyEnabled, final String hostname, final String port) {
		// save settings
		cp.m_iniCfg.sidplay2().setEnableProxy(proxyEnabled);
		cp.m_iniCfg.sidplay2().setProxyHostname(hostname);
		cp.m_iniCfg.sidplay2().setProxyPort(port.length() > 0 ? Integer.valueOf(port) : 80);
	}

	public Proxy getProxy() {
		Proxy proxy;
		if (!cp.m_iniCfg.sidplay2().isEnableProxy()) {
			proxy = null;
		} else {
			final SocketAddress addr = new InetSocketAddress(cp.m_iniCfg.sidplay2().getProxyHostname(), cp.m_iniCfg.sidplay2().getProxyPort());
			proxy = new Proxy(Proxy.Type.HTTP, addr);
		}
		return proxy;
	}

	public favorites_section getFavorites() {
		return cp.m_iniCfg.favorites();
	}

	/**
	 * Create the GUI.
	 */
	private void createGUI() {
		// use system L&F
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (final Exception e) {
			e.printStackTrace();
		}

		setLayout(new BorderLayout());

		fButtonsPanel = new JPanel(new GridLayout(0, 4, 0, 0));
		add(fButtonsPanel, BorderLayout.NORTH);
		
		// create GUI
		tabbedPane = new JTabbedPane();
		fVideoView = new VideoView(this);
		addTab(tabbedPane, "Video", fVideoView);
		addTab(tabbedPane, "Sound Settings", new GeneralCanvas(this));
		addTab(tabbedPane, "Joystick", new JoystickTab());
		addTab(tabbedPane, "Emulation Settings", new SettingsPanel(this));
		addTab(tabbedPane, "Graphics", new GraphicalView(this));
		addTab(tabbedPane, "Memory", new MemoryView(this));
		addTab(tabbedPane, "STIL", new STILView(this));
		addTab(tabbedPane, "HVSC", new HVSCView(this));
		addTab(tabbedPane, "CGSC", new CGSCView(this));
		addTab(tabbedPane, "Favorites", new FavoritesView(this));
		addTab(tabbedPane, "Network SID Device", new HardSIDView(this));
		addTab(tabbedPane, "SIDDump", new SIDDumpView(this));
		addTab(tabbedPane, "SIDReg", new SIDRegView(this));
		addTab(tabbedPane, "Console", consoleView);
		addTab(tabbedPane, "About", new AboutView(this));
		add(tabbedPane, BorderLayout.CENTER);

		setLockModel(getLockModel());

		if (this.cp.m_iniCfg.sidplay2().getHvsc() != null) {
			support.firePropertyChange(PROP_HVSC, "", this.cp.m_iniCfg.sidplay2().getHvsc());
		}
		if (this.cp.m_iniCfg.sidplay2().getCgsc() != null) {
			support.firePropertyChange(PROP_CGSC, "", this.cp.m_iniCfg.sidplay2().getCgsc());
		}
		fLoadButton = new JButton("Load Program...", new ImageIcon(JSIDPlay2.class.getResource("icons/fileopen.png")));
		fLoadButton.setToolTipText(fLoadButton.getText());
		fLoadButton.setMnemonic(LOAD_KEY);
		fLoadButton.addActionListener(this);
		fButtonsPanel.add(fLoadButton);

		final JButton insertDiskButton = new JButton("Insert Disk...", new ImageIcon(JSIDPlay2.class.getResource("icons/floppy.png")));
		insertDiskButton.setToolTipText(insertDiskButton.getText());
		insertDiskButton.addActionListener(this);
		insertDiskButton.setMnemonic(DISK_KEY);
		fButtonsPanel.add(insertDiskButton);

		final JButton insertTapeButton = new JButton("Insert Tape...", new ImageIcon(JSIDPlay2.class.getResource("icons/tape.png")));
		insertTapeButton.setToolTipText(insertTapeButton.getText());
		insertTapeButton.addActionListener(this);
		insertTapeButton.setMnemonic(TAPE_KEY);
		fButtonsPanel.add(insertTapeButton);

		final JButton cartButton = new JButton("Insert Cart...", new ImageIcon(JSIDPlay2.class.getResource("icons/cart.png")));
		cartButton.setToolTipText(cartButton.getText());
		cartButton.addActionListener(this);
		cartButton.setMnemonic(CART_KEY);
		fButtonsPanel.add(cartButton);

		fPauseButton = new JButton("Pause", new ImageIcon(JSIDPlay2.class.getResource("icons/player_pause.png")));
		fPauseButton.setToolTipText(fPauseButton.getText());
		fPauseButton.setMnemonic(PAUSE_KEY);
		fPauseButton.addActionListener(this);
		fButtonsPanel.add(fPauseButton);

		fPreviousButton = new JButton("Previous Song", new ImageIcon(JSIDPlay2.class.getResource("icons/1leftarrow.png")));
		fPreviousButton.setToolTipText(fPreviousButton.getText());
		fPreviousButton.setMnemonic(PREVIOUS_KEY);
		fPreviousButton.addActionListener(this);
		fButtonsPanel.add(fPreviousButton);

		fNextButton = new JButton("Next Song", new ImageIcon(JSIDPlay2.class.getResource("icons/1rightarrow.png")));
		fNextButton.setToolTipText(fNextButton.getText());
		fNextButton.setMnemonic(NEXT_KEY);
		fNextButton.addActionListener(this);
		fButtonsPanel.add(fNextButton);

		final JButton resetButton = new JButton("Reset C64", new ImageIcon(JSIDPlay2.class.getResource("icons/reset.png")));
		resetButton.setToolTipText(resetButton.getText());
		resetButton.setMnemonic(RESET_KEY);
		resetButton.addActionListener(this);
		fButtonsPanel.add(resetButton);

		final JButton normalButton = new JButton("Normal Speed", new ImageIcon(JSIDPlay2.class.getResource("icons/mini/1rightarrow.png")));
		normalButton.setToolTipText(normalButton.getText());
		normalButton.setMnemonic(NORMAL_KEY);
		normalButton.addActionListener(this);
		fButtonsPanel.add(normalButton);

		final JButton fastButton = new JButton("Fast Forward", new ImageIcon(JSIDPlay2.class.getResource("icons/mini/2rightarrow.png")));
		fastButton.setToolTipText(fastButton.getText());
		fastButton.setMnemonic(FASTER_KEY);
		fastButton.addActionListener(this);
		fButtonsPanel.add(fastButton);

		fAddFavButton = new JButton("Add to Favorites...", new ImageIcon(JSIDPlay2.class.getResource("icons/bookmark_add.png")));
		fAddFavButton.setToolTipText(fAddFavButton.getText());
		fAddFavButton.setMnemonic(ADD_FAV_KEY);
		fAddFavButton.addActionListener(this);
		fButtonsPanel.add(fAddFavButton);
		
		final JButton freezeButton = new JButton("Freeze Cart.", new ImageIcon(JSIDPlay2.class.getResource("icons/freeze.png")));
		freezeButton.setToolTipText(freezeButton.getText());
		freezeButton.setMnemonic(FREEZE_KEY);
		freezeButton.addActionListener(this);
		fButtonsPanel.add(freezeButton);

		// StatusBar
		final Box sbar = new Box(BoxLayout.X_AXIS);
		sbar.setBackground(Color.LIGHT_GRAY);
		fProgressBar = new JProgressBar(0, 100) {
			@Override
			public boolean isVisible() {
				return getValue() != 100;
			}
		};
		fProgressBar.setValue(100);
		fProgressBar.setMaximumSize(new Dimension(50, 25));
		sbar.add(fProgressBar);
		sbar.add(Box.createRigidArea(new Dimension(10, 0)));
		fStatusLine = new JLabel();
		sbar.add(fStatusLine);
		sbar.add(Box.createHorizontalGlue());
		add(sbar, BorderLayout.SOUTH);

		tabbedPane.transferFocus();
	}

	private void addTab(final JTabbedPane tabPane, final String title, final TuneTab tab) {
		tab.createContents();
		tabPane.add(title, tab);
		tabs.add(tab);
	}

	/**
	 * Update the buttons according to the currently played song number
	 */
	private void updateButtonTexts() {
		fPreviousButton.setEnabled(cp.getCurrentSongCount() != 0);		
		fNextButton.setEnabled(cp.getCurrentSongCount() != 0);
		
		int prevSong = cp.getCurrentSong() - 1;
		if (prevSong < 1) {
			prevSong = cp.getCurrentSongCount();
		}
		fPreviousButton.setText(String.format("Prev. Song (%d/%d)", prevSong, cp.getCurrentSongCount()));
		fPreviousButton.setToolTipText(fPreviousButton.getText());

		int nextSong = cp.getCurrentSong() + 1;
		if (nextSong > cp.getCurrentSongCount()) {
			nextSong = 1;
		}
		fNextButton.setText(String.format("Next Song (%d/%d)", nextSong, cp.getCurrentSongCount()));
		fNextButton.setToolTipText(fNextButton.getText());

		fPauseButton.setText("Pause");
		fPauseButton.setToolTipText(fPauseButton.getText());
	}

	private void setUI() {
		for (final TuneTab tab : tabs) {
			tab.setTune(cp.m_engine, cp.m_tune);
		}
		setTune();

		updateButtonTexts();
	}

	public void setTune() {
		final SidTune tune = cp.m_tune;
		fAddFavButton.setEnabled(tune != null);
		fSongLength = null;
		if (tune == null) {
			return;
		}

		int stilIndex = 0;
		for (int i = 0; i < tabbedPane.getComponentCount(); i++) {
			if (tabbedPane.getComponent(i) instanceof STILView) {
				stilIndex = i;
			}
		}
		if (getSTIL(tune.getInfo().filename) != null) {
			tabbedPane.setIconAt(stilIndex, stilIcon);
		} else {
			tabbedPane.setIconAt(stilIndex, stilNoIcon);
		}
		final int songLength = getSongLength(tune);
		if (songLength > 0) {
			fSongLength = String.format("%02d:%02d", (songLength / 60 % 100), (songLength % 60));
		}
		tabbedPane.repaint();
	}

	private void setStatusLine() {
		Runtime runtime = Runtime.getRuntime();
		int totalMemory = (int) (runtime.totalMemory() / 1000000);
		int freeMemory = (int) (runtime.freeMemory() / 1000000);
		
		C1541 c1541 = cp.m_engine.getC1541();
		boolean motor = fDriveSound && cp.state() == ConsolePlayer.playerRunning && c1541.getMotor();
		if (!oldMotor && motor) {
			motorSound.loop();
		} else if (oldMotor && !motor) {
			motorSound.stop();
		}
		oldMotor = motor;
		
		int track = c1541.getTrack();
		if (fDriveSound && oldTrack != track) {
			trackSound.play();
		}
		oldTrack = track;
		
		Datasette datasette = cp.m_engine.getDatasette();
		if (datasette.getMotor()) {
			setProgressBar(datasette.getProgress());
		}

		final int seconds = cp.m_engine.time();
		String text = 
			String.format("Time: %02d:%02d", seconds / 60 % 100, seconds % 60)
			+ (fSongLength != null ? " / " + fSongLength : "")
			+ ", JRE V" + System.getProperty("java.version")
			+ ", Mem: " + (totalMemory - freeMemory) + "MB / " + totalMemory + "MB"
			+ ", Version 2.0"
			+ String.format(", floppy: %03d%s", c1541.getTrack(), c1541.getLED() ? "+" : "-")
			+ String.format(", tape: %03d%s", datasette.getCounter(), datasette.getMotor() ? "+" : "-");
		
		fStatusLine.setText(text);
	}

	private void startC64() {
		fPlayerThread = new Thread(playerRunnable);
		fPlayerThread.setPriority(Thread.MAX_PRIORITY);
		fPlayerThread.start();
	}

	public void playFile(final File file, final Component comp) {
		stopC64();
		// Next time player is used, the track is reset
		cp.resetTrack();
		setPlayingIcon(comp);
		if (cp.loadTune(file != null ? file.getAbsolutePath() : null) < 0) {
			return;
		}
		startC64();
	}

	private void setPlayingIcon(final Component comp) {
		for (int i = 0; i < tabbedPane.getTabCount(); i++) {
			if (tabbedPane.getIconAt(i) == PLAY_ICON) {
				tabbedPane.setIconAt(i, null);
			}
			if (comp == tabbedPane.getComponentAt(i)) {
				tabbedPane.setIconAt(i, PLAY_ICON);
			}
		}
		tabbedPane.repaint();
	}

	public String getPlayingFile() {
		return cp.m_filename;
	}

	//
	// Action listener
	//

	public void actionPerformed(final ActionEvent ae) {
		try {
			/* First row buttons */
		final String cmd = ae.getActionCommand();
		if (cmd.startsWith("Load Program")) {
			final JFileChooser fileDialog = new JFileChooser(cp.m_iniCfg.sidplay2().getLastDirectory());
			fileDialog.setFileFilter(new C64FileFilter());
			final int rc = fileDialog.showOpenDialog(this);
			if (rc == JFileChooser.APPROVE_OPTION
					&& fileDialog.getSelectedFile() != null) {
				final File selectedFile = fileDialog.getSelectedFile();
				cp.m_iniCfg.sidplay2().setLastDirectory(selectedFile.getParentFile().getAbsolutePath());
				// play tune
				playFile(selectedFile, null);
			}
		} else if (cmd.startsWith("Insert Disk")) {
				final JFileChooser fileDialog = new JFileChooser(cp.m_iniCfg
						.sidplay2().getLastDirectory());
				fileDialog.setFileFilter(new D64FileFilter());
				final int rc = fileDialog.showOpenDialog(JSIDPlay2.this);
				if (rc == JFileChooser.APPROVE_OPTION
						&& fileDialog.getSelectedFile() != null) {
					final File selectedFile = fileDialog.getSelectedFile();
					cp.m_iniCfg.sidplay2().setLastDirectory(selectedFile
							.getParentFile().getAbsolutePath());
					// attach selected disk drive
					final C1541 c1541 = cp.m_engine.getC1541();
					c1541.insertDisk(selectedFile.getAbsolutePath());
				}
		} else if (cmd.startsWith("Insert Tape")) {
			final Datasette datasette = cp.m_engine.getDatasette();
			// press eject on tape
			datasette.ejectTape();

			final JFileChooser fileDialog = new JFileChooser(cp.m_iniCfg.sidplay2().getLastDirectory());
			fileDialog.setFileFilter(new TapeFileFilter());
			final int rc = fileDialog.showOpenDialog(JSIDPlay2.this);
			if (rc == JFileChooser.APPROVE_OPTION
					&& fileDialog.getSelectedFile() != null) {
				final File selectedFile = fileDialog.getSelectedFile();
				cp.m_iniCfg.sidplay2().setLastDirectory(selectedFile.getParentFile().getAbsolutePath());
				if (!selectedFile.getName().toLowerCase().endsWith(".tap")) {
					try {
						// convert to tape first
						File tmp = File.createTempFile("wav2prg", ".tap");
						tmp.deleteOnExit();
						String[] args = new String[] {
								selectedFile.getAbsolutePath(),
								tmp.getAbsolutePath() };
						PRG2TAP.main(args);
						// attach temporary tape
						datasette.insertTape(tmp.getAbsolutePath());
					} catch (IOException e1) {
						e1.printStackTrace();
					}
				} else {
					// attach selected tape
					datasette.insertTape(selectedFile.getAbsolutePath());
				}
			}
		} else if (cmd.startsWith("Insert Cart")) {
			final JFileChooser fileDialog = new JFileChooser(cp.m_iniCfg.sidplay2().getLastDirectory());
			fileDialog.setFileFilter(new CartFileFilter());
			final int rc = fileDialog.showOpenDialog(this);
			if (rc == JFileChooser.CANCEL_OPTION) {
				cp.m_engine.getC64().insertCartridge(null);
				reset(null);
			} else {
				if (rc == JFileChooser.APPROVE_OPTION && fileDialog.getSelectedFile() != null) {
					cp.m_iniCfg.sidplay2().setLastDirectory(fileDialog.getSelectedFile().getParentFile().getAbsolutePath());
					final File selectedFile = fileDialog.getSelectedFile();
					try {
						cp.m_engine.getC64().insertCartridge(selectedFile.getAbsolutePath());
						gotoVideoTab();
						reset(null);
					} catch (FileNotFoundException e) {
						e.printStackTrace();
					}
				}
			}
			
			/* second row buttons */
		} else if (cmd.startsWith("Pause")) {
			fPauseButton.setText("Play");
			fPauseButton.setToolTipText(fPauseButton.getText());
			cp.pause();
		} else if (cmd.startsWith("Play")) {
			fPauseButton.setText("Pause");
			fPauseButton.setToolTipText(fPauseButton.getText());
			if (fPlayerThread.isAlive()) {
				cp.pause();
			} else {
				startC64();
			}
		} else if (cmd.startsWith("Prev")) {
			cp.previousSong();
		} else if (cmd.startsWith("Next")) {
			cp.nextSong();
		} else if (cmd.startsWith("Reset")) {
			gotoVideoTab();
			reset(null);
			
			/* third row buttons */
		} else if (cmd.startsWith("Normal")) {
			cp.normalSpeed();
		} else if (cmd.startsWith("Fast")) {
			cp.fastForward();
		} else if (cmd.startsWith("Add")) {
			addToFavorites((Component) ae.getSource());
		} else if (cmd.startsWith("Freeze")) {
			cp.m_engine.getC64().getCartridge().freeze();

			/* unknown command (shouldn't happen, of course) */
		} else {
			throw new RuntimeException("Unknown command: " + cmd);
		}
		} catch (Exception e) {
			CharArrayWriter cw = new CharArrayWriter();
			e.printStackTrace(new PrintWriter(cw));
			JOptionPane.showMessageDialog(this, cw.toString());
		}
	}

	public void setDriveSound(final boolean driveSound) {
		fDriveSound = driveSound;
	}

	public void reset(final Component comp) {
		playFile(null, comp);
	}

	public void gotoVideoTab() {
		int videoIndex = 0;
		for (int i = 0; i < tabbedPane.getComponentCount(); i++) {
			if (tabbedPane.getComponent(i) instanceof VideoView) {
				videoIndex = i;
				break;
			}
		}
		tabbedPane.setSelectedIndex(videoIndex);
	}

	public void restartC64() {
		stopC64();
		startC64();
	}

	private void addToFavorites(final Component component) {
		fContextPopup = new JPopupMenu("Add to Favorites");
		support.firePropertyChange(IFavoriteTabNames.class.getName(), null, new IFavoriteTabNames() {

			public void setFavoriteTabNames(final String[] names, final String selected) {
				for (int i = 0; i < names.length; i++) {
					final int index = i;
					final String title = names[i];
					final JMenuItem tabItem = new JMenuItem(title);
					tabItem.addActionListener(new ActionListener() {

						public void actionPerformed(final ActionEvent e) {
							support.firePropertyChange(IGetFavorites.class.getName(), null, new IGetFavorites() {

								public int getIndex() {
									return index;
								}

								public void setFavorites(final IFavorites favorites) {
									final String filename = cp.m_tune.getInfo().filename;
									if (filename == null) {
										return;
									}
									favorites.addToFavorites(new File[] { new File(filename) });
								}

							});
						}

					});
					fContextPopup.add(tabItem);
				}
			}

		});
		fContextPopup.show(component, 0, 0);
	}

	public static void main(final String[] args) {
		final JSIDPlay2 sp = new JSIDPlay2(args);
//		JMenuBar menuBar = new JSIDPlayMenu();
//		sp.setJMenuBar(menuBar );
		sp.init();
		
		final JFrame window = new JFrame(FRAME_TITLE);
		window.setIconImage(WINDOW_ICON.getImage());
		window.add(sp);

		/* Set default position and size */
		final sidplay2_section section = sp.cp.m_iniCfg.sidplay2();
		if (section.getFrameX() != -1 && section.getFrameY() != -1) {
			window.setLocation(section.getFrameX(), section.getFrameY());
			window.setSize(section.getFrameWidth(), section.getFrameHeight());
		} else {
			window.setSize(620, 780);
		}
	
		/* Handle close button: execute applet's stop() */
		window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		window.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(final WindowEvent e) {
				sp.stopC64();
				sp.stop();
			}
		});
		
		/* Handle resize: record new window size */
		window.addComponentListener(new ComponentListener() {
			public void componentShown(final ComponentEvent e) {
			}
			
			public void componentResized(final ComponentEvent e) {
				final Dimension size = window.getSize();
				sp.cp.m_iniCfg.sidplay2().setFrameWidth(size.width);
				sp.cp.m_iniCfg.sidplay2().setFrameHeight(size.height);
			}
			
			public void componentMoved(final ComponentEvent e) {
				final Point loc = window.getLocation();
				sp.cp.m_iniCfg.sidplay2().setFrameX(loc.x);
				sp.cp.m_iniCfg.sidplay2().setFrameY(loc.y);
			}
			
			public void componentHidden(final ComponentEvent e) {
			}
		});
		window.setVisible(true);

		/* Start applet */
		sp.start();
	}

	public void propertyChange(final PropertyChangeEvent evt) {
		if (IAddFavoritesTab.class.getName().equals(evt.getPropertyName())) {
			final IAddFavoritesTab ifObj = (IAddFavoritesTab) evt.getNewValue();
			final favorites_section favorites = cp.m_iniCfg.favorites();
			final String[] newFavoritesTitles = new String[favorites.getFavorites_titles().length + 1];
			for (int i = 0; i < favorites.getFavorites_titles().length; i++) {
				newFavoritesTitles[i] = favorites.getFavorites_titles()[i];
			}
			newFavoritesTitles[favorites.getFavorites_titles().length] = ifObj.getTitle();
			favorites.setFavorites_titles(newFavoritesTitles);
			final String[] newFavoritesFilenames = new String[favorites.getFavorites_filenames().length + 1];
			for (int i = 0; i < favorites.getFavorites_filenames().length; i++) {
				newFavoritesFilenames[i] = favorites.getFavorites_filenames()[i];
			}
			newFavoritesFilenames[favorites.getFavorites_filenames().length] = null;
			favorites.setFavorites_filenames(newFavoritesFilenames);
			// System.err.println("Add title=" + ifObj.getTitle());
		} else if (IRemoveFavoritesTab.class.getName().equals(evt.getPropertyName())) {
			final IRemoveFavoritesTab ifObj = (IRemoveFavoritesTab) evt.getNewValue();
			final favorites_section favorites = cp.m_iniCfg.favorites();
			final String[] newFavoriteTitles = new String[favorites.getFavorites_titles().length - 1];
			int j = 0;
			for (int i = 0; i < favorites.getFavorites_titles().length; i++) {
				if (i != ifObj.getIndex()) {
					newFavoriteTitles[j] = favorites.getFavorites_titles()[i];
					j++;
				}
			}
			favorites.setFavorites_titles(newFavoriteTitles);

			final String[] newFavoriteFilenames = new String[favorites.getFavorites_filenames().length - 1];
			int k = 0;
			for (int i = 0; i < favorites.getFavorites_filenames().length; i++) {
				if (i != ifObj.getIndex()) {
					newFavoriteFilenames[k] = favorites.getFavorites_filenames()[i];
					k++;
				}
			}
			favorites.setFavorites_filenames(newFavoriteFilenames);

			// System.err.println("Remove index=" + ifObj.getIndex() + ",
			// title="
			// + ifObj.getTitle());
		} else if (IChangeFavoritesTab.class.getName().equals(evt.getPropertyName())) {
			final IChangeFavoritesTab ifObj = (IChangeFavoritesTab) evt.getNewValue();
			final favorites_section favorites = cp.m_iniCfg.favorites();
			String[] newTitles = new String[favorites.getFavorites_titles().length];
			for (int i = 0; i < newTitles.length; i++) {
				if (i == ifObj.getIndex()) {
					newTitles[i] = ifObj.getTitle();
				} else {
					newTitles[i] = favorites.getFavorites_titles()[i];
				}
			}
			favorites.setFavorites_titles(newTitles);
			String[] newFilenames = new String[favorites.getFavorites_filenames().length];
			for (int i = 0; i < newFilenames.length; i++) {
				if (i == ifObj.getIndex()) {
					newFilenames[i] = ifObj.getFileName();
				} else {
					newFilenames[i] = favorites.getFavorites_filenames()[i];
				}
			}
			if (ifObj.getFileName() != null) {
				final String relativePath = RelativePath.getRelativePath(new File(System.getProperty("user.dir")), new File(ifObj.getFileName()));
				if (relativePath != null) {
					newFilenames[ifObj.getIndex()] = relativePath;
				}
			}
			favorites.setFavorites_filenames(newFilenames);
			if (ifObj.isSelected()) {
				favorites.setFavorites_current(ifObj.getTitle());
			}
			// System.err.println("Change index=" + ifObj.getTitle()
			// + ", filename=" + ifObj.getFileName());
		} else if (ITuneStateChanged.class.getName().equals(evt.getPropertyName())) {
			setPlayingIcon(null);
			fPauseButton.setText("Play");
			repaint();
		}
	}

	public SIDEmu getSIDEmu(final int num) {
		return cp.m_engine.getSid(num);
	}

	public SID getSID(final int num) {
		final SIDEmu sid = cp.m_engine.getSid(num);
		if (sid instanceof ReSID) {
			return ((ReSID) sid).sid();
		} else {
			return null;
		}
	}
	
	public void setProgressBar(int value) {
		fProgressBar.setValue(value);
	}
}
