package applet;

import java.io.File;
import java.io.IOException;

import javax.swing.filechooser.FileFilter;

import libpsid64.Psid64;

public class SidTuneConverter {
	/**
	 * file filter for tunes
	 */
	private final FileFilter fFileFilter = new TuneFileFilter();

	private final JSIDPlay2 fSidplay2;

	public SidTuneConverter(final JSIDPlay2 sidplay2) {
		fSidplay2 = sidplay2;
	}

	public void convertFiles(final File[] files, final File target) {
		for (final File file : files) {
			if (file.isDirectory()) {
				convertFiles(file.listFiles(), target);
			} else {
				if (fFileFilter.accept(file)) {
					convertToPSID64(file, target);
					break;
				}
			}
		}
	}

	private void convertToPSID64(final File file, final File target) {
		final String filename = file.getAbsolutePath();
		final Psid64 psid64 = new Psid64();
		psid64.setHVSC(fSidplay2.getHVSCDir());
		try {
			psid64.setStilEntry(fSidplay2.getSTIL(file.getCanonicalPath()));
		} catch (final IOException e) {
		}
		if (!psid64.load(filename)) {
			System.err.println(psid64.getStatus());
			return;
		}
		if (!psid64.convert()) {
			System.err.println(psid64.getStatus());
			return;
		}
		if (!psid64.save(new File(target, ext(file, ".prg")).getAbsolutePath())) {
			System.err.println(psid64.getStatus());
			return;
		}
	}

	private String ext(final File file, final String ext) {
		final String name = file.getName();
		final int lastIndexOf = name.lastIndexOf('.');
		if (lastIndexOf != -1) {
			return name.substring(0, lastIndexOf) + ext;
		} else {
			return name + ext;
		}
	}
}
