package applet.disassembler;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.io.FileOutputStream;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;
import libsidplay.sidtune.SidTuneInfo;
import applet.JSIDPlay2;
import applet.TuneTab;

@SuppressWarnings("serial")
public class MemoryView extends TuneTab {
	// jump targets
	protected JButton fDriverAddress = new JButton("?");
	protected JButton fInitAddress = new JButton("?");
	protected JButton fLoadAddress = new JButton("?");
	protected JButton fPlayAddress = new JButton("?");

	protected JSIDPlay2 fSidplay2;

	//
	// GUI elements
	//

	protected JTable fTable;
	protected JScrollPane fTableScrollPane;
	protected JTextField address;
	protected MyTableModel memoryTableModel;
	private JTextField saveStart;
	private JTextField saveEnd;

	public MemoryView(final JSIDPlay2 sidplay) {
		this.fSidplay2 = sidplay;
		sidplay.support.addPropertyChangeListener(this);
	}

	private static void bolden(final JComponent lbl) {
		lbl.setFont(lbl.getFont().deriveFont(Font.BOLD));
	}

	@Override
	public void createContents() {
		setLayout(new GridLayout(1, 2));

		final JPanel memTablePanel = new JPanel();
		memTablePanel.setLayout(new GridLayout(2, 1));
		add(memTablePanel);

		fTable = new JTable();

		fTableScrollPane = new JScrollPane(fTable);
		fTableScrollPane.setBorder(new TitledBorder("Memory Browser"));
		add(fTableScrollPane);
		add(memTablePanel);
		
		final JPanel gotoPanel = new JPanel();
		gotoPanel.setBorder(new TitledBorder("Jump"));
		gotoPanel.setLayout(new GridBagLayout());

		final GridBagConstraints cLabel = new GridBagConstraints();
		cLabel.weightx = 0.5;
		cLabel.fill = GridBagConstraints.BOTH;

		final GridBagConstraints cControl = new GridBagConstraints();
		cControl.weightx = 0.5;
		cControl.fill = GridBagConstraints.BOTH;
		cControl.anchor = GridBagConstraints.LINE_START;
		cControl.gridwidth = GridBagConstraints.REMAINDER;

		final JLabel addr = new JLabel("Address:");
		bolden(addr);
		gotoPanel.add(addr, cLabel);

		address = new JTextField("0x0000");
		address.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				gotoMem(address.getText());
			}
		});
		gotoPanel.add(address, cControl);

		final JLabel driverLbl = new JLabel("Driver Address:");
		bolden(driverLbl);
		gotoPanel.add(driverLbl, cLabel);
		gotoPanel.add(fDriverAddress, cControl);

		final JLabel loadAdrLbl = new JLabel("Load Address:");
		bolden(loadAdrLbl);
		gotoPanel.add(loadAdrLbl, cLabel);
		gotoPanel.add(fLoadAddress, cControl);

		final JLabel initAdrLbl = new JLabel("Init Address:");
		bolden(initAdrLbl);
		gotoPanel.add(initAdrLbl, cLabel);
		gotoPanel.add(fInitAddress, cControl);

		final JLabel playLbl = new JLabel("Player Address:");
		bolden(playLbl);
		gotoPanel.add(playLbl, cLabel);
		gotoPanel.add(fPlayAddress, cControl);

		for (final JButton b : new JButton[] { fDriverAddress, fLoadAddress, fInitAddress, fPlayAddress }) {
			b.addActionListener(new ActionListener() {
				public void actionPerformed(final ActionEvent e) {
					final JButton btn = (JButton) e.getSource();
					final String value = btn.getText();
					/* parse $1234 -> 1234 */
					final int hexValueBegin = value.indexOf('$');
					if (hexValueBegin > -1) {
						gotoMem("0x" + value.substring(hexValueBegin + 1, hexValueBegin + 1 + 4));
					}
				}
			});
		}
		
		memTablePanel.add(gotoPanel);

		final JPanel savePanel = new JPanel();
		savePanel.setBorder(new TitledBorder("Save memory"));
		savePanel.setLayout(new GridBagLayout());
		
		final JLabel saveStartLabel = new JLabel("Start:");
		bolden(saveStartLabel);
		saveStart = new JTextField();
		savePanel.add(saveStartLabel, cLabel);
		savePanel.add(saveStart, cControl);
		
		final JLabel saveEndLabel = new JLabel("End:");
		bolden(saveEndLabel);
		saveEnd = new JTextField();
		savePanel.add(saveEndLabel, cLabel);
		savePanel.add(saveEnd, cControl);
		
		final JButton saveButton = new JButton("Save...");
		savePanel.add(new JLabel(""), cLabel);
		savePanel.add(saveButton, cControl);
		saveButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent buttonDown) {
				JFileChooser chooser = new JFileChooser(fSidplay2.cp.m_iniCfg.sidplay2().getLastDirectory());
				int rc = chooser.showOpenDialog(getTopLevelAncestor());
				if (rc == JFileChooser.APPROVE_OPTION) {
					try {
						FileOutputStream fos = new FileOutputStream(chooser.getSelectedFile());
						int start = Integer.decode(saveStart.getText());
						int end = Integer.decode(saveEnd.getText()) + 1;
						start &= 0xffff;
						end &= 0xffff;
						fos.write(start & 0xff);
						fos.write(start >> 8);
						byte[] ram = fSidplay2.cp.m_engine.getC64().getRAM();
						while (start != end) {
							fos.write(ram[start]);
							start = start + 1 & 0xffff;
						}
					}
					catch (Exception e) {
						JOptionPane.showMessageDialog(getTopLevelAncestor(), e.getMessage());
					}
				}
				
			}
		});
		
		memTablePanel.add(savePanel);
	}

	protected void gotoMem(final String destination) {
		int dest = 0;
		try {
			dest = Integer.decode(destination) & 0xffff;
		}
		catch (final NumberFormatException e) {
		}
		address.setText(String.format("0x%04x", dest));
		fTableScrollPane.getViewport().setViewPosition(fTable.getCellRect(fTable.convertRowIndexToView(dest), 0, true).getLocation());
		fTable.repaint();
	}

	public void propertyChange(final PropertyChangeEvent arg0) {
	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
		memoryTableModel = new MyTableModel(this, m_engine.getC64().getRAM());
		fTable.setModel(memoryTableModel);
		fTable.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
		fTable.getColumnModel().getColumn(MyTableModel.COL_ADDR).setPreferredWidth(50);
		fTable.getColumnModel().getColumn(MyTableModel.COL_BYTES).setPreferredWidth(100);
		fTable.getColumnModel().getColumn(MyTableModel.COL_INSTR).setPreferredWidth(100);
		fTable.getColumnModel().getColumn(MyTableModel.COL_OPS).setPreferredWidth(100);
		fTable.getColumnModel().getColumn(MyTableModel.COL_CYCLES).setPreferredWidth(50);
		fTable.repaint();

		if (m_tune == null) {
			return;
		}
		final SidTuneInfo tuneInfo = m_tune.getInfo();
		// Display PSID Driver location
		String line;
		if (tuneInfo.determinedDriverAddr == 0) {
			line = "N/A";
		} else {
			line = String.format("$%04x - $%04x", tuneInfo.determinedDriverAddr, tuneInfo.determinedDriverAddr + tuneInfo.determinedDriverLength - 1);
		}
		fDriverAddress.setText(line);

		// Display PSID Load address
		line = String.format("$%04x - $%04x", tuneInfo.loadAddr,  tuneInfo.loadAddr + tuneInfo.c64dataLen - 1);
		fLoadAddress.setText(line);

		// Display PSID Init address
		if (tuneInfo.playAddr == 0xffff) {
			line = String.format("SYS $%04x", tuneInfo.initAddr);
		} else {
			line = String.format("$%04x", tuneInfo.initAddr);
		}
		fInitAddress.setText(line);

		// Display PSID Play address
		if (tuneInfo.playAddr == 0xffff) {
			line = "N/A";
		} else {
			line = String.format("$%04x", tuneInfo.playAddr);
		}
		fPlayAddress.setText(line);
	}
}
