package applet.favorites;

import java.awt.Color;
import java.awt.Component;
import java.io.File;
import java.io.IOException;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JTable;
import javax.swing.table.DefaultTableCellRenderer;

import applet.JSIDPlay2;

@SuppressWarnings("serial")
public class FavoritesCellRenderer extends DefaultTableCellRenderer {

	private Icon fStilIcon;

	private Icon fStilNoIcon;

	@Override
	public Component getTableCellRendererComponent(JTable table, Object value,
			boolean isSelected, boolean hasFocus, int row, int column) {
		if (column == 0) {
			if (value != null) {
				File file = fModel.getFile(value);
				try {
					String filename = file.getCanonicalPath();
					if (!file.exists()) {
						setBackground(Color.RED);
						setToolTipText("File not found!");
					} else if (isCurrentlyPlayed(filename)) {
						setBackground(Color.LIGHT_GRAY);
						setToolTipText(null);
					} else {
						setBackground(Color.WHITE);
						setToolTipText(null);
					}
				} catch (IOException e1) {
					setBackground(Color.WHITE);
					setToolTipText(null);
				}
				try {
					if (fSidplay2.getSTIL(file.getCanonicalPath()) != null) {
						// set STIL icon
						if (fStilIcon == null) {
							fStilIcon = new ImageIcon(JSIDPlay2.class
									.getResource("icons/stil.png"));
						}
						setIcon(fStilIcon);
					} else {
						// set NO STIL icon
						if (fStilNoIcon == null) {
							fStilNoIcon = new ImageIcon(JSIDPlay2.class
									.getResource("icons/stil_no.png"));
						}
						setIcon(fStilNoIcon);
					}
				} catch (IOException e) {
					setIcon(null);
				}
			}
		} else {
			setIcon(null);
		}
		return super.getTableCellRendererComponent(table, value, isSelected,
				hasFocus, row, column);
	}

	private JSIDPlay2 fSidplay2;
	private FavoritesModel fModel;

	public FavoritesCellRenderer(JSIDPlay2 sidplay2, FavoritesModel model) {
		fSidplay2 = sidplay2;
		fModel = model;
	}

	/**
	 * Check if the current tree element is being played
	 * 
	 * @param file
	 * 
	 * @return is it played now?
	 */
	private boolean isCurrentlyPlayed(String file) {
		if (fSidplay2.getPlayingFile() == null || file == null) {
			return false;
		}
		try {
			return new File(fSidplay2.getPlayingFile()).getCanonicalPath().equals(new File(file).getCanonicalPath());
		} catch (IOException e) {
			return false;
		}
	}
}
