package applet.favorites;

import java.io.File;
import java.util.ArrayList;

import javax.swing.table.DefaultTableModel;

import applet.JSIDPlay2;
import applet.sidtune.SidTuneInfoCache;

@SuppressWarnings("serial")
public class FavoritesModel extends DefaultTableModel {

	public static final String HVSC_PREFIX = "<HVSC>";

	public static final String CGSC_PREFIX = "<CGSC>";

	private final SidTuneInfoCache infoCache;

	ArrayList<Integer> propertyIndices = new ArrayList<Integer>(1);

	private final JSIDPlay2 fSidPlay;

	public FavoritesModel(JSIDPlay2 sidplay2) {
		super(new Object[] { "Filename" }, 0);
		this.fSidPlay = sidplay2;
		infoCache = new SidTuneInfoCache(sidplay2);
		propertyIndices.add(0);
	}

	@Override
	public String getColumnName(int column) {
		int modelColumn = propertyIndices.get(column);
		if (modelColumn == 0) {
			return "Path name";
		} else {
			return SidTuneInfoCache.SIDTUNE_INFOS[modelColumn-1];
		}
	}

	@Override
	public Class<?> getColumnClass(int column) {
		if (column == 0) {
			return super.getColumnClass(column);
		}
		int modelColumn = propertyIndices.get(column);
		return SidTuneInfoCache.SIDTUNE_TYPES[modelColumn-1];
	}

	@Override
	public Object getValueAt(int row, int column) {
		if (column == 0) {
			return super.getValueAt(row, column);
		}
		int modelColumn = propertyIndices.get(column);
		String pathName = getFile(getValueAt(row, 0)).getAbsolutePath();
		return infoCache.getInfo(pathName)[modelColumn-1];
	}

	@Override
	public boolean isCellEditable(int row, int column) {
		return false;
	}

	public boolean contains(Object obj, int column) {
		int modelColumn = propertyIndices.get(column);
		for (int i = 0; i < getRowCount(); i++) {
			Object value = getValueAt(i, modelColumn);
			String pathName = getFile(value).getAbsolutePath();
			if (pathName.equals(obj.toString())) {
				return true;
			}
		}
		return false;
	}

	@Override
	public void addColumn(Object columnName) {
		int modelIndex = 1;
		for (int i = 0; i < SidTuneInfoCache.SIDTUNE_INFOS.length; i++) {
			if (columnName.equals(SidTuneInfoCache.SIDTUNE_INFOS[i])) {
				modelIndex = i+1;
			}
		}
		// System.err.println("Add: " + modelIndex);
		propertyIndices.add(modelIndex);
		super.addColumn(columnName);
		// for (int i = 0; i < propertyIndices.size(); i++) {
		// System.err.println("aa: " + propertyIndices.get(i));
		// }
	}

	public void removeColumn(int column) {
		int modelIndex = propertyIndices.get(column);
		// System.err.println("Remove: " + modelIndex);
		for (int i = 0; i < propertyIndices.size(); i++) {
			if (propertyIndices.get(i).equals(modelIndex)) {
				propertyIndices.remove(propertyIndices.get(i));
				setColumnCount(getColumnCount() - 1);
				break;
			}
		}
		// for (int i = 0; i < propertyIndices.size(); i++) {
		// System.err.println("ar: " + propertyIndices.get(i));
		// }
	}

	@Override
	public int getColumnCount() {
		return propertyIndices.size();
	}

	public File getFile(Object value) {
		String filename = value.toString();
		if (filename.startsWith(HVSC_PREFIX)) {
			filename = fSidPlay.getHVSCDir()
					+ filename.substring(HVSC_PREFIX.length());
		} else if (filename.startsWith(CGSC_PREFIX)) {
			filename = fSidPlay.getCGSCDir()
					+ filename.substring(CGSC_PREFIX.length());
		}
		return new File(filename);
	}

}
