package applet.favorites;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.dnd.DropTarget;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.RowFilter;
import javax.swing.RowSorter;
import javax.swing.SortOrder;
import javax.swing.SwingUtilities;
import javax.swing.RowSorter.SortKey;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.RowSorterEvent;
import javax.swing.event.RowSorterListener;
import javax.swing.plaf.basic.BasicButtonUI;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import applet.JSIDPlay2;
import applet.RelativePath;
import applet.SidTuneConverter;
import applet.dnd.FileDrop;
import applet.events.favorites.IFavoriteTabNames;
import applet.sidtune.SidTuneInfoCache;
import applet.ui.JNiceButton;

@SuppressWarnings("serial")
public class FavoritesPanel extends JPanel implements IFavorites, ActionListener {

	protected JTable fFavoritesTable;

	protected RowSorter<TableModel> fRowSorter;

	private final FavoritesModel fFavoriteModel;

	private final FavoritesView fFavoritesView;

	protected JSIDPlay2 fSidplay2;

	protected JPopupMenu fPopup;

	protected int fContextColumn;

	protected JPopupMenu fContextPopup;

	protected File fLastDir;

	protected JMenuItem fRemoveMenuItem;

	private String fFilename;

	private Random fRandom;

	protected IFavoritesSelectionListener fListener;

	private final JButton fMoveUp;

	private final JButton fMoveDown;

	private final JButton fUnsort;

	public FavoritesPanel(final FavoritesView favoritesView, JSIDPlay2 sidplay2) {
		fFavoritesView = favoritesView;
		fSidplay2 = sidplay2;

		setLayout(new BorderLayout());

		fFavoriteModel = new FavoritesModel(sidplay2);
		fFavoritesTable = new FavoritesTable(fFavoriteModel, fSidplay2);

		fPopup = new JPopupMenu("Customize Column");
		fRemoveMenuItem = new JMenuItem("Remove Column");
		fRemoveMenuItem.addActionListener(FavoritesPanel.this);
		fPopup.add(fRemoveMenuItem);
		for (String element : SidTuneInfoCache.SIDTUNE_INFOS) {
			JMenuItem menuItem = new JMenuItem("Add Column '" + element + "'");
			menuItem.addActionListener(FavoritesPanel.this);
			fPopup.add(menuItem);
		}
		JTableHeader header = fFavoritesTable.getTableHeader();
		header.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(MouseEvent e) {
				if (e.getButton() == MouseEvent.BUTTON3) {
					JTableHeader popupHeader = (JTableHeader) e.getSource();
					fContextColumn = popupHeader.columnAtPoint(new Point(e.getPoint()));
					int columnModelIndex = fFavoritesTable.convertColumnIndexToModel(fContextColumn);
					if (columnModelIndex == 0) {
						fRemoveMenuItem.setEnabled(false);
					} else {
						fRemoveMenuItem.setEnabled(true);
					}
					fPopup.show((Component) e.getSource(), e.getX(), e.getY());
				}
			}

		});
		fRowSorter = new TableRowSorter<TableModel>(fFavoriteModel);
		fRowSorter.addRowSorterListener(new RowSorterListener() {

			public void sorterChanged(RowSorterEvent e) {
				setMoveEnabledState(getSelection());
			}

		});
		fFavoritesTable.setRowSorter(fRowSorter);
		fFavoritesTable.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(KeyEvent e) {
				if (e.getModifiers() == 0 && KeyEvent.VK_BACK_SPACE == e.getKeyCode() || KeyEvent.VK_DELETE == e.getKeyCode()) {
					removeSelectedRows();
				} else if (e.getModifiers() == 0 && KeyEvent.VK_ENTER == e.getKeyCode()) {
					playSelectedRow();
				}

			}

		});
		fFavoritesTable.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(MouseEvent mouseEvent) {
				if (mouseEvent.getButton() == MouseEvent.BUTTON1 && mouseEvent.getClickCount() == 2) {
					playSelectedRow();
				} else if (mouseEvent.getButton() == MouseEvent.BUTTON3 && mouseEvent.getClickCount() == 1 && fFavoritesTable.getSelectionModel().getMinSelectionIndex() != -1) {

					fContextPopup = new JPopupMenu("Tune Actions");
					JMenuItem fileExportItem = new JMenuItem("Export to dir...");
					fileExportItem.addActionListener(FavoritesPanel.this);
					fContextPopup.add(fileExportItem);

					final JMenu moveToTab = new JMenu("Move to Tab");
					fSidplay2.support.firePropertyChange(IFavoriteTabNames.class.getName(), null, new IFavoriteTabNames() {

						public void setFavoriteTabNames(String[] names, String selected) {
							for (String name1 : names) {
								if (!name1.equals(selected)) {
									final String title = name1;
									JMenuItem tabItem = new JMenuItem(title);
									tabItem.addActionListener(new ActionListener() {

										public void actionPerformed(ActionEvent e) {
											moveSelectedFavoritesToTab(title, false);
										}

									});
									moveToTab.add(tabItem);
								}
							}
						}
					});
					fContextPopup.add(moveToTab);
					if (moveToTab.getMenuComponentCount() == 0) {
						moveToTab.setEnabled(false);
					}
					final JMenu copyToTab = new JMenu("Copy to Tab");
					fSidplay2.support.firePropertyChange(IFavoriteTabNames.class.getName(), null, new IFavoriteTabNames() {

						public void setFavoriteTabNames(String[] names, String selected) {
							for (String name1 : names) {
								if (!name1.equals(selected)) {
									final String title = name1;
									JMenuItem tabItem = new JMenuItem(title);
									tabItem.addActionListener(new ActionListener() {

										public void actionPerformed(ActionEvent e) {
											moveSelectedFavoritesToTab(title, true);
										}

									});
									copyToTab.add(tabItem);
								}
							}
						}
					});
					fContextPopup.add(copyToTab);
					if (copyToTab.getMenuComponentCount() == 0) {
						copyToTab.setEnabled(false);
					}

					JMenu convertItem = new JMenu("Convert to...");

					JMenuItem psid64 = new JMenuItem("PSID64 (.prg)");
					psid64.addActionListener(new ActionListener() {

						public void actionPerformed(ActionEvent e) {
							JFileChooser fc = new JFileChooser(fLastDir);
							fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
							int result = fc.showOpenDialog(fSidplay2);
							if (result == JFileChooser.APPROVE_OPTION && fc.getSelectedFile() != null) {
								fLastDir = fc.getSelectedFile();
								convertSelectedTunes();
							}
						}

					});
					convertItem.add(psid64);

					fContextPopup.add(convertItem);

					JMenuItem makeRelative = new JMenuItem("Make Path Relative");
					makeRelative.addActionListener(FavoritesPanel.this);
					fContextPopup.add(makeRelative);

					fContextPopup.show((Component) mouseEvent.getSource(), mouseEvent.getX(), mouseEvent.getY());
				}
			}

		});

		fFavoritesTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {

			public void valueChanged(ListSelectionEvent e) {
				if (fListener == null) {
					return;
				}
				String[] selection = getSelection();
				fListener.selectionChanged(selection);
				setMoveEnabledState(selection);
			}

		});
		new FileDrop(this, new FileDrop.Listener() {
			public void filesDropped(java.io.File[] files, Object source, Point point) {
				DropTarget t = (DropTarget) source;
				if (t.getComponent() == FavoritesPanel.this) {
					addToFavorites(files);
				}
			}
		});

		JPanel infos = new JPanel();
		infos.setLayout(new GridLayout(3, 1));
		JLabel label = new JLabel("Open context menu on column header to add/remove columns with additional tune details!");
		label.setFont(label.getFont().deriveFont(Font.ITALIC));
		label.setLayout(new FlowLayout(FlowLayout.LEFT));
		infos.add(label);

		JLabel label2 = new JLabel("Open context menu on selected tunes to start various tune actions!");
		label2.setFont(label2.getFont().deriveFont(Font.ITALIC));
		label2.setLayout(new FlowLayout(FlowLayout.LEFT));
		infos.add(label2);

		JPanel filterLine = new JPanel();
		filterLine.setLayout(new BorderLayout());

		JLabel filterLbl = new JLabel("Reg. Expr. filter:");
		filterLbl.setFont(filterLbl.getFont().deriveFont(Font.ITALIC));
		filterLbl.setLayout(new FlowLayout(FlowLayout.LEFT));
		filterLine.add(filterLbl, BorderLayout.WEST);

		final JTextField filterField = new JTextField();
		filterField.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(KeyEvent e) {
				if (filterField.getText().trim().length() == 0) {
					((TableRowSorter<TableModel>) fRowSorter).setRowFilter(null);
				} else {
					if (validatePattern()) {
						RowFilter<TableModel, Integer> filter = RowFilter.regexFilter(filterField.getText());
						((TableRowSorter<TableModel>) fRowSorter).setRowFilter(filter);
					}
				}
			}

			private boolean validatePattern() {
				boolean ok = true;
				filterField.setToolTipText(null);
				filterField.setBackground(Color.white);
				try {
					Pattern.compile(filterField.getText());
				} catch (PatternSyntaxException e) {
					filterField.setToolTipText(e.getMessage());
					filterField.setBackground(Color.red);
					ok = false;
				}
				return ok;
			}

		});
		filterLine.add(filterField, BorderLayout.CENTER);

		infos.add(filterLine);

		add(infos, BorderLayout.NORTH);
		add(new JScrollPane(fFavoritesTable), BorderLayout.CENTER);

		JPanel movePanel = new JPanel();
		movePanel.setLayout(new BoxLayout(movePanel, BoxLayout.Y_AXIS));

		fUnsort = new JNiceButton();
		fUnsort.setActionCommand("Unsort");
		fUnsort.setIcon(new ImageIcon(JSIDPlay2.class.getResource("icons/fav_unsort.png")));
		fUnsort.addActionListener(this);
		fUnsort.setPreferredSize(new Dimension(20, 20));
		fUnsort.setUI(new BasicButtonUI());

		fUnsort.setEnabled(false);
		fUnsort.setToolTipText("Unsort Favorites");
		movePanel.add(fUnsort);

		fMoveUp = new JNiceButton();
		fMoveUp.setActionCommand("Up");
		fMoveUp.setIcon(new ImageIcon(JSIDPlay2.class.getResource("icons/fav_up.png")));
		fMoveUp.addActionListener(this);
		fMoveUp.setPreferredSize(new Dimension(20, 20));
		fMoveUp.setEnabled(false);
		fMoveUp.setToolTipText("Move Up Favorite");
		fMoveUp.setUI(new BasicButtonUI());
		movePanel.add(fMoveUp);

		fMoveDown = new JNiceButton();
		fMoveDown.setActionCommand("Down");
		fMoveDown.setIcon(new ImageIcon(JSIDPlay2.class.getResource("icons/fav_down.png")));
		fMoveDown.addActionListener(this);
		fMoveDown.setPreferredSize(new Dimension(20, 20));
		fMoveDown.setEnabled(false);
		fMoveDown.setToolTipText("Move Down Favorite");
		movePanel.add(fMoveDown);

		add(movePanel, BorderLayout.EAST);
	}

	public void addToFavorites(File[] files) {
		for (int i = 0; files != null && i < files.length; i++) {
			final File file = files[i];
			if (file.isDirectory()) {
				addToFavorites(file.listFiles());
			} else {
				if (fFavoritesView.getTuneFilter().accept(file)) {
					if (!fFavoriteModel.contains(file, 0)) {
						fFavoriteModel.addRow(new Object[] { createRelativePath(file.getAbsolutePath()) });
					}
				}
			}
		}
	}

	protected void playSelectedRow() {
		int selectedRow = fFavoritesTable.getSelectedRow();
		if (selectedRow == -1) {
			return;
		}
		int selectedModelRow = fRowSorter.convertRowIndexToModel(selectedRow);
		if (selectedModelRow != -1) {
			// play tune
			fSidplay2.playFile(fFavoriteModel.getFile(fFavoriteModel.getValueAt(selectedModelRow, 0)), fFavoritesView);
			fFavoritesView.setCurrentlyPlayedFavorites(this);
			fFavoritesTable.repaint();
		}
	}

	public void removeSelectedRows() {
		int[] rows = fFavoritesTable.getSelectedRows();
		if (rows.length == 0) {
			return;
		}
		int response = JOptionPane.showConfirmDialog(this, String.format("Remove %d of my favorites?", rows.length), "Remove Favorites", JOptionPane.YES_NO_OPTION);
		if (response == JOptionPane.YES_OPTION) {
			for (int i = 0; i < rows.length; i++) {
				int row = fRowSorter.convertRowIndexToModel(rows[i]);
				fFavoriteModel.removeRow(row);
				// shift row numbers
				for (int j = i + 1; j < rows.length; j++) {
					rows[j] = rows[j] - 1;
				}
			}
		}
	}

	@SuppressWarnings("unchecked")
	public void loadFavorites(String filename) {
		try {
			if (filename.endsWith(".xml")) {
				// old version of favorites file
				ObjectInputStream fileIn = new ObjectInputStream(new FileInputStream(filename));
				ArrayList<IFavorite> favorites = (ArrayList<IFavorite>) fileIn.readObject();
				fileIn.close();
				fFavoriteModel.setRowCount(0);
				for (IFavorite favorite : favorites) {
					fFavoriteModel.addRow(new Object[] { favorite.getFilename() });
				}
				// automatic migration to new format
				filename = filename.substring(0, filename.lastIndexOf('.')) + FavoritesView.EXT_FAVORITES;
				saveFavorites(filename);
			} else {
				// new favorites file format
				final BufferedReader r = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "ISO-8859-1"));
				String line;
				fFavoriteModel.setRowCount(0);
				while ((line = r.readLine()) != null) {
					fFavoriteModel.addRow(new Object[] { line });
				}
				r.close();
			}
			this.fFilename = filename;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
	}

	public void saveFavorites(String filename) {
		try {
			final PrintStream p = new PrintStream(filename, "ISO-8859-1");
			for (int i = 0; i < fFavoriteModel.getRowCount(); i++) {
				p.println(fFavoriteModel.getValueAt(i, 0));
			}
			p.close();
			this.fFilename = filename;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void deselectFavorites() {
		fFavoritesTable.getSelectionModel().clearSelection();
		String[] selection = new String[0];
		fListener.selectionChanged(selection);
		setMoveEnabledState(selection);
	}

	protected void setMoveEnabledState(String[] selection) {
		List<? extends SortKey> keys = fRowSorter.getSortKeys();
		int keySize = keys.size();
		if (keySize == 0 || keys.get(0).getSortOrder() == SortOrder.UNSORTED) {
			fMoveUp.setEnabled(selection.length == 1);
			fMoveDown.setEnabled(selection.length == 1);
			fUnsort.setEnabled(false);
		} else {
			fMoveUp.setEnabled(false);
			fMoveDown.setEnabled(false);
			fUnsort.setEnabled(true);
		}
	}

	public void selectFavorites() {
		if (fFavoriteModel.getRowCount() == 0) {
			return;
		}
		fFavoritesTable.getSelectionModel().setSelectionInterval(0, fFavoriteModel.getRowCount() - 1);
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().startsWith("Add Column '")) {
			String property = e.getActionCommand().substring("Add Column '".length());
			property = property.substring(0, property.indexOf('\''));
			try {
				fFavoritesTable.getColumn(property);
				// if column already exist, do nothing
			} catch (IllegalArgumentException e1) {
				fFavoriteModel.addColumn(property);
			}
		} else if (e.getActionCommand().startsWith("Down")) {
			int row = fFavoritesTable.getSelectedRow();
			int start = row;
			int to = row < fFavoriteModel.getRowCount() - 1 ? row + 1 : row;
			fFavoriteModel.moveRow(start, start, to);
			fFavoritesTable.getSelectionModel().setSelectionInterval(to, to);
		} else if (e.getActionCommand().startsWith("Up")) {
			int row = fFavoritesTable.getSelectedRow();
			int start = row;
			int to = row > 0 ? row - 1 : row;
			fFavoriteModel.moveRow(start, start, to);
			fFavoritesTable.getSelectionModel().setSelectionInterval(to, to);
		} else if (e.getActionCommand().startsWith("Unsort")) {
			fRowSorter.setSortKeys(new ArrayList<SortKey>());
		} else if (e.getActionCommand().startsWith("Remove")) {
			removeColumn();
		} else if (e.getActionCommand().startsWith("Export")) {
			JFileChooser fc = new JFileChooser(fLastDir);
			fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			int result = fc.showOpenDialog(fSidplay2);
			if (result == JFileChooser.APPROVE_OPTION && fc.getSelectedFile() != null) {
				fLastDir = fc.getSelectedFile();
				int[] rows = fFavoritesTable.getSelectedRows();
				for (int row1 : rows) {
					int row = fRowSorter.convertRowIndexToModel(row1);
					File file = fFavoriteModel.getFile(fFavoriteModel.getValueAt(row, 0));
					try {
						copy(file.getAbsolutePath(), fc.getSelectedFile().getAbsolutePath());
					} catch (IOException e1) {
						System.err.println(e1.getMessage());
					}
				}
			}
		} else if (e.getActionCommand().startsWith("Make")) {
			int[] rows = fFavoritesTable.getSelectedRows();
			for (int row1 : rows) {
				int row = fRowSorter.convertRowIndexToModel(row1);
				fFavoriteModel.setValueAt(createRelativePath(fFavoriteModel.getValueAt(row, 0).toString()), row, 0);
			}
		}
	}

	private String createRelativePath(String nameToConvert) {
		boolean converted = false;
		String result = nameToConvert;
		String hvscName = fSidplay2.getHVSCName(nameToConvert);
		if (hvscName != null) {
			result = FavoritesModel.HVSC_PREFIX + hvscName;
			converted = true;
		}
		String cgscName = fSidplay2.getCGSCName(nameToConvert);
		if (!converted && cgscName != null) {
			result = FavoritesModel.CGSC_PREFIX + cgscName;
			converted = true;
		}
		if (!converted && new File(nameToConvert).isAbsolute()) {
			String relativePath = RelativePath.getRelativePath(new File(System.getProperty("user.dir")), new File(nameToConvert));
			result = relativePath != null ? relativePath : result;
		}
		return result;
	}

	public void copy(String fromFileName, String toFileName) throws IOException {
		File fromFile = new File(fromFileName);
		File toFile = new File(toFileName);

		if (!fromFile.exists()) {
			throw new IOException("FileCopy: " + "no such source file: " + fromFileName);
		}
		if (!fromFile.isFile()) {
			throw new IOException("FileCopy: " + "can't copy directory: " + fromFileName);
		}
		if (!fromFile.canRead()) {
			throw new IOException("FileCopy: " + "source file is unreadable: " + fromFileName);
		}

		if (toFile.isDirectory()) {
			toFile = new File(toFile, fromFile.getName());
		}

		if (toFile.exists()) {
			if (!toFile.canWrite()) {
				throw new IOException("FileCopy: " + "destination file is unwriteable: " + toFileName);
			}
			int response = JOptionPane.showConfirmDialog(fSidplay2, String.format("Overwrite existing file %s ?: ", toFile.getName()), "Overwrite File", JOptionPane.YES_NO_OPTION);
			if (response != JOptionPane.YES_OPTION) {
				throw new IOException(String.format("FileCopy: " + "existing file %s was not overwritten.", toFile.getName()));
			}
		} else {
			String parent = toFile.getParent();
			if (parent == null) {
				parent = System.getProperty("user.dir");
			}
			File dir = new File(parent);
			if (!dir.exists()) {
				throw new IOException("FileCopy: " + "destination directory doesn't exist: " + parent);
			}
			if (dir.isFile()) {
				throw new IOException("FileCopy: " + "destination is not a directory: " + parent);
			}
			if (!dir.canWrite()) {
				throw new IOException("FileCopy: " + "destination directory is unwriteable: " + parent);
			}
		}

		FileInputStream from = null;
		FileOutputStream to = null;
		try {
			from = new FileInputStream(fromFile);
			to = new FileOutputStream(toFile);
			byte[] buffer = new byte[4096];
			int bytesRead;

			while ((bytesRead = from.read(buffer)) != -1) {
				to.write(buffer, 0, bytesRead); // write
			}
		} finally {
			if (from != null) {
				try {
					from.close();
				} catch (IOException e) {
					/* XXX deal with this exception */
				}
			}
			if (to != null) {
				try {
					to.close();
				} catch (IOException e) {
					/* XXX deal with this exception */
				}
			}
		}
	}

	private void removeColumn() {
		int columnModelIndex = fFavoritesTable.convertColumnIndexToModel(fContextColumn);
		if (columnModelIndex == 0) {
			// do not remove filenames
			return;
		}
		fFavoriteModel.removeColumn(columnModelIndex);
		fFavoriteModel.fireTableStructureChanged();
	}

	protected void moveSelectedFavoritesToTab(final String title, final boolean copy) {
		JTabbedPane pane = fFavoritesView.getTabbedPane();
		int index = pane.indexOfTab(title);
		// target panel
		IFavorites panel = (IFavorites) pane.getComponentAt(index);

		int[] rows = fFavoritesTable.getSelectedRows();
		for (int i = 0; i < rows.length; i++) {
			int row = fRowSorter.convertRowIndexToModel(rows[i]);
			// add next row to target tab
			panel.addToFavorites(new File[] { fFavoriteModel.getFile(fFavoriteModel.getValueAt(row, 0)) });
			if (!copy) {
				// remove row from source tab
				fFavoriteModel.removeRow(row);
				// shift row numbers
				for (int j = i + 1; j < rows.length; j++) {
					rows[j] = rows[j] - 1;
				}
			}
		}
	}

	protected void convertSelectedTunes() {
		final ArrayList<File> files = new ArrayList<File>();
		int[] rows = fFavoritesTable.getSelectedRows();
		for (int row1 : rows) {
			int row = fRowSorter.convertRowIndexToModel(row1);
			files.add(fFavoriteModel.getFile(fFavoriteModel.getValueAt(row, 0)));
		}

		SidTuneConverter c = new SidTuneConverter(fSidplay2);
		c.convertFiles(files.toArray(new File[0]), fLastDir);
	}

	public String getFileName() {
		return fFilename;
	}

	public String getNextFilename(String filename) {
		int playedRow = -1;
		int rowCount = fFavoritesTable.getRowCount();
		for (int i = 0; i < rowCount; i++) {
			int row = fRowSorter.convertRowIndexToModel(i);
			try {
				String currFilename = fFavoriteModel.getFile(fFavoriteModel.getValueAt(row, 0)).getCanonicalPath();
				if (currFilename.equals(new File(filename).getCanonicalPath())) {
					playedRow = i;
					System.out.println("played row = " + i);
					break;
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		final int nextRow = playedRow + 1;
		if (nextRow == fFavoritesTable.getRowCount()) {
			System.err.println("next is null!");
			return null;
		}
		int row = fRowSorter.convertRowIndexToModel(nextRow);
		SwingUtilities.invokeLater(new Runnable() {

			public void run() {
				fFavoritesTable.scrollRectToVisible(fFavoritesTable.getCellRect(nextRow, 0, true));
			}

		});
		return fFavoriteModel.getFile(fFavoriteModel.getValueAt(row, 0)).getAbsolutePath();
	}

	public String getNextRandomFilename(String filename) {
		int rowCount = fFavoritesTable.getRowCount();
		fRandom = new Random();
		final int randomRow = Math.abs(fRandom.nextInt()) % rowCount;
		int row = fRowSorter.convertRowIndexToModel(randomRow);
		SwingUtilities.invokeLater(new Runnable() {

			public void run() {
				fFavoritesTable.scrollRectToVisible(fFavoritesTable.getCellRect(randomRow, 0, true));
			}

		});
		return fFavoriteModel.getFile(fFavoriteModel.getValueAt(row, 0)).getAbsolutePath();
	}

	public void addSelectionListener(IFavoritesSelectionListener listener) {
		fListener = listener;
	}

	public String[] getSelection() {
		int[] rows = fFavoritesTable.getSelectedRows();
		if (rows.length == 0) {
			return new String[0];
		}
		ArrayList<String> filenames = new ArrayList<String>();
		for (int row1 : rows) {
			int row = fRowSorter.convertRowIndexToModel(row1);
			filenames.add(fFavoriteModel.getFile(fFavoriteModel.getValueAt(row, 0)).getAbsolutePath());
		}
		String[] retValue = filenames.toArray(new String[filenames.size()]);
		return retValue;
	}

	public boolean isEmpty() {
		return fFavoriteModel.getRowCount() == 0;
	}

}
