package applet.favorites;

import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.Random;

import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileFilter;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;
import sidplay.IniConfig.favorites_section;
import applet.C64FileFilter;
import applet.JSIDPlay2;
import applet.TuneTab;
import applet.events.ITuneStateChanged;
import applet.events.favorites.IAddFavoritesTab;
import applet.events.favorites.IChangeFavoritesTab;
import applet.events.favorites.IFavoriteTabNames;
import applet.events.favorites.IGetFavorites;
import applet.events.favorites.IRemoveFavoritesTab;

@SuppressWarnings("serial")
public class FavoritesView extends TuneTab implements ActionListener,
		IFavoritesSelectionListener {

	static final String EXT_FAVORITES = ".js2";

	/**
	 * file filter for tunes
	 */
	private final FileFilter fPlayListFilter = new FavoritesFileFilter(EXT_FAVORITES);

	/**
	 * file filter for tunes
	 */
	private final FileFilter fTuneFilter = new C64FileFilter();

	protected JTabbedPane fTabbedPane;

	protected JSIDPlay2 fSidplay2;

	private File fLastDir;

	protected JCheckBox fPlaybackEnable;

	protected JRadioButton fNormalPlaybackButton;

	protected JRadioButton fRandomOneButton;

	protected JRadioButton fRandomAllButton;

	protected JCheckBox fRepeatEnable;

	protected JRadioButton fRepeatOneButton;

	private FavoritesPanel fCurrentlyPlayedFavorites;

	private final Random random = new Random();

	protected JButton fRemove;

	public FavoritesView(JSIDPlay2 sidplay) {
		sidplay.support.addPropertyChangeListener(this);
		this.fSidplay2 = sidplay;
	}

	@Override
	public void createContents() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		JLabel label = new JLabel("Use Drag & Drop to add your favorite tunes!");
		label.setFont(label.getFont().deriveFont(Font.ITALIC));
		label.setLayout(new FlowLayout(FlowLayout.LEFT));
		add(label);

		JPanel favoritesGroup = new JPanel();
		favoritesGroup.setBorder(new TitledBorder("Playlist"));
		favoritesGroup
				.setLayout(new BoxLayout(favoritesGroup, BoxLayout.Y_AXIS));
		add(favoritesGroup);

		JPanel buttonsPanel = new JPanel();
		buttonsPanel.setLayout(new BoxLayout(buttonsPanel, BoxLayout.X_AXIS));
		favoritesGroup.add(buttonsPanel);

		JButton importFilesFolders = new JButton("Add...");
		importFilesFolders.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_add.png")));
		importFilesFolders.addActionListener(this);
		buttonsPanel.add(importFilesFolders);

		fRemove = new JButton("Remove");
		fRemove.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_remove.png")));
		fRemove.addActionListener(this);
		fRemove.setEnabled(false);
		buttonsPanel.add(fRemove);

		JButton selectAll = new JButton("Select All");
		selectAll.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_selectall.png")));
		selectAll.addActionListener(this);
		buttonsPanel.add(selectAll);

		JButton deselectAll = new JButton("Deselect All");
		deselectAll.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_deselectall.png")));
		deselectAll.addActionListener(this);
		buttonsPanel.add(deselectAll);

		JPanel buttonsPanel2 = new JPanel();
		buttonsPanel2.setLayout(new BoxLayout(buttonsPanel2, BoxLayout.X_AXIS));
		favoritesGroup.add(buttonsPanel2);

		JButton load = new JButton("Load...");
		load.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_load.png")));
		load.addActionListener(this);
		buttonsPanel2.add(load);

		JButton save = new JButton("Save");
		save.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_save.png")));
		save.addActionListener(this);
		buttonsPanel2.add(save);

		JButton saveAs = new JButton("Save...");
		saveAs.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_save.png")));
		saveAs.addActionListener(this);
		buttonsPanel2.add(saveAs);

		fTabbedPane = new JTabbedPane();

		fTabbedPane.add("", new JLabel(
				"Please click on the plus icon to add a new Playlist."));
		fTabbedPane.setTabComponentAt(0, new ButtonTabComponent(fSidplay2,
				fTabbedPane, "", "icons/_active__add.png", "Add a new tab",
				new ActionListener() {

					public void actionPerformed(ActionEvent e) {
						fSidplay2.support.firePropertyChange(
								IAddFavoritesTab.class.getName(), null,
								new IAddFavoritesTab() {

									public String getTitle() {
										return "New Tab";
									}

									public void setFavorites(
											IFavorites favorites) {

									}

								});
					}

				}));

		fTabbedPane.setTabLayoutPolicy(JTabbedPane.WRAP_TAB_LAYOUT);

		restoreFavorites();
		reloadRestoredFavorites();

		fTabbedPane.addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				final int index = fTabbedPane.getSelectedIndex();
				if (index == -1 || index == fTabbedPane.getTabCount() - 1) {
					return;
				}
				{
					Component comp = fTabbedPane.getComponentAt(index);
					if (comp instanceof IFavorites) {
						IFavorites fav = (IFavorites) comp;
						fRemove.setEnabled(fav.getSelection().length > 0);
					}
				}
				fSidplay2.support.firePropertyChange(IChangeFavoritesTab.class
						.getName(), null, new IChangeFavoritesTab() {
					public int getIndex() {
						return index;
					}

					public String getTitle() {
						return fTabbedPane.getTitleAt(index);
					}

					public String getFileName() {
						Component comp = fTabbedPane.getComponentAt(index);
						if (comp instanceof IFavorites) {
							IFavorites fav = (IFavorites) comp;
							return fav.getFileName();
						}
						return null;
					}

					public boolean isSelected() {
						return true;
					}
				});
			}

		});
		favoritesGroup.add(fTabbedPane);

		JPanel playbackOptionsGroup = new JPanel();
		playbackOptionsGroup.setBorder(new TitledBorder("Playback Options"));
		playbackOptionsGroup.setLayout(new GridBagLayout());
		add(playbackOptionsGroup/* , BorderLayout.SOUTH */);

		JLabel note = new JLabel(
				"Set the \"Default Song Length\" and check \"Respect Song Length\" option on the \"General\" tab!");
		Font font = note.getFont().deriveFont(Font.ITALIC);
		note.setFont(font);
		note.setForeground(Color.BLUE);
		GridBagConstraints c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 4;
		c.gridheight = 1;
		c.fill = GridBagConstraints.BOTH;
		playbackOptionsGroup.add(note, c);

		fPlaybackEnable = new JCheckBox("Playback On/Off");
		fPlaybackEnable.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent arg0) {
				boolean playbackSelected = fPlaybackEnable.isSelected();
				fNormalPlaybackButton.setEnabled(playbackSelected);
				fRandomOneButton.setEnabled(playbackSelected);
				fRandomAllButton.setEnabled(playbackSelected);
				fRepeatEnable.setEnabled(playbackSelected);
				fRepeatOneButton.setEnabled(playbackSelected);
				boolean repeatSelected = fRepeatEnable.isSelected();
				fRepeatOneButton.setEnabled(repeatSelected);
			}
		});
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		playbackOptionsGroup.add(fPlaybackEnable, c);

		fNormalPlaybackButton = new JRadioButton("Normal");
		fNormalPlaybackButton.setSelected(true);
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		playbackOptionsGroup.add(fNormalPlaybackButton, c);

		fRandomOneButton = new JRadioButton("Random (one playlist)");
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		playbackOptionsGroup.add(fRandomOneButton, c);

		fRandomAllButton = new JRadioButton("Random (all playlists)");
		c = new GridBagConstraints();
		c.gridx = 3;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		playbackOptionsGroup.add(fRandomAllButton, c);

		ButtonGroup playbackGroup = new ButtonGroup();
		playbackGroup.add(fNormalPlaybackButton);
		playbackGroup.add(fRandomOneButton);
		playbackGroup.add(fRandomAllButton);

		fNormalPlaybackButton.setEnabled(false);
		fRandomOneButton.setEnabled(false);
		fRandomAllButton.setEnabled(false);

		fRepeatEnable = new JCheckBox("Repeat On/Off");
		fRepeatEnable.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent arg0) {
				boolean repeatSelected = fRepeatEnable.isSelected();
				fRepeatOneButton.setEnabled(repeatSelected);
				fRepeatOneButton.setSelected(repeatSelected);
			}
		});
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		playbackOptionsGroup.add(fRepeatEnable, c);

		fRepeatOneButton = new JRadioButton("Repeat (one tune)");
		fRepeatOneButton.setSelected(true);
		fRepeatOneButton.setEnabled(false);
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		playbackOptionsGroup.add(fRepeatOneButton, c);

		ButtonGroup repeatGroup = new ButtonGroup();
		repeatGroup.add(fRepeatOneButton);

		fRepeatEnable.setEnabled(false);
	}

	private void reloadRestoredFavorites() {
		favorites_section favorites = fSidplay2.getFavorites();
		for (int i = 0; i < favorites.getFavorites_filenames().length; i++) {
			final String filename = favorites.getFavorites_filenames()[i];
			if (filename == null) {
				continue;
			}
			final int index = i;
			SwingUtilities.invokeLater(new Runnable() {

				public void run() {
					Component comp = fTabbedPane.getComponentAt(index);
					if (comp instanceof IFavorites) {
						IFavorites fav = (IFavorites) comp;
						fav.loadFavorites(filename);

						// BEGIN change event for migrated files
						String baseName = new File(filename).getName();
						int lastIndexOf = baseName.lastIndexOf('.');
						final String title;
						if (lastIndexOf != -1) {
							title = baseName.substring(0, lastIndexOf);
						} else {
							title = baseName;
						}
						fSidplay2.support.firePropertyChange(IChangeFavoritesTab.class
								.getName(), null, new IChangeFavoritesTab() {
							public int getIndex() {
								return index;
							}

							public String getTitle() {
								return title;
							}

							public String getFileName() {
								return filename.substring(0, filename.lastIndexOf('.'))
								+ EXT_FAVORITES;
							}

							public boolean isSelected() {
								return false;
							}
						});
						// END change event for migrated files
					}
				}

			});
		}
	}

	private void restoreFavorites() {
		favorites_section favorites = fSidplay2.getFavorites();
		for (int i = 0; i < favorites.getFavorites_titles().length; i++) {
			final String title = favorites.getFavorites_titles()[i];
			addTab(title);
		}
		if (favorites.getFavorites_titles().length == 0) {
			// add a first tab
			fSidplay2.support.firePropertyChange(IAddFavoritesTab.class
					.getName(), null, new IAddFavoritesTab() {

				public String getTitle() {
					return "New Tab";
				}

				public void setFavorites(IFavorites newFavorites) {
				}

			});
		} else {
			String title = favorites.getFavorites_current();
			int index = fTabbedPane.indexOfTab(title);
			fTabbedPane.setSelectedIndex(Math.max(index, 0));
		}
	}

	private IFavorites addTab(String newTitle) {
		final int lastIndex = fTabbedPane.getTabCount() - 1;
		final FavoritesPanel favorites = new FavoritesPanel(this, fSidplay2);
		favorites.addSelectionListener(this);
		fTabbedPane.insertTab(newTitle, null, favorites, null, lastIndex);
		fTabbedPane.setTabComponentAt(lastIndex, new ButtonTabComponent(
				fSidplay2, fTabbedPane, newTitle, "icons/_active__no.png",
				"close this tab", new ActionListener() {

					public void actionPerformed(final ActionEvent e) {
						ButtonTabComponent.TabButton button = (ButtonTabComponent.TabButton) e
								.getSource();
						final int index = fTabbedPane
								.indexOfTabComponent(button.getComp());
						final String title = fTabbedPane.getTitleAt(index);
						if (fTabbedPane.getTabCount() > 2) {
							if (index < fTabbedPane.getComponentCount() - 1) {
								int response = JOptionPane.NO_OPTION;
								if (!favorites.isEmpty()) {
									response = JOptionPane
											.showConfirmDialog(
													FavoritesView.this,
													String
															.format(
																	"Remove all \"%s\" favorites? Your favorites file on disk will NOT be removed.",
																	title),
													"Remove Favorites",
													JOptionPane.YES_NO_OPTION);
								} else {
									response = JOptionPane.YES_OPTION;
								}

								if (response == JOptionPane.YES_OPTION) {
									// remove tab
									fSidplay2.support
											.firePropertyChange(
													IRemoveFavoritesTab.class
															.getName(), null,
													new IRemoveFavoritesTab() {

														public int getIndex() {
															return index;
														}

														public String getTitle() {
															return title;
														}

													});
								}
							}
						}
					}

				}));

		fTabbedPane.setSelectedIndex(lastIndex);
		return favorites;
	}

	private void removeTab(int index) {
		fTabbedPane.removeTabAt(index);
		fTabbedPane.setSelectedIndex(index > 0 ? index - 1 : 0);
	}

	private void changeTab(IChangeFavoritesTab changeFavoritesTab) {
		fTabbedPane.setTitleAt(changeFavoritesTab.getIndex(),
				changeFavoritesTab.getTitle());
	}

	private void playNextTune(ITuneStateChanged stateChanged) {
		if (fPlaybackEnable == null) {
			// gui not created yet, ignore!
			return;
		}
		if (fPlaybackEnable.isSelected()) {
			String tune = stateChanged.getTune();
			String nextFilename = null;
			if (fCurrentlyPlayedFavorites != null
					&& fRepeatEnable.isSelected() && fRepeatOneButton.isSelected()) {
				// repeat one tune
				setPlayingIcon(fCurrentlyPlayedFavorites);
				nextFilename = tune;
			} else if (fRandomAllButton.isSelected()) {
				// random all playlists
				int index = Math.abs(random.nextInt())
						% (fTabbedPane.getTabCount() - 1);
				Component comp = fTabbedPane.getComponentAt(index);
				setPlayingIcon(comp);
				fTabbedPane.setSelectedComponent(comp);
				IFavorites fav = (IFavorites) comp;
				nextFilename = fav.getNextRandomFilename(tune);
			} else if (fCurrentlyPlayedFavorites != null
					&& fRandomOneButton.isSelected()) {
				// random one playlist
				IFavorites fav = fCurrentlyPlayedFavorites;
				setPlayingIcon(fCurrentlyPlayedFavorites);
				nextFilename = fav.getNextRandomFilename(tune);
			} else if (fCurrentlyPlayedFavorites != null
					&& !fRepeatEnable.isSelected()) {
				// normal playback
				IFavorites fav = fCurrentlyPlayedFavorites;
				setPlayingIcon(fCurrentlyPlayedFavorites);
				nextFilename = fav.getNextFilename(tune);
			}
			if (nextFilename != null) {
				// System.err.println("Play Next File: " + nextFilename);
				final File file = new File(nextFilename);
				SwingUtilities.invokeLater(new Runnable() {

					public void run() {
						// play tune
						fSidplay2.playFile(file, FavoritesView.this);
					}

				});
			}
		}
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().startsWith("Load")) {
			JFileChooser fileDialog = new JFileChooser(fLastDir);
			fileDialog.setFileFilter(fPlayListFilter);
			int rc = fileDialog.showOpenDialog(fSidplay2);
			if (rc == JFileChooser.APPROVE_OPTION
					&& fileDialog.getSelectedFile() != null) {
				fLastDir = fileDialog.getSelectedFile().getParentFile();
				final String name;
				if (!fileDialog.getSelectedFile().getAbsolutePath().endsWith(
						EXT_FAVORITES)
						&& !fileDialog.getSelectedFile().getAbsolutePath()
								.endsWith(".xml")) {
					name = fileDialog.getSelectedFile().getAbsolutePath()
							+ EXT_FAVORITES;
				} else {
					name = fileDialog.getSelectedFile().getAbsolutePath();
				}
				String baseName = new File(name).getName();
				int lastIndexOf = baseName.lastIndexOf('.');
				final String title;
				if (lastIndexOf != -1) {
					title = baseName.substring(0, lastIndexOf);
				} else {
					title = baseName;
				}

				// add a first tab
				fSidplay2.support.firePropertyChange(IAddFavoritesTab.class
						.getName(), null, new IAddFavoritesTab() {

					public String getTitle() {
						return title;
					}

					public void setFavorites(IFavorites favorites) {

					}

				});

				final int index = fTabbedPane.getSelectedIndex();
				Component comp = fTabbedPane.getComponentAt(index);
				if (comp instanceof IFavorites) {
					IFavorites fav = (IFavorites) comp;
					fav.loadFavorites(name);
				}
				fSidplay2.support.firePropertyChange(IChangeFavoritesTab.class
						.getName(), null, new IChangeFavoritesTab() {
					public int getIndex() {
						return index;
					}

					public String getTitle() {
						return title;
					}

					public String getFileName() {
						if (name.endsWith(".xml")) {
							return name.substring(0, name.lastIndexOf('.'))
									+ EXT_FAVORITES;
						} else {
							return name;
						}
					}

					public boolean isSelected() {
						return false;
					}
				});
			}
		} else if (e.getActionCommand().startsWith("Save...")) {
			saveAs();
		} else if (e.getActionCommand().startsWith("Save")) {
			final int index = fTabbedPane.getSelectedIndex();
			Component comp = fTabbedPane.getComponentAt(index);
			if (comp instanceof IFavorites) {
				IFavorites fav = (IFavorites) comp;
				if (fav.getFileName() == null) {
					saveAs();
				} else {
					fav.saveFavorites(fav.getFileName());
				}
			}
		} else if (e.getActionCommand().startsWith("Add")) {
			int index = fTabbedPane.getSelectedIndex();
			Component comp = fTabbedPane.getComponentAt(index);
			if (comp instanceof IFavorites) {
				IFavorites fav = (IFavorites) comp;
				JFileChooser fileDialog = new JFileChooser(fLastDir);
				fileDialog.setFileFilter(fTuneFilter);
				fileDialog.setMultiSelectionEnabled(true);
				fileDialog
						.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
				int rc = fileDialog.showOpenDialog(this);
				if (rc == JFileChooser.APPROVE_OPTION
						&& fileDialog.getSelectedFile() != null) {
					fLastDir = fileDialog.getSelectedFile().getParentFile();
					File files[] = fileDialog.getSelectedFiles();
					fav.addToFavorites(files);
				}
			}
		} else if (e.getActionCommand().startsWith("Select")) {
			int index = fTabbedPane.getSelectedIndex();
			Component comp = fTabbedPane.getComponentAt(index);
			if (comp instanceof IFavorites) {
				IFavorites fav = (IFavorites) comp;
				fav.selectFavorites();
			}
		} else if (e.getActionCommand().startsWith("Deselect")) {
			int index = fTabbedPane.getSelectedIndex();
			Component comp = fTabbedPane.getComponentAt(index);
			if (comp instanceof IFavorites) {
				IFavorites fav = (IFavorites) comp;
				fav.deselectFavorites();
			}
		} else if (e.getActionCommand().startsWith("Remove")) {
			int index = fTabbedPane.getSelectedIndex();
			Component comp = fTabbedPane.getComponentAt(index);
			if (comp instanceof IFavorites) {
				IFavorites fav = (IFavorites) comp;
				fav.removeSelectedRows();
			}
		}
	}

	private void saveAs() {
		JFileChooser fileDialog = new JFileChooser(fLastDir);
		fileDialog.setFileFilter(fPlayListFilter);
		int rc = fileDialog.showSaveDialog(fSidplay2);
		if (rc == JFileChooser.APPROVE_OPTION
				&& fileDialog.getSelectedFile() != null) {
			fLastDir = fileDialog.getSelectedFile().getParentFile();
			final String name;
			if (!fileDialog.getSelectedFile().getAbsolutePath().endsWith(
					EXT_FAVORITES)) {
				name = fileDialog.getSelectedFile().getAbsolutePath()
						+ EXT_FAVORITES;
			} else {
				name = fileDialog.getSelectedFile().getAbsolutePath();
			}
			final int index = fTabbedPane.getSelectedIndex();
			Component comp = fTabbedPane.getComponentAt(index);
			if (comp instanceof IFavorites) {
				IFavorites fav = (IFavorites) comp;
				fav.saveFavorites(name);
			}
			String baseName = new File(name).getName();
			int lastIndexOf = baseName.lastIndexOf('.');
			final String title;
			if (lastIndexOf != -1) {
				title = baseName.substring(0, lastIndexOf);
			} else {
				title = baseName;
			}
			fSidplay2.support.firePropertyChange(IChangeFavoritesTab.class
					.getName(), null, new IChangeFavoritesTab() {
				public int getIndex() {
					return index;
				}

				public String getTitle() {
					return title;
				}

				public String getFileName() {
					return name;
				}

				public boolean isSelected() {
					return false;
				}
			});
		}
	}

	public void propertyChange(PropertyChangeEvent event) {
		if (IAddFavoritesTab.class.getName().equals(event.getPropertyName())) {
			IAddFavoritesTab addFavoritesTab = (IAddFavoritesTab) event
					.getNewValue();
			IFavorites favorites = addTab(addFavoritesTab.getTitle());
			addFavoritesTab.setFavorites(favorites);
		} else if (IRemoveFavoritesTab.class.getName().equals(
				event.getPropertyName())) {
			IRemoveFavoritesTab removeFavoritesTab = (IRemoveFavoritesTab) event
					.getNewValue();
			removeTab(removeFavoritesTab.getIndex());
		} else if (IChangeFavoritesTab.class.getName().equals(
				event.getPropertyName())) {
			IChangeFavoritesTab changeFavoritesTab = (IChangeFavoritesTab) event
					.getNewValue();
			changeTab(changeFavoritesTab);
		} else if (IFavoriteTabNames.class.getName().equals(
				event.getPropertyName())) {
			IFavoriteTabNames favoritListNames = (IFavoriteTabNames) event
					.getNewValue();

			ArrayList<String> result = new ArrayList<String>();
			for (int i = 0; i < fTabbedPane.getTabCount() - 1; i++) {
				result.add(fTabbedPane.getTitleAt(i));
			}

			favoritListNames.setFavoriteTabNames(result
					.toArray(new String[result.size()]), fTabbedPane
					.getTitleAt(fTabbedPane.getSelectedIndex()));
		} else if (IGetFavorites.class.getName()
				.equals(event.getPropertyName())) {
			IGetFavorites favoritesByName = (IGetFavorites) event.getNewValue();
			int index = favoritesByName.getIndex();
			IFavorites favorites = (IFavorites) fTabbedPane
					.getComponentAt(index);
			favoritesByName.setFavorites(favorites);
		} else if (ITuneStateChanged.class.getName().equals(
				event.getPropertyName())) {
			ITuneStateChanged stateChanged = (ITuneStateChanged) event
					.getNewValue();
			setPlayingIcon(null);
			if (stateChanged.naturalFinished()) {
				playNextTune(stateChanged);
			}
		}

	}

	@Override
	public void setTune(Player m_engine, SidTune m_tune) {
	}

	public JTabbedPane getTabbedPane() {
		return fTabbedPane;
	}

	public void setCurrentlyPlayedFavorites(FavoritesPanel favoritesPanel) {
		fCurrentlyPlayedFavorites = favoritesPanel;
		setPlayingIcon(favoritesPanel);
	}

	private void setPlayingIcon(Component comp) {
		for (int i = 0; i < fTabbedPane.getTabCount(); i++) {
			ButtonTabComponent tabComponent = (ButtonTabComponent) fTabbedPane
					.getTabComponentAt(i);
			if (tabComponent == null)
				continue;
			if (fTabbedPane.getComponentAt(i) != comp) {
				tabComponent.fPlayButton.setVisible(false);
			} else {
				tabComponent.fPlayButton.setVisible(true);
			}
		}
	}

	public void selectionChanged(String[] selection) {
		fRemove.setEnabled(selection.length > 0);
	}

	public FileFilter getTuneFilter() {
		return fTuneFilter;
	}

}
