package applet.fileviewer;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.io.IOException;

import javax.swing.ImageIcon;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.tree.DefaultTreeCellRenderer;

import applet.JSIDPlay2;

@SuppressWarnings("serial")
final class TuneTreeRenderer extends DefaultTreeCellRenderer {

	private JSIDPlay2 fSidplay2;

	private final JPanel view;
	
	private TuneFile file;

	private ImageIcon fStilIcon;

	private ImageIcon fStilNoIcon;

	/**
	 * @param view
	 * @param sidplay
	 */
	public TuneTreeRenderer(final JPanel view, final JSIDPlay2 sidplay) {
		this.view = view;
		this.fSidplay2 = sidplay;
	}

	@Override
	public Component getTreeCellRendererComponent(JTree tree, Object value,
			boolean sel, boolean expanded, boolean leaf, int row,
			boolean _hasFocus) {
		if (value instanceof TuneFile) {
			file = (TuneFile) value;
			if (leaf && file != null) {
				try {
					if (fSidplay2.getSTIL(file.getCanonicalPath()) != null) {
						// set STIL icon
						if (fStilIcon == null) {
							fStilIcon = new ImageIcon(JSIDPlay2.class
									.getResource("icons/stil.png"));
						}
						setLeafIcon(fStilIcon);
					} else {
						// set NO STIL icon
						if (fStilNoIcon == null) {
							fStilNoIcon = new ImageIcon(JSIDPlay2.class
									.getResource("icons/stil_no.png"));
						}
						setLeafIcon(fStilNoIcon);
					}
				} catch (IOException e) {
					setLeafIcon(null);
				}
			}
		}

		Component treeCellRendererComponent = super
				.getTreeCellRendererComponent(tree, value, sel, expanded, leaf,
						row, _hasFocus);
		return treeCellRendererComponent;
	}

	@Override
	protected void paintComponent(Graphics g) {
		try {
			if (file!=null && isCurrentlyPlayed(file.getCanonicalPath())) {
				// highlight currently played tune
				g.setColor(Color.LIGHT_GRAY);
				int x = (getIcon() == null) ? 0 : getIcon().getIconWidth()
						+ getIconTextGap() - 1;
				g.fillRect(x, 0, this.view.getWidth(), this.view.getHeight());
			}
		} catch (IOException e) {
		}
		super.paintComponent(g);
	}

	/**
	 * Check if the current tree element is being played
	 * @param filename file name
	 * 
	 * @return is it played now?
	 */
	private boolean isCurrentlyPlayed(String filename) {
		if (fSidplay2.getPlayingFile() == null || file == null) {
			return false;
		}
		return fSidplay2.getPlayingFile().startsWith(
				filename + System.getProperty("file.separator")) ||
				fSidplay2.getPlayingFile().equals(filename);
	}
}