package applet.fileviewer.search;

import java.io.File;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.swing.JOptionPane;

public class SearchInIndexThread extends SearchThread {

	private final Connection fConnection;
	private final String fRoot;
	private ResultSet rs;

	private String fieldName, fieldValue;
	private boolean caseSensitive;

	public SearchInIndexThread(String root, Connection conn, boolean forward) {
		super(forward);
		fConnection = conn;
		fRoot = root;
	}

	private void select() throws SQLException {
		if (rs != null) {
			rs.close();
		}

		String fv = fieldValue;
		if (!caseSensitive) {
			fv = fv.toLowerCase();
		}
		String[] pieces = fv.split("\\s+");

		String query = "SELECT \"Full Path\" FROM ";
		query += fieldName.startsWith("STIL") && !"STIL glb. Comment".equals(fieldName) ? "stil" : "collection";
		query += " WHERE 1=1";

		/* split by spaces, find entries where all match */
		for (@SuppressWarnings("unused") String piece : pieces) {
			if (!caseSensitive) {
				query += " AND lower(\"" + fieldName + "\")";
			} else {
				query += " AND \"" + fieldName + "\"";
			}
			query += " LIKE ?";
		}
		query += " ORDER BY \"Full Path\" ASC";
		PreparedStatement stmt = fConnection.prepareStatement(query, ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);

		for (int i = 0; i < pieces.length; i++) {
			stmt.setString(i + 1, "%" + pieces[i] + "%");
		}

		rs = stmt.executeQuery();
	}

	@Override
	public void run() {
		for (ISearchListener listener : fListeners) {
			listener.searchStart();
		}
		try {
			if (rs == null) {
				select();
			}
			// search the collection
			while (!fAborted && (fForward ? next() : prev())) {
				String filePath = rs.getString(1);
				for (ISearchListener listener : fListeners) {
					listener.searchHit(new File(fRoot, filePath));
				}
			}
			if (!fAborted) {
				rs.close();
				rs = null;
			}
		} catch (SQLException e) {
			JOptionPane.showMessageDialog(null, e);
		}
		for (ISearchListener listener : fListeners) {
			listener.searchStop(fAborted);
		}
	}

	private boolean prev() throws SQLException {
		if (rs.isFirst()) {
			return false;// rs.last();
		} else {
			return rs.previous();
		}
	}

	private boolean next() throws SQLException {
		if (rs.isLast()) {
			return false;// rs.first();
		} else {
			return rs.next();
		}
	}

	@Override
	public Object getSearchState() {
		return rs;
	}

	@Override
	public void restoreSearchState(Object state) {
		if (state instanceof ResultSet) {
			rs = (ResultSet) state;
		}
	}

	public String getFieldName() {
		return fieldName;
	}

	public void setFieldName(String fieldName) {
		this.fieldName = fieldName;
	}

	public String getFieldValue() {
		return fieldValue;
	}

	public void setFieldValue(String fieldValue) {
		this.fieldValue = fieldValue;
	}

	public boolean isCaseSensitive() {
		return caseSensitive;
	}

	public void setCaseSensitive(boolean caseSensitive) {
		this.caseSensitive = caseSensitive;
	}
}
