package applet.general;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.Proxy;
import java.net.URL;
import java.net.URLConnection;

import applet.JSIDPlay2;

public class DownloadThread extends Thread {
	private static final int MAX_BUFFER_SIZE = 8192;

	private boolean isAborted;
	private final JSIDPlay2 fSidplay2;
	private final String fURL;
	private final GeneralCanvas fListener;

	public DownloadThread(JSIDPlay2 sidplay2, GeneralCanvas listener,
			String url) {
		fSidplay2 = sidplay2;
		fURL = url;
		fListener = listener;
	}

	@Override
	public void run() {
		try {
			Proxy proxy = fSidplay2.getProxy();

			URL url = new URL(fURL);
			URLConnection conn;
			if (proxy != null) {
				conn = url.openConnection(proxy);
			} else {
				conn = url.openConnection();
			}
			conn.connect();
			InputStream is = conn.getInputStream();
			File downloadedFile = new File(System.getProperty("user.home"), new File(url.getPath()).getName());
			downloadedFile.deleteOnExit();
			FileOutputStream out = new FileOutputStream(downloadedFile);

			int total = 0;
			int bytesRead;
			byte[] buffer = new byte[MAX_BUFFER_SIZE];
			while ((bytesRead = is.read(buffer)) != -1) {
				out.write(buffer, 0, bytesRead);
				total += bytesRead;
				if (isAborted) {
					fListener.downloadStop(null);
					out.close();
					is.close();
					return;
				}
				fListener.downloadStep(total * 100 / conn.getContentLength());
			}
			out.close();
			is.close();
			fListener.downloadStop(downloadedFile);
		} catch (MalformedURLException e1) {
			e1.printStackTrace();
			fListener.downloadStop(null);
		} catch (IOException e1) {
			e1.printStackTrace();
			fListener.downloadStop(null);
		}

	}

	public void setAborted(boolean aborted) {
		this.isAborted = aborted;
	}
}
