package applet.general;

import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.io.File;
import java.text.MessageFormat;

import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;
import javax.swing.filechooser.FileFilter;

import libsidplay.C64;
import libsidplay.Player;
import libsidplay.common.Event;
import libsidplay.common.EventScheduler;
import libsidplay.common.ISID2Types;
import libsidplay.sidtune.SidTune;
import libsidplay.sidtune.SidTuneInfo;
import libsidutils.IniReader;
import resid_builder.resid.ISIDDefs.SamplingMethod;
import sidplay.ConsolePlayer.OUTPUTS;
import sidplay.ConsolePlayer.SIDEMUS;
import applet.JSIDPlay2;
import applet.TuneTab;

public class GeneralCanvas extends TuneTab {
	protected JLabel fTuneSpeedLbl;
	protected JLabel fPlayerId;
	protected JSIDPlay2 fSidplay2;
	protected JCheckBox fEnableSldb;
	protected JTextField fOutputDir;
	protected JButton fOutputDirBrwButton;
	protected JComboBox fSoundOutput;
	protected JComboBox fHardSIDDevice6581;
	protected JTextField fMP3File;
	protected JButton fMP3BrwButton;
	protected JRadioButton fPlayOriginal;
	protected JLabel fDirectory;
	protected JRadioButton fPlayEmulation;
	protected String fSOASCName;
	protected DownloadThread fDownloadThread;
	protected int fSOASCCurrentSong;
	protected JTextField fDefaultTime;
	protected JCheckBox fSingleTrack;
	protected int hardsid6581 = 1;
	protected int hardsid8580 = 1;
	protected JLabel fHardsidLbl8580;
	protected JComboBox fHardSIDDevice8580;

	protected JTextField fSOASCDwnlUrl6581R2;
	protected JTextField fSOASCDwnlUrl6581R4;
	protected JTextField fSOASCDwnlUrl8580R5;
	private JButton fDownload6581R2Btn;
	private JButton fDownload6581R4Btn;
	private JButton fDownload8580R5Btn;
	private JCheckBox fEnableProxy;
	private JTextField fProxyHost;
	private JTextField fProxyPort;
	private JComboBox frequency;
	private JComboBox fSamplingMethod;
	
	public GeneralCanvas(final JSIDPlay2 sidplay) {
		this.fSidplay2 = sidplay;
		sidplay.support.addPropertyChangeListener(this);
	}

	@Override
	public void createContents() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		final JPanel fSongTimePanel = new JPanel();
		fSongTimePanel.setBorder(new TitledBorder("Timer"));
		fSongTimePanel.setLayout(new GridBagLayout());
		add(fSongTimePanel);

		final GridBagConstraints c = new GridBagConstraints();
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.anchor = GridBagConstraints.WEST;

		final JLabel fDefaultTimeLbl = new JLabel("Default Length: ");
		c.gridx = 0;
		c.gridy = 2;
		fSongTimePanel.add(fDefaultTimeLbl, c);
		fDefaultTime = new JTextField();
		fDefaultTime.setToolTipText("Default Song Length (00:00 - unlimited)");
		c.gridx = 1;
		c.gridy = 2;
		c.fill = GridBagConstraints.NONE;
		fDefaultTime.setColumns(5);
		fSongTimePanel.add(fDefaultTime, c);
		c.fill = GridBagConstraints.HORIZONTAL;

		final int seconds = fSidplay2.getDefaultPlayLength();
		fDefaultTime.setText(String.format("%02d:%02d", seconds / 60, seconds % 60));
		fDefaultTime.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String time = fDefaultTime.getText();
				final int secs = IniReader.parseTime(time);
				if (secs != -1) {
					fSidplay2.setDefaultPlayLength(secs);
					fDefaultTime.setToolTipText("Default Song Length (00:00 - unlimited)");
					fDefaultTime.setBackground(Color.white);
				} else {
					fDefaultTime.setToolTipText("The time format is mm:ss");
					fDefaultTime.setBackground(Color.RED);
				}
			}

		});

		fEnableSldb = new JCheckBox("Respect Song Length");
		c.gridx = 0;
		c.gridy = 3;
		fSongTimePanel.add(fEnableSldb, c);

		fSingleTrack = new JCheckBox("Single Song");
		fSingleTrack.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent arg0) {
				fSidplay2.setSingleTrack(fSingleTrack.isSelected());
			}
		});
		c.gridx = 1;
		c.gridy = 3;
		fSongTimePanel.add(fSingleTrack, c);

		fEnableSldb.setSelected(fSidplay2.getEnableSLDb());
		fEnableSldb.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setEnableSLDb(fEnableSldb.isSelected());
			}

		});

		final JPanel fOutputPanel = new JPanel();
		fOutputPanel.setBorder(new TitledBorder("Output Device"));
		fOutputPanel.setLayout(new GridBagLayout());
		add(fOutputPanel);

		final JLabel outputLbl = new JLabel("Device:");
		c.gridx = 0;
		c.gridy = 0;
		fOutputPanel.add(outputLbl, c);

		fSoundOutput = new JComboBox(new String[] { "Soundcard", "HardSID4U", "WAV (Live) Recorder", "Compare to MP3/WAV/OGG Recording" });
		fSoundOutput.setSelectedIndex(0);
		fSidplay2.setOutputDevice(OUTPUTS.OUT_SOUNDCARD, SIDEMUS.EMU_RESID);
		c.gridx = 1;
		fOutputPanel.add(fSoundOutput, c);
		fSoundOutput.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				enableDisableFields();

				switch (fSoundOutput.getSelectedIndex()) {
				case 0:
					fSidplay2.setOutputDevice(OUTPUTS.OUT_SOUNDCARD, SIDEMUS.EMU_RESID);
					break;

				case 1:
					fSidplay2.setOutputDevice(OUTPUTS.OUT_NULL, SIDEMUS.EMU_HARDSID);
					break;

				case 2:
					fSidplay2.setOutputDevice(OUTPUTS.OUT_LIVE_WAV, SIDEMUS.EMU_RESID);
					break;

				case 3:
					fSidplay2.setOutputDevice(OUTPUTS.OUT_COMPARE, SIDEMUS.EMU_RESID);
					break;
				}
				fSidplay2.restartC64();
			}
			
		});

		hardsid6581 = fSidplay2.cp.m_iniCfg.emulation().getHardsid6581();
		hardsid8580 = fSidplay2.cp.m_iniCfg.emulation().getHardsid8580();

		final JPanel hardsidPanel2 = new JPanel();
		hardsidPanel2.setLayout(new FlowLayout(FlowLayout.LEFT));
		JLabel fHardsidLbl = new JLabel("6581:");
		hardsidPanel2.add(fHardsidLbl);
		// Put in enough devices for:
		// 4 HardSID PCI cards +
		// 4 HardSID USB chips +
		// 8 Network SID Device
		fHardSIDDevice6581 = new JComboBox(new Integer[] { 1, 2, 3, 4, 5, 6, 7, 8,
				9, 10, 11, 12, 13, 14, 15, 16 });
		fHardSIDDevice6581.setSelectedItem(hardsid6581);
		fHardSIDDevice6581.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				hardsid6581 = (Integer) fHardSIDDevice6581.getSelectedItem();
				fSidplay2.setHardSIDDevices(hardsid6581, hardsid8580);
			}

		});
		hardsidPanel2.add(fHardSIDDevice6581);

		fHardsidLbl8580 = new JLabel("8580:");
		hardsidPanel2.add(fHardsidLbl8580);
		fHardSIDDevice8580 = new JComboBox(new Integer[] { 1, 2, 3, 4, 5, 6, 7, 8,
				9, 10, 11, 12, 13, 14, 15, 16 });
		fHardSIDDevice8580.setSelectedItem(hardsid8580);
		fHardSIDDevice8580.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				hardsid8580 = (Integer) fHardSIDDevice8580.getSelectedItem();
				fSidplay2.setHardSIDDevices(hardsid6581, hardsid8580);
			}

		});
		hardsidPanel2.add(fHardSIDDevice8580);

		c.gridx = 2;
		fOutputPanel.add(hardsidPanel2, c);

		c.gridx = 0;
		c.gridy ++;
		fOutputPanel.add(new JLabel("Sampling rate:"), c);
		frequency = new JComboBox(new Integer[] { 44100, 48000, 96000 });
		frequency.setSelectedItem(fSidplay2.getFrequency());
		frequency.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setFrequency((Integer) frequency.getSelectedItem());
				fSidplay2.restartC64();
			}
		});
		c.gridx = 1;
		fOutputPanel.add(frequency, c);

		c.gridx = 0;
		c.gridy ++;
		fOutputPanel.add(new JLabel("Sampling:"), c);
		fSamplingMethod = new JComboBox(SamplingMethod.values());
		fSamplingMethod.setSelectedItem(fSidplay2.getSamplingMethod());
		fSamplingMethod.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setSamplingMethod((SamplingMethod) fSamplingMethod.getSelectedItem());
			}

		});
		c.gridx = 1;
		fOutputPanel.add(fSamplingMethod, c);

		fDirectory = new JLabel("Directory:");
		c.gridx = 0;
		c.gridy ++;
		fOutputPanel.add(fDirectory, c);

		fOutputDir = new JTextField();
		fOutputDir.setEditable(false);
		c.gridx = 1;
		fOutputPanel.add(fOutputDir, c);

		fOutputDirBrwButton = new JButton("Browse...", new ImageIcon(JSIDPlay2.class.getResource("icons/_active__open.png")));
		JSIDPlay2.makeDisabledIcon(fOutputDirBrwButton);
		fOutputDirBrwButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				final JFileChooser fc = new JFileChooser();
				fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				final int result = fc.showOpenDialog(fSidplay2);
				if (result == JFileChooser.APPROVE_OPTION && fc.getSelectedFile() != null) {
					fOutputDir.setText(fc.getSelectedFile().getAbsolutePath());
					fSidplay2.setOutputDir(fOutputDir.getText());
				}
			}

		});
		c.gridx = 2;
		fOutputPanel.add(fOutputDirBrwButton, c);
		fOutputDir.setText(fSidplay2.getOutputDir());

		final JPanel emuOrigPanel = new JPanel();
		emuOrigPanel.setLayout(new BoxLayout(emuOrigPanel, BoxLayout.LINE_AXIS));
		c.gridx = 0;
		c.gridy ++;
		fOutputPanel.add(emuOrigPanel, c);

		fPlayEmulation = new JRadioButton("Emu", false);
		fPlayEmulation.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setPlayOriginal(false);
			}

		});
		emuOrigPanel.add(fPlayEmulation);

		fPlayOriginal = new JRadioButton("Recording", true);
		fPlayOriginal.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setPlayOriginal(true);
			}

		});
		emuOrigPanel.add(fPlayOriginal);

		final ButtonGroup grp = new ButtonGroup();
		grp.add(fPlayEmulation);
		grp.add(fPlayOriginal);

		fMP3File = new JTextField("", 20);
		fMP3File.setEditable(false);
		fMP3File.setText(fSidplay2.cp.m_iniCfg.audio().getMp3File());
		fMP3File.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.cp.m_iniCfg.audio().setMp3File(fMP3File.getText());
			}

		});
		c.gridx = 1;
		fOutputPanel.add(fMP3File, c);

		fMP3BrwButton = new JButton("Browse...", new ImageIcon(JSIDPlay2.class.getResource("icons/_active__open.png")));
		JSIDPlay2.makeDisabledIcon(fMP3BrwButton);
		fMP3BrwButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				final JFileChooser fc = new JFileChooser();
				fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				fc.setFileFilter(new FileFilter() {

					@Override
					public String getDescription() {
						return "MP3/WAV/OGG file";
					}

					@Override
					public boolean accept(final File f) {
						if (f.isDirectory()) {
							return true;
						}
						return f.isFile() && (f.getName().endsWith(".mp3") || f.getName().endsWith(".ogg") || f.getName().endsWith(".wav"));
					}

				});
				final int result = fc.showOpenDialog(fSidplay2);
				if (result == JFileChooser.APPROVE_OPTION && fc.getSelectedFile() != null) {
					fMP3File.setText(fc.getSelectedFile().getAbsolutePath());
					fSidplay2.cp.m_iniCfg.audio().setMp3File(fMP3File.getText());
					fSidplay2.restartC64();
				}
			}

		});
		c.gridx = 2;
		fOutputPanel.add(fMP3BrwButton, c);

		add(createContentsSOASC());
		
		enableDisableFields();
		
		final JPanel options = new JPanel();
		options.setBorder(new TitledBorder("Options"));
		options.setLayout(new GridBagLayout());
		add(options);

		{
			c.gridx = 0;
			c.gridy = 2;
			final JLabel PlayerIdLbl = new JLabel("Player:");
			options.add(PlayerIdLbl, c);
			fPlayerId = new JLabel();
			c.gridx = 1;
			c.gridwidth = 2;
			options.add(fPlayerId, c);
			c.gridwidth = 1;
		}

		{
			c.gridx = 3;
			c.gridy = 2;
			final JLabel songSpeedLbl = new JLabel("Song Speed:");
			options.add(songSpeedLbl, c);
			fTuneSpeedLbl = new JLabel("?");
			c.gridx = 4;
			options.add(fTuneSpeedLbl, c);
		}
	}
	
	private JComponent createContentsSOASC() {
		GridBagConstraints c = new GridBagConstraints();
		c.weighty = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		
		final JPanel soascGroup = new JPanel();
		soascGroup.setBorder(new TitledBorder("Stone Oakvalley's Authentic SID Collection - www.6581-8580.com"));
		soascGroup.setLayout(new GridBagLayout());
		add(soascGroup);

		ActionListener proxyEnableEvent = new ActionListener() {
			public void actionPerformed(final ActionEvent actionevent) {
				fProxyHost.setEnabled(fEnableProxy.isSelected());
				fProxyPort.setEnabled(fEnableProxy.isSelected());
				fSidplay2.setProxy(fEnableProxy.isSelected(), fProxyHost.getText(),
						fProxyPort.getText());
			}
		};
		
		fEnableProxy = new JCheckBox("Enable Proxy", fSidplay2.cp.m_iniCfg.sidplay2().isEnableProxy());
		fEnableProxy.addActionListener(proxyEnableEvent);
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 3;
		soascGroup.add(fEnableProxy, c);
		c.gridwidth = 1;

		ActionListener proxyChangeEvent = new ActionListener() {
			public void actionPerformed(final ActionEvent actionevent) {
				fSidplay2.setProxy(fEnableProxy.isSelected(), fProxyHost.getText(),
						fProxyPort.getText());
			}
		};
		
		fProxyHost = new JTextField(fSidplay2.cp.m_iniCfg.sidplay2().getProxyHostname(), 20);
		fProxyHost.addActionListener(proxyChangeEvent);
		c.gridx = 0;
		c.gridy ++;
		soascGroup.add(new JLabel("Hostname and port:"), c);
		c.gridx = 1;
		soascGroup.add(fProxyHost, c);
		fProxyPort = new JTextField(String.valueOf(fSidplay2.cp.m_iniCfg.sidplay2().getProxyPort()), 5);
		fProxyPort.addActionListener(proxyChangeEvent);
		c.gridx = 2;
		soascGroup.add(fProxyPort, c);

		proxyEnableEvent.actionPerformed(null);
		
		fSOASCDwnlUrl6581R2 = new JTextField(fSidplay2.cp.m_iniCfg.sidplay2().getSoasc6581R2(), 80);
		fSOASCDwnlUrl6581R2.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				fSidplay2.cp.m_iniCfg.sidplay2().setSoasc6581R2(fSOASCDwnlUrl6581R2.getText());
			}
		});
		
		final JLabel soascDwnlURL6581R2Lbl = new JLabel("Download URL (6581R2):");
		c.gridx = 0;
		c.gridy ++;
		c.weightx = 1;
		soascGroup.add(soascDwnlURL6581R2Lbl, c);
		c.gridx = 1;
		c.weightx = 8;
		soascGroup.add(fSOASCDwnlUrl6581R2, c);
		c.gridx = 2;
		c.weightx = 1;
		fDownload6581R2Btn = new JButton("Download");
		fDownload6581R2Btn.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				String url = fSidplay2.cp.m_iniCfg.sidplay2().getSoasc6581R2();
				url = MessageFormat.format(url, fSOASCName, fSOASCCurrentSong);
				downloadStart(url);
			}

		});
		soascGroup.add(fDownload6581R2Btn, c);
		
		final JLabel soascDwnlURL6581R4Lbl = new JLabel("Download URL (6581R4):");
		c.gridx = 0;
		c.gridy ++;
		c.weightx = 1;
		soascGroup.add(soascDwnlURL6581R4Lbl, c);
		c.gridx = 1;
		c.weightx = 8;
		fSOASCDwnlUrl6581R4 = new JTextField(fSidplay2.cp.m_iniCfg.sidplay2().getSoasc6581R4(), 80);
		fSOASCDwnlUrl6581R4.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				fSidplay2.cp.m_iniCfg.sidplay2().setSoasc6581R4(fSOASCDwnlUrl6581R4.getText());
			}
		});
		soascGroup.add(fSOASCDwnlUrl6581R4, c);
		c.gridx = 2;
		c.weightx = 1;
		fDownload6581R4Btn = new JButton("Download");
		fDownload6581R4Btn.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				String url = fSidplay2.cp.m_iniCfg.sidplay2().getSoasc6581R4();
				url = MessageFormat.format(url, fSOASCName, fSOASCCurrentSong);
				downloadStart(url);
			}
		});
		soascGroup.add(fDownload6581R4Btn, c);
		
		final JLabel soascDwnlURL8580R5Lbl = new JLabel("Download URL (8580R5):");
		c.gridx = 0;
		c.gridy ++;
		c.weightx = 1;
		soascGroup.add(soascDwnlURL8580R5Lbl, c);
		c.gridx = 1;
		c.weightx = 8;
		fSOASCDwnlUrl8580R5 = new JTextField(fSidplay2.cp.m_iniCfg.sidplay2().getSoasc8580R5(), 80);
		fSOASCDwnlUrl8580R5.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				fSidplay2.cp.m_iniCfg.sidplay2().setSoasc8580R5(fSOASCDwnlUrl8580R5.getText());
			}
		});
		soascGroup.add(fSOASCDwnlUrl8580R5, c);
		c.gridx = 2;
		c.weightx = 1;
		fDownload8580R5Btn = new JButton("Download");
		fDownload8580R5Btn.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				String url = fSidplay2.cp.m_iniCfg.sidplay2().getSoasc8580R5();
				url = MessageFormat.format(url, fSOASCName, fSOASCCurrentSong);
				downloadStart(url);
			}
		});
		soascGroup.add(fDownload8580R5Btn, c);

		return soascGroup;
	}

	private void enableDisableFields() {
		final boolean fileEnabled = fSoundOutput.getSelectedIndex() == 2;
		final boolean mp3Enabled = fSoundOutput.getSelectedIndex() == 3;
		fDirectory.setEnabled(fileEnabled);
		fOutputDir.setEnabled(fileEnabled);
		fOutputDirBrwButton.setEnabled(fileEnabled);
		fMP3File.setEnabled(mp3Enabled);
		fMP3BrwButton.setEnabled(mp3Enabled);
		fHardSIDDevice6581.setEnabled(fSoundOutput.getSelectedIndex() == 1);
		fHardSIDDevice8580.setEnabled(fSoundOutput.getSelectedIndex() == 1);
		frequency.setEnabled(fSoundOutput.getSelectedIndex() != 1);
		fSamplingMethod.setEnabled(fSoundOutput.getSelectedIndex() != 1);
	}

	@Override
	public void setTune(final Player sidplay2, final SidTune sidTuneMod) {
		lastUpdate = 0;

		fPlayOriginal.setEnabled(fSidplay2.getOutputDevice() == OUTPUTS.OUT_COMPARE);
		fPlayEmulation.setEnabled(fSidplay2.getOutputDevice() == OUTPUTS.OUT_COMPARE);
		
		if (sidTuneMod == null) {
			fSOASCName = null;
			enableOrDisableDownloadButtons();
			return;
		}

		final SidTuneInfo tuneInfo = sidTuneMod.getInfo();
		final String name = fSidplay2.getHVSCName(sidTuneMod.getInfo().filename);
		if (name != null) {
			fSOASCName = name.replace(".sid", "");
			fSOASCCurrentSong = tuneInfo.currentSong;
		}
		enableOrDisableDownloadButtons();
		
		fEnableSldb.setEnabled(!"".equals(fSidplay2.getHVSCDir()));

		setTuneSpeed("");

		final StringBuilder ids = new StringBuilder();
		for (final String s : sidTuneMod.identify()) {
			if (ids.length() > 0) {
				ids.append(", ");
			}
			ids.append(s);
		}
		fPlayerId.setText(ids.toString());

		fSingleTrack.setSelected(fSidplay2.getSingleTrack());
	}

	private void enableOrDisableDownloadButtons() {
		boolean canDownload = fSOASCName != null && fDownloadThread == null;
		fDownload6581R2Btn.setEnabled(canDownload);
		fDownload6581R4Btn.setEnabled(canDownload);
		fDownload8580R5Btn.setEnabled(canDownload);
	}

	private void setTuneSpeed(final String tuneSpeed) {
		fTuneSpeedLbl.setText(tuneSpeed);
	}

	/**
	 * If some settings of the GUI changes, some settings of this panel must be set accordingly.
	 */
	public void propertyChange(final PropertyChangeEvent event) {
		if (JSIDPlay2.PROP_PLAY_ORIGINAL.equals(event.getPropertyName())) {
			final boolean playOriginal = (Boolean) event.getNewValue();
			fPlayOriginal.setSelected(playOriginal);
			fPlayEmulation.setSelected(!playOriginal);
		}
	}

	public void downloadStart(String url) {
		System.out.println("Download URL: <" + url + ">");
		fDownloadThread = new DownloadThread(fSidplay2, GeneralCanvas.this, url);
		fDownloadThread.start();

		fSidplay2.setProgressBar(0);
		enableOrDisableDownloadButtons();
	}

	public void downloadStep(final int pct) {
		fSidplay2.setProgressBar(pct);
	}

	public void downloadStop(File downloadedFile) {
		fDownloadThread = null;
		
		if (downloadedFile == null) {
			fSidplay2.setProgressBar(100);
		} else {
			fSoundOutput.setSelectedIndex(3);
			fMP3File.setText(downloadedFile.getAbsolutePath());
			fSidplay2.cp.m_iniCfg.audio().setMp3File(fMP3File.getText());
			fSidplay2.setPlayOriginal(true);
			fPlayOriginal.setSelected(true);
			fSidplay2.restartC64();
		}
		enableOrDisableDownloadButtons();
	}

	long lastUpdate;

	@Override
	public void updateTab() {
		C64 c64 = fSidplay2.cp.m_engine.getC64();
		
		final EventScheduler ctx = c64.getEventScheduler();
		final ISID2Types.Clock systemClock = c64.getClock();
		if (systemClock != null) {
			final double waitClocks = systemClock.getCpuFrequency();
	
			final long now = ctx.getTime(Event.Phase.PHI1);
			final double interval = now - lastUpdate;
			if (interval < waitClocks) {
				return;
			}
			lastUpdate = now;
	
			final double callsSinceLastRead = c64.callsToPlayRoutineSinceLastTime() * waitClocks / interval;
			/* convert to number of calls per frame */
			setTuneSpeed(String.format("%.1f calls/frame", callsSinceLastRead / systemClock.getRefresh()));
		}
	}
}
