package applet.graphical;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.GridLayout;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.border.TitledBorder;

@SuppressWarnings("serial")
abstract class Gauge extends JPanel {
	protected static final Color[] gaugeColors = new Color[256];

	static {
		for (int i = 0; i < gaugeColors.length; i ++) {
			int color;
			if (i > 128) {
				color = 256 - i << 8;
			} else {
				color = i << 8;
			}
			color |= 0x008000;
			color |= i << 16;
			gaugeColors[i] = new Color(color);
		}
	}

	/* source of color */
	protected byte[] brightness;
	/** data plots normalized between -128 and 127 */
	protected byte[] dataMin = new byte[256];
	/** data plots normalized between -128 and 127 */
	protected byte[] dataMax = new byte[256];
	/** Position within data buffer */
	protected int dataPos = 0;
	/** Our label */
	private final TitledBorder border;

	private final RenderArea area;

	private String oldLabel;

	protected class RenderArea extends JComponent {
		protected RenderArea() {
			setBackground(Color.BLACK);
			setOpaque(true);
			setIgnoreRepaint(true);
		}

		@Override
		public void update(final Graphics g) {
			paint(g);
		}

		@Override
		public void paint(final Graphics g) {
			final int width = getWidth();
			final int height = getHeight();

			g.setColor(Color.BLACK);
			g.fillRect(0, 0, width, height);

			g.setColor(Color.GREEN);
			for (int x = 0; x < width; x ++) {
				final int readPos = dataPos - width + x & dataMin.length - 1;

				if (brightness != null) {
					final int shade = brightness[readPos] + 128;
					g.setColor(gaugeColors[shade]);
				}

				g.drawLine(x, (height - 1) * (-dataMin[readPos] + 127) / 255,
						x, (height - 1) * (-dataMax[readPos] + 127) / 255);
			}
		}
	}

	public Gauge(final String label) {
		border = new TitledBorder(label);
		oldLabel = label;
		setLayout(new GridLayout());
		setBorder(border);
		area = new RenderArea();
		add(area);
	}

	protected void accumulate(final byte value) {
		if (value < dataMin[dataPos]) {
			dataMin[dataPos] = value;
		}
		if (value > dataMax[dataPos]) {
			dataMax[dataPos] = value;
		}
	}

	protected void advance() {
		final byte min = dataMin[dataPos];
		final byte max = dataMax[dataPos];
		dataPos = dataPos + 1 & dataMin.length - 1;
		dataMin[dataPos] = max;
		dataMax[dataPos] = min;
	}

	protected void setColorSource(final Gauge g) {
		brightness = g.dataMax;
	}

	public void setLabel(final String label) {
		if (label.equals(oldLabel)) {
			return;
		}
		oldLabel = label;
		border.setTitle(label);
		repaint();
	}

	public void updateGauge() {
		area.repaint();
	}

	public abstract void sample();
}
