package applet.graphical;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;

import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.TitledBorder;

import libsidplay.Player;
import libsidplay.common.Event;
import libsidplay.common.EventScheduler;
import libsidplay.common.SIDEmu;
import libsidplay.sidtune.SidTune;
import resid_builder.resid.Filter6581;
import resid_builder.resid.SID;
import resid_builder.resid.WaveformGenerator;
import sidplay.ConsolePlayer.SIDEMUS;
import applet.JSIDPlay2;
import applet.TuneTab;
import applet.events.IMuteVoice;

/**
 * @author Ken Hndel
 */
@SuppressWarnings("serial")
public class GraphicalView extends TuneTab {
	protected JSIDPlay2 fSIDPlay2;

	protected int fActivateGraphics;

	protected JCheckBox fVoice1Button;

	protected JCheckBox fVoice2Button;

	protected JCheckBox fVoice3Button;

	protected JCheckBox fDualVoice1Button;

	protected JCheckBox fDualVoice2Button;

	protected JCheckBox fDualVoice3Button;

	protected final Gauge[][][] gauges = new Gauge[2][4][3];

	int repaint = 0;

	protected final HighResolutionEvent highResolutionEvent = new HighResolutionEvent();

	protected class HighResolutionEvent extends Event {
		private EventScheduler ctx;

		public HighResolutionEvent() {
			super("High Resolution SID Register Sampler");
		}

		public void beginScheduling(final EventScheduler _ctx) {
			ctx = _ctx;
			ctx.schedule(this, 0, Event.Phase.PHI2);
		}

		@Override
		public void event() {
			for (int i = 0; i < fActivateGraphics; i++) {
				SIDEmu sidemu = fSIDPlay2.getSIDEmu(i);
				if (sidemu == null) {
					continue;
				}

				sidemu.clock();
				for (int j = 0; j < 4; j ++) {
					gauges[i][j][0].sample();
					gauges[i][j][1].sample();
					gauges[i][j][2].sample();

					gauges[i][j][0].advance();
					if ((repaint & 127) == 0) {
						gauges[i][j][1].advance();
						gauges[i][j][2].advance();
					}

					if (isVisible() && (repaint & 255) == 0) {
						gauges[i][j][0].updateGauge();
						gauges[i][j][1].updateGauge();
						gauges[i][j][2].updateGauge();
					}
				}
			}


			++ repaint;
			ctx.schedule(this, 128);
		}
	}

	public GraphicalView(final JSIDPlay2 sidplay) {
		this.fSIDPlay2 = sidplay;
		sidplay.support.addPropertyChangeListener(this);
	}

	@Override
	public void createContents() {
		setLayout(new GridBagLayout());

		final GridBagConstraints c = new GridBagConstraints();
		c.weightx = 1;
		c.fill = GridBagConstraints.BOTH;

		final JPanel muteVoices = new JPanel();
		muteVoices.setBorder(new TitledBorder("Voice Control"));
		muteVoices.setLayout(new GridLayout(2, 6));
		c.gridwidth = GridBagConstraints.REMAINDER;
		add(muteVoices, c);

		final JLabel monoSidLbl = new JLabel("Mono SID:");
		Font deriveFont = monoSidLbl.getFont().deriveFont(Font.BOLD);
		monoSidLbl.setFont(deriveFont);
		muteVoices.add(monoSidLbl);

		fVoice1Button = new JCheckBox("Mute Voice 1");
		muteVoices.add(fVoice1Button);
		fVoice2Button = new JCheckBox("Mute Voice 2");
		muteVoices.add(fVoice2Button);
		fVoice3Button = new JCheckBox("Mute Voice 3");
		muteVoices.add(fVoice3Button);
		fVoice1Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				setVoiceMute(0, 0, fVoice1Button.isSelected());
			}

		});
		fVoice2Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				setVoiceMute(0, 1, fVoice2Button.isSelected());
			}

		});
		fVoice3Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				setVoiceMute(0, 2, fVoice3Button.isSelected());
			}

		});

		final JLabel stereoSidLbl = new JLabel("Stereo SID:");
		deriveFont = stereoSidLbl.getFont().deriveFont(Font.BOLD);
		stereoSidLbl.setFont(deriveFont);
		muteVoices.add(stereoSidLbl);

		fDualVoice1Button = new JCheckBox("Mute Voice 1");
		muteVoices.add(fDualVoice1Button);
		fDualVoice2Button = new JCheckBox("Mute Voice 2");
		muteVoices.add(fDualVoice2Button);
		fDualVoice3Button = new JCheckBox("Mute Voice 3");
		muteVoices.add(fDualVoice3Button);
		fDualVoice1Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				setVoiceMute(1, 0, fDualVoice1Button.isSelected());
			}

		});
		fDualVoice2Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				setVoiceMute(1, 1, fDualVoice2Button.isSelected());
			}

		});
		fDualVoice3Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				setVoiceMute(1, 2, fDualVoice3Button.isSelected());
			}

		});

		// mono/stereo SID
		final JPanel sidCurves = new JPanel();
		sidCurves.setLayout(new GridLayout(1, 2));
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weighty = 1.0;
		add(sidCurves, c);

		final JPanel curves1 = new JPanel();
		curves1.setBorder(new TitledBorder("Mono SID"));
		curves1.setLayout(new GridLayout(4, 3));
		sidCurves.add(curves1);
		constructGauges(0, curves1);

		final JPanel curves2 = new JPanel();
		curves2.setBorder(new TitledBorder("Stereo SID"));
		curves2.setLayout(new GridLayout(4, 3));
		sidCurves.add(curves2);
		constructGauges(1, curves2);
	}

	private void constructGauges(final int sidNum, final JPanel curves) {
		Gauge g;
		for (int voice = 0; voice < 3; voice ++) {
			final int v = voice;
			g = new Gauge("Wave") {
				@Override
				public void sample() {
					SID sid = fSIDPlay2.getSID(sidNum);
					final WaveformGenerator wave = sid.voice[v].wave;
					final WaveformGenerator waveRing = sid.voice[(v + 2) % 3].wave;
					int sampleValue = (int) (wave.output(waveRing) - wave.getZeroLevel());
					sampleValue /= 2048/128;
					sampleValue -= 128;
					if (sampleValue > 127) {
						sampleValue = 127;
					}
					if (sampleValue < -128) {
						sampleValue = -128;
					}
					accumulate((byte) sampleValue);
				}

				@Override
				public void updateGauge() {
					super.updateGauge();
					SIDEmu sidemu = fSIDPlay2.getSIDEmu(sidNum);
					if (sidemu == null) {
						setLabel("Wave");
					} else {
						final byte wf = sidemu.readInternalRegister(4 + 7 * v);
						final byte filt = sidemu.readInternalRegister(0x17);
						setLabel(String.format("Wave %X %s%s%s%s",
								wf >> 4 & 0xf,
								(wf & 2) != 0 ? "S" : "",
										(wf & 4) != 0 ? "R" : "",
												(wf & 8) != 0 ? "T" : "",
														(filt & 1 << v) != 0 ? "F" : ""));
					}
				}
			};
			gauges[sidNum][voice][0] = g;
			curves.add(g);

			g = new Gauge("Envelope") {
				@Override
				public void sample() {
					SID sid = fSIDPlay2.getSID(sidNum);
					final float envOutput = sid.voice[v].envelope.output();
					int value = (int) (Math.log(envOutput / 256.0) / Math.log(10) * 20 * 8) + 127;
					if (value < -128) {
						value = -128;
					}
					accumulate((byte) value);
				}
			};
			gauges[sidNum][voice][1] = g;
			curves.add(g);

			g = new Gauge("Frequency") {
				@Override
				public void sample() {
					SIDEmu sidemu = fSIDPlay2.getSIDEmu(sidNum);
					int frqValue = (sidemu.readInternalRegister(1 + v * 7) & 0xff) << 8 | sidemu.readInternalRegister(0 + v * 7) & 0xff;

					if (frqValue != 0) {
						/* Tuned for FR=7382 as approximate for 440 Hz at 1 MHz clock. */
						frqValue = (int) (Math.log(frqValue) / Math.log(2) * 12 * 2 - 0.19755 + 0.5);
					}
					frqValue -= 12 * 32 + 1;
					frqValue += 127;
					if (frqValue < -128) {
						frqValue = -128;
					}
					if (frqValue > 127) {
						frqValue = 127;
					}

					accumulate((byte) frqValue);
				}
			};
			g.setColorSource(gauges[sidNum][voice][1]);

			gauges[sidNum][voice][2] = g;
			curves.add(g);
		}

		g = new Gauge("Master volume") {
			@Override
			public void sample() {
				SIDEmu sidemu = fSIDPlay2.getSIDEmu(sidNum);
				int sample = sidemu.readInternalRegister(0x18) & 0xf;
				sample *= 255;
				sample /= 15;
				sample -= 128;
				accumulate((byte) sample);
			}
		};
		gauges[sidNum][3][0] = g;
		curves.add(g);

		g = new Gauge("Resonance") {
			@Override
			public void sample() {
				SIDEmu sidemu = fSIDPlay2.getSIDEmu(sidNum);
				int res = sidemu.readInternalRegister(0x17) >> 4 & 0xf;
					res *= 255;
					res /= 15;
					res -= 128;
					accumulate((byte) res);
			}
		};

		gauges[sidNum][3][1] = g;
		curves.add(g);

		g = new Gauge("Filter") {
			int brightAvg;
			int brightN;

			{
				brightness = new byte[dataMin.length];
			}

			@Override
			public void sample() {
				SIDEmu sidemu = fSIDPlay2.getSIDEmu(sidNum);
				int fc = sidemu.readInternalRegister(0x15) & 7;
				fc |= (sidemu.readInternalRegister(0x16) & 0xff) << 3;
				fc *= 255;
				fc /= 2047;
				fc -= 128;
				accumulate((byte) fc);

				SID sid = fSIDPlay2.getSID(sidNum);
				final Filter6581 filter = sid.getFilter6581();
				final int distortion = (int) (filter.estimateCurrentDistortion() * 2e4);
				brightAvg += distortion;
				brightN ++;
			}

			@Override
			protected void advance() {
				int brightnessAvg = brightN != 0 ? brightAvg / brightN : 0;
				brightN = 0;
				brightAvg = 0;
				brightnessAvg -= 128;
				if (brightnessAvg < -128) {
					brightnessAvg = -128;
				}
				if (brightnessAvg > 127) {
					brightnessAvg = 127;
				}
				brightness[dataPos] = (byte) brightnessAvg;
				super.advance();
			}

			@Override
			public void updateGauge() {
				super.updateGauge();
				SIDEmu sidemu = fSIDPlay2.getSIDEmu(sidNum);
				if (sidemu == null) {
					setLabel("Filter");
				} else {
					final byte vol = sidemu.readInternalRegister(0x18);
					setLabel("Filter "
							+ ((vol & 0x10) != 0 ? "L" : "")
							+ ((vol & 0x20) != 0 ? "B" : "")
							+ ((vol & 0x40) != 0 ? "H" : ""));
				}
			}
		};
		gauges[sidNum][3][2] = g;
		curves.add(g);
	}

	protected void setVoiceMute(final int sidNum, final int voiceNo, final boolean mute) {
		fSIDPlay2.setMute(sidNum, voiceNo, mute);
	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
		/* enable 1 or 2 displays? */
		fActivateGraphics = m_engine.getConfig().channels;

		if (fSIDPlay2.getOutputDeviceEmu() != SIDEMUS.EMU_HARDSID) {
			final EventScheduler ctx = fSIDPlay2.cp.m_engine.getC64().getEventScheduler();
			/* sample oscillator buffer */
			highResolutionEvent.beginScheduling(ctx);
		}

		setVoiceMute(0, 0, fVoice1Button.isSelected());
		setVoiceMute(0, 1, fVoice2Button.isSelected());
		setVoiceMute(0, 2, fVoice3Button.isSelected());
		setVoiceMute(1, 0, fDualVoice1Button.isSelected());
		setVoiceMute(1, 1, fDualVoice2Button.isSelected());
		setVoiceMute(1, 2, fDualVoice3Button.isSelected());
	}

	public void propertyChange(final PropertyChangeEvent evt) {
		if (IMuteVoice.class.getName().equals(evt.getPropertyName())) {
			final IMuteVoice muteVoice = (IMuteVoice) evt.getNewValue();
			final int sidNum = muteVoice.getSidNum();
			final int voice = muteVoice.getVoice();
			if (sidNum == 0) {
				if (voice == 0 && !evt.getSource().equals(fVoice1Button)) {
					fVoice1Button.setSelected(muteVoice.isMute());
				} else if (voice == 1 && !evt.getSource().equals(fVoice2Button)) {
					fVoice2Button.setSelected(muteVoice.isMute());
				} else if (voice == 2 && !evt.getSource().equals(fVoice3Button)) {
					fVoice3Button.setSelected(muteVoice.isMute());
				}
			} else {
				if (voice == 0 && !evt.getSource().equals(fDualVoice1Button)) {
					fDualVoice1Button.setSelected(muteVoice.isMute());
				} else if (voice == 1
						&& !evt.getSource().equals(fDualVoice2Button)) {
					fDualVoice2Button.setSelected(muteVoice.isMute());
				} else if (voice == 2
						&& !evt.getSource().equals(fDualVoice3Button)) {
					fDualVoice3Button.setSelected(muteVoice.isMute());
				}
			}
		}
	}
}
