package applet.hardsid;

import static hardsid_builder.HardSID.SID_DEVICES;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.border.TitledBorder;

import netsiddev.NetworkSIDDevice;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;
import applet.JSIDPlay2;
import applet.TuneTab;

/**
 * Virtual HardSID functionality installs a HardSID.dll to communicate
 * between the HardSID.dll and JSidplay2 via a network interface.
 * 
 * Protocol details can be found in NetworkSidDevice class.
 */
@SuppressWarnings("serial")
public class HardSIDView extends TuneTab {
	/** Applet instance */
	protected JSIDPlay2 fSidplay2;

	private NetworkSIDDevice sidDevice;

	public HardSIDView(final JSIDPlay2 sidplay) {
		fSidplay2 = sidplay;
		sidplay.support.addPropertyChangeListener(this);

		sidDevice = new NetworkSIDDevice(SID_DEVICES);
		sidDevice.start();
	}

	@Override
	public void createContents() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		final JPanel integrationPanel = new JPanel();
		integrationPanel.setBorder(new TitledBorder("Network SID Device"));
		integrationPanel.setLayout(new GridBagLayout());
		add(integrationPanel);

		final JTextArea note = new JTextArea();
		note.setEditable(false);
		note.setOpaque(false);
		note.setForeground(Color.BLUE);
		note.setLineWrap(true);
		note.setWrapStyleWord(true);
		note.append("This feature allows you to use this player as a sound device with ACID64 by Wilfred Bos (www.acid64.com)\n");
		note.append("Network port 6581 is used for the connection between JSIDPlay2 and ACID64.\n");
		note.append("Usable Network SID Devices can be chosen in the external player:\n\n");
		note.append("1: MOS6581, PAL, Resample\n");
		note.append("2: MOS8580, PAL, Resample\n");
		note.append("3: MOS6581, NTSC, Resample\n");
		note.append("4: MOS8580, NTSC, Resample\n");
		note.append("5: MOS6581, PAL, Decimate\n");
		note.append("6: MOS8580, PAL, Decimate\n");
		note.append("7: MOS6581, NTSC, Decimate\n");
		note.append("8: MOS8580, NTSC, Decimate\n\n");
		note.append("Other HardSID enabled software may or may not run as supposed\n");
		note.append("and will require to install the HardSID libraries, first\n");
		note.append("by clicking on the browse button, while ACID64 runs out of the box!");

		GridBagConstraints c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.BOTH;
		integrationPanel.add(note, c);

		final JLabel applicationLbl = new JLabel(
				"Directory of any other HardSID enabled application:");
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		integrationPanel.add(applicationLbl, c);

		final JButton fApplicationDirBrwButton = new JButton("Browse...",
				new ImageIcon(
						JSIDPlay2.class.getResource("icons/_active__open.png")));
		JSIDPlay2.makeDisabledIcon(fApplicationDirBrwButton);
		fApplicationDirBrwButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				final JFileChooser fc = new JFileChooser();
				fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				final int result = fc.showOpenDialog(fSidplay2);
				if (result == JFileChooser.APPROVE_OPTION
						&& fc.getSelectedFile() != null) {
					installDLL(fc.getSelectedFile(),
							"/hardsid/cpp/Debug/HardSID.dll", "HardSID.dll");
					installDLL(fc.getSelectedFile(),
							"/hardsid/cpp/Debug/HardSID_orig.dll",
							"HardSID_orig.dll");
				}
			}

		});
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		integrationPanel.add(fApplicationDirBrwButton, c);

	}

	protected void copyFile(final File src, final File dst) throws IOException {
		final InputStream in = new FileInputStream(src);
		final OutputStream out = new FileOutputStream(dst);
		copyFile(in, out);
		in.close();
		out.close();
	}

	protected void copyFile(final InputStream in, final OutputStream out)
			throws IOException {
		// Transfer bytes from in to out
		final byte[] buf = new byte[1024];
		int len;
		while ((len = in.read(buf)) > 0) {
			out.write(buf, 0, len);
		}
	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
	}

	private void installDLL(final File dir, final String dllResource,
			final String dllName) {
		final InputStream str = getClass().getResourceAsStream(dllResource);
		final File f = new File(dir, dllName);
		try {
			final File backupFile = new File(f.getAbsolutePath() + ".bak");
			if (f.exists() && !backupFile.exists()) {
				// create backup
				System.out.println("Create backup: "
						+ backupFile.getAbsolutePath());
				copyFile(f, backupFile);
			}
			if (f.exists()) {
				// remove old library
				f.delete();
			}
			// install new library
			System.out.println("Save HardSID library: " + f.getAbsolutePath());
			final FileOutputStream bout = new FileOutputStream(f);
			copyFile(str, bout);
			str.close();
			bout.close();
		} catch (final IOException e1) {
			e1.printStackTrace();
		}
	}

	public void propertyChange(final PropertyChangeEvent arg0) {
	}
}
