package applet.joystick;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;

import libsidplay.Player;
import libsidplay.components.joystick.IJoystick;
import libsidplay.sidtune.SidTune;
import net.java.games.input.Component;
import net.java.games.input.Controller;
import net.java.games.input.ControllerEnvironment;
import applet.TuneTab;

@SuppressWarnings("serial")
public class JoystickTab extends TuneTab {
	private static final String LBL_DEVICE = "Device";
	private static final String LBL_UP = "Up";
	private static final String LBL_DOWN = "Down";
	private static final String LBL_LEFT = "Left";
	private static final String LBL_RIGHT = "Right";
	private static final String LBL_FIRE = "Fire";

	protected static ControllerEnvironment fControllerEnv = ControllerEnvironment.getDefaultEnvironment();

	protected class JoystickPort {
		protected final int portNumber;
		protected final JCheckBox master;

		protected final JComboBox deviceCombo, upCombo, downCombo, leftCombo, rightCombo, fireCombo;
		protected final JTextField upText, downText, leftText, rightText, fireText;
		protected final JoystickTestModel model = new JoystickTestModel();
		protected final JTable testTable = new JTable(model);

		public JoystickPort(final int portNumber) {
			this.portNumber = portNumber;
			master = new JCheckBox("Joystick in port " + portNumber + " enabled");

			deviceCombo = new JComboBox();
			upCombo = new JComboBox();
			downCombo = new JComboBox();
			leftCombo = new JComboBox();
			rightCombo = new JComboBox();
			fireCombo = new JComboBox();
			upText = new JTextField();
			downText = new JTextField();
			leftText = new JTextField();
			rightText = new JTextField();
			fireText = new JTextField();

			enableComponents();
		}

		protected Controller device;

		protected Component up, down, left, right, fire;

		protected float upValue, downValue, leftValue, rightValue, fireValue;

		protected void enableComponents() {
			boolean state = master.isSelected();

			deviceCombo.setEnabled(state);

			if (device == null) {
				state = false;
			}

			upCombo.setEnabled(state);
			downCombo.setEnabled(state);
			leftCombo.setEnabled(state);
			rightCombo.setEnabled(state);
			fireCombo.setEnabled(state);

			upText.setEnabled(state);
			downText.setEnabled(state);
			leftText.setEnabled(state);
			rightText.setEnabled(state);
			fireText.setEnabled(state);
		}

		protected JPanel buildComponents() {
			final GridBagConstraints inputs = new GridBagConstraints();
			inputs.fill = GridBagConstraints.HORIZONTAL;
			final GridBagConstraints labels = (GridBagConstraints) inputs.clone();
			inputs.weightx = 1.0;

			final JPanel deviceSelection = new JPanel();
			deviceSelection.setLayout(new GridBagLayout());
			deviceSelection.setBorder(new TitledBorder("Port " + portNumber));

			deviceSelection.add(new JLabel(LBL_DEVICE), labels);
			inputs.gridwidth = GridBagConstraints.REMAINDER;
			deviceSelection.add(deviceCombo, inputs);

			deviceSelection.add(new JLabel(LBL_UP), labels);
			inputs.gridwidth = GridBagConstraints.RELATIVE;
			deviceSelection.add(upCombo, inputs);
			inputs.gridwidth = GridBagConstraints.REMAINDER;
			deviceSelection.add(upText, inputs);

			deviceSelection.add(new JLabel(LBL_DOWN), labels);
			inputs.gridwidth = GridBagConstraints.RELATIVE;
			deviceSelection.add(downCombo, inputs);
			inputs.gridwidth = GridBagConstraints.REMAINDER;
			deviceSelection.add(downText, inputs);

			deviceSelection.add(new JLabel(LBL_LEFT), labels);
			inputs.gridwidth = GridBagConstraints.RELATIVE;
			deviceSelection.add(leftCombo, inputs);
			inputs.gridwidth = GridBagConstraints.REMAINDER;
			deviceSelection.add(leftText, inputs);

			deviceSelection.add(new JLabel(LBL_RIGHT), labels);
			inputs.gridwidth = GridBagConstraints.RELATIVE;
			deviceSelection.add(rightCombo, inputs);
			inputs.gridwidth = GridBagConstraints.REMAINDER;
			deviceSelection.add(rightText, inputs);

			deviceSelection.add(new JLabel(LBL_FIRE), labels);
			inputs.gridwidth = GridBagConstraints.RELATIVE;
			deviceSelection.add(fireCombo, inputs);
			inputs.gridwidth = GridBagConstraints.REMAINDER;
			deviceSelection.add(fireText, inputs);

			for (final JComboBox jcb : new JComboBox[] { upCombo, downCombo, leftCombo, rightCombo, fireCombo }) {
				jcb.addActionListener(new ActionListener() {
					public void actionPerformed(final ActionEvent e) {
						final JComboBox direction = (JComboBox) e.getSource();
						final int index = direction.getSelectedIndex();
						if (index < 0 || device == null) {
							return;
						}

						final Component c = device.getComponents()[index];
						/* I can't figure out how to associate the JComboBox to my Component */
						if (direction == upCombo) {
							up = c;
						} else if (direction == downCombo) {
							down = c;
						} else if (direction == leftCombo) {
							left = c;
						} else if (direction == rightCombo) {
							right = c;
						} else if (direction == fireCombo) {
							fire = c;
						}
					}
				});
			}

			for (final JTextField jtf : new JTextField[] { upText, downText, leftText, rightText, fireText }) {
				jtf.addActionListener(new ActionListener() {
					public void actionPerformed(final ActionEvent e) {
						final JTextField valueField = (JTextField) e.getSource();
						componentValueChanged(valueField);
					}
				});
				jtf.setText("1.0");
				componentValueChanged(jtf);
			}

			addJoystickSelection();

			return deviceSelection;
		}

		protected void addJoystickSelection() {
			deviceCombo.addActionListener(new ActionListener() {
				public void actionPerformed(final ActionEvent e) {
					final int index = deviceCombo.getSelectedIndex();
					if (index == -1) {
						return;
					}

					device = fControllerEnv.getControllers()[index];
					model.setInput(device);
					testTable.repaint();

					upCombo.removeAllItems();
					downCombo.removeAllItems();
					leftCombo.removeAllItems();
					rightCombo.removeAllItems();
					fireCombo.removeAllItems();

					int componentIndex = 0;
					for (final Component c : device.getComponents()) {
						final String componentName = componentIndex++ + ": " + c.getName();
						upCombo.addItem(componentName);
						downCombo.addItem(componentName);
						leftCombo.addItem(componentName);
						rightCombo.addItem(componentName);
						fireCombo.addItem(componentName);
					}

					enableComponents();

					upCombo.repaint();
					downCombo.repaint();
					leftCombo.repaint();
					rightCombo.repaint();
					fireCombo.repaint();
				}
			});

			if (fControllerEnv.isSupported()) {
				for (final Controller c : fControllerEnv.getControllers()) {
					deviceCombo.addItem(c.getType() + " (" + c.getPortType() + ")");
				}
			}
		}

		protected void componentValueChanged(final JTextField valueField) {
			final float value = Float.valueOf(valueField.getText());
			if (valueField == upText) {
				upValue = value;
			} else if (valueField == downText) {
				downValue = value;
			} else if (valueField == leftText) {
				leftValue = value;
			} else if (valueField == rightText) {
				rightValue = value;
			} else if (valueField == fireText) {
				fireValue = value;
			}
		}

	}

	protected JoystickPort[] joystickPorts = new JoystickPort[] {
			new JoystickPort(1),
			new JoystickPort(2),
	};


	long oldRenderTime;
	@Override
	public void updateTab() {
		final long currentTime = System.currentTimeMillis();
		if (currentTime - oldRenderTime > 500) {
			oldRenderTime = currentTime;
			joystickPorts[0].testTable.repaint();
			joystickPorts[1].testTable.repaint();
		}
	}

	public JoystickTab() {
	}

	/**
	 * Create GUI elements here
	 */
	@Override
	public void createContents() {
		setBorder(new TitledBorder("Joystick Configuration"));

		setLayout(new GridBagLayout());

		final GridBagConstraints constraints = new GridBagConstraints();
		constraints.weightx = 1.0;
		constraints.fill = GridBagConstraints.HORIZONTAL;

		{
			final JCheckBox fActivateJoy1 = joystickPorts[0].master;
			final Font deriveFont = fActivateJoy1.getFont().deriveFont(Font.BOLD);
			fActivateJoy1.setFont(deriveFont);
			constraints.gridwidth = GridBagConstraints.RELATIVE;
			add(fActivateJoy1, constraints);
			fActivateJoy1.addActionListener(new ActionListener() {
				public void actionPerformed(final ActionEvent e) {
					joystickPorts[0].enableComponents();
				}
			});
		}

		{
			final JCheckBox fActivateJoy2 = joystickPorts[1].master;
			final Font deriveFont = fActivateJoy2.getFont().deriveFont(Font.BOLD);
			fActivateJoy2.setFont(deriveFont);
			constraints.gridwidth = GridBagConstraints.REMAINDER;
			add(fActivateJoy2, constraints);
			fActivateJoy2.addActionListener(new ActionListener() {
				public void actionPerformed(final ActionEvent e) {
					joystickPorts[1].enableComponents();
				}
			});
		}

		constraints.gridwidth = GridBagConstraints.RELATIVE;
		add(joystickPorts[0].buildComponents(), constraints);
		constraints.gridwidth = GridBagConstraints.REMAINDER;
		add(joystickPorts[1].buildComponents(), constraints);

		constraints.fill = GridBagConstraints.BOTH;
		constraints.weighty = 1.0;
		constraints.gridwidth = GridBagConstraints.RELATIVE;
		add(new JScrollPane(joystickPorts[0].testTable), constraints);
		constraints.gridwidth = GridBagConstraints.REMAINDER;
		add(new JScrollPane(joystickPorts[1].testTable), constraints);
	}

	protected IJoystick getJoystickReader(final JoystickPort port) {
		return new IJoystick() {
			private long lastPollingTime;
			private byte bits;
			public byte getValue() {
				if (!port.master.isSelected()) {
					return (byte) 0xff;
				}

				/* throttle polling to max. once every 5 ms */
				final long currentTime = System.currentTimeMillis();
				if (currentTime > lastPollingTime + 5) {
					port.device.poll();
					lastPollingTime = currentTime;

					bits = (byte) 0xff;
					if (port.up != null
							&& Math.abs(port.up.getPollData() - port.upValue) < 0.1) {
						bits ^= 1;
					}
					if (port.down != null
							&& Math.abs(port.down.getPollData() - port.downValue) < 0.1) {
						bits ^= 2;
					}
					if (port.left != null
							&& Math.abs(port.left.getPollData() - port.leftValue) < 0.1) {
						bits ^= 4;
					}
					if (port.right != null
							&& Math.abs(port.right.getPollData() - port.rightValue) < 0.1) {
						bits ^= 8;
					}
					if (port.fire != null
							&& Math.abs(port.fire.getPollData() - port.fireValue) < 0.1) {
						bits ^= 16;
					}
				}

				return bits;
			}
		};
	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
		for (int i = 0; i < 2; i ++) {
			m_engine.setJoystick(i, getJoystickReader(joystickPorts[i]));
		}
	}

	public void propertyChange(final PropertyChangeEvent arg0) {
	}
}