package applet.settings;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JPanel;

import resid_builder.FilterConfig;
import resid_builder.resid.ISIDDefs.ChipModel;
import resid_builder.resid.SID;
import applet.JSIDPlay2;

@SuppressWarnings("serial")
public class CurveFit extends JPanel implements PropertyChangeListener {
	protected final JSIDPlay2 sidplay2;

	private static final int xmin = 0;
	private static final int xmax = 2048;
	private static final int ymin = 160;
	private static final int ymax = 24000;
	private boolean logScale;

	private final Font fLabelFont = new Font("Courier", Font.BOLD | Font.ITALIC, 18);
	protected FilterConfig filter;

	protected class HighResolutionPoint {
		protected double x, y;

		protected HighResolutionPoint(final double x, final double y) {
			this.x = x;
			this.y = y;
		}

		protected Point toPoint() {
			return new Point((int) (x + 0.5), (int) (y + 0.5));
		}
	}

	protected final int distance(final int x1, final int y1, final int x2, final int y2) {
		final int dx = x1 - x2;
		final int dy = y1 - y2;
		return (int) (Math.sqrt(dx * dx + dy * dy) + 0.5);
	}

	public CurveFit(final JSIDPlay2 play2, final SettingsPanel settingsPanel) {
		sidplay2 = play2;
		play2.support.addPropertyChangeListener(this);
		// setBackground(Color.LIGHT_GRAY.brighter());
		addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				final int x = e.getX();
				final int y = e.getY();

				/* ok, now we need to find the curve closest to here. */
				String filterName = null;
				int maximumScan = 100;

				for (final String name : sidplay2.cp.m_iniCfg.getFilterList(filter.getK() != 0 ? ChipModel.MOS8580 : ChipModel.MOS6581)) {
					final FilterConfig thisFilter = sidplay2.cp.m_iniCfg.getFilter(name);

					int scanRadius = 0;
					while (scanRadius < maximumScan) {
						/* find the curve closest to clicked point. */
						Point p;
						int distance;

						final int xMinusScanInv = xmin + (xmax - xmin) * (x - scanRadius) / getWidth();
						if (xMinusScanInv >= xmin) {
							p = viewtransform(xMinusScanInv, evaluate_filter(thisFilter, xMinusScanInv)).toPoint();
							distance = distance(x, y, p.x, p.y);
							if (distance < maximumScan) {
								maximumScan = distance;
								filterName = name;
							}
						}

						final int xPlusScanInv = xmin + (xmax - xmin) * (x + scanRadius) / getWidth();
						if (xPlusScanInv < xmax) {
							p = viewtransform(xPlusScanInv, evaluate_filter(thisFilter, xPlusScanInv)).toPoint();
							distance = distance(x, y, p.x, p.y);
							if (distance < maximumScan) {
								maximumScan = distance;
								filterName = name;
							}
						}
						scanRadius++;
					}
				}

				if (filterName != null) {
					settingsPanel.fFilter.setSelectedItem(filterName);
				}
			}
		});
	}

	@Override
	protected void paintComponent(final Graphics g) {
		super.paintComponent(g);

		final Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);

		{
			g2.setPaint(new Color(0, 0, 0, 255));
			Point from = viewtransform(xmin, ymin).toPoint();
			Point to = viewtransform(xmax, ymin).toPoint();
			g2.drawLine(from.x, from.y, to.x, to.y);
			from = to;
			to = viewtransform(xmax, ymax).toPoint();
			g2.drawLine(from.x, from.y, to.x, to.y);
			from = to;
			to = viewtransform(xmin, ymax).toPoint();
			g2.drawLine(from.x, from.y, to.x, to.y);
			from = to;
			to = viewtransform(xmin, ymin).toPoint();
			g2.drawLine(from.x, from.y, to.x, to.y);
		}

		for (int x = xmin - xmin % 128; x < xmax; x += 128) {
			final Point from = viewtransform(x, ymin).toPoint();
			final Point to = viewtransform(x, ymax).toPoint();
			g2.setPaint(x % 512 == 0 ? new Color(32, 32, 32, 255) : new Color(128, 128, 128, 255));
			g2.drawLine(from.x, from.y, to.x, to.y);
		}

		for (int y = ymin - ymin % 100; y < ymax; y += 100) {
			final Point from = viewtransform(0, y).toPoint();
			final Point to = viewtransform(xmax, y).toPoint();
			g2.setPaint(y % 1000 == 0 ? new Color(32, 32, 32, 255) : new Color(192, 192, 192, 255));
			g2.drawLine(from.x, from.y, to.x, to.y);
		}

		if (filter == null) {
			return;
		}

		/* render all known filters, and ours last */
		g2.setPaint(new Color(224, 224, 224, 128));
		for (final String name : sidplay2.cp.m_iniCfg.getFilterList(filter.getK() != 0 ? ChipModel.MOS8580 : ChipModel.MOS6581)) {

			final FilterConfig thisFilter = sidplay2.cp.m_iniCfg.getFilter(name);
			if (filter == thisFilter) {
				continue;
			}

			renderFilter(g2, thisFilter);
		}

		if (filter.getK() == 0f) {
			// default filter
			g2.setPaint(Color.green.darker());
		} else {
			// type 3 filter
			g2.setPaint(Color.blue.darker());
		}

		renderFilter(g2, filter);

		{
			final String xAxis = "FC value";
			final String yAxis = "Center frequency (Hz)";
			g2.setFont(fLabelFont);
			final FontMetrics fontMetrics = g.getFontMetrics();

			g2.setPaint(Color.black.darker());
			g2.drawString(yAxis, 10, 30);
			Rectangle bounds = fontMetrics.getStringBounds(xAxis, g).getBounds();
			g2.drawString(xAxis, (getWidth() - bounds.width - 10), (getHeight() - bounds.height));

			String type = "(unknown filter)";
			if (filter != null && filter.getBaseresistance() != 0f) {
				type = "Type 3";
			} else if (filter != null && filter.getK() != 0f) {
				type = "Type 4";
			}

			bounds = fontMetrics.getStringBounds(yAxis, g).getBounds();
			g2.drawString(type, 10, (30 + bounds.height));
		}
	}

	private void renderFilter(final Graphics2D g, final FilterConfig currentFilter) {
		HighResolutionPoint last = viewtransform(0, evaluate_filter(currentFilter, 0));
		HighResolutionPoint tesselation = last;
		Double prevAngle = null;
		for (int i = 1; i < 2048; i += 1) {
			final HighResolutionPoint tesselationPrev = tesselation;

			tesselation = viewtransform(i, evaluate_filter(currentFilter, i));
			/* don't bother if we'd be rendering on the very same X coordinate eventually. */
			if ((int) tesselation.x - (int) last.x == 0) {
				continue;
			}

			final Double angle = Math.atan2(last.y - tesselation.y, last.x - tesselation.x);

			/*
			 * does this point continue in roughly the same direction as the
			 * other points so far? (If first point, record the current angle
			 * from last point as the trial direction.)
			 */
			if (prevAngle == null) {
				prevAngle = angle;
				continue;
			}

			if (Math.abs(angle - prevAngle) < 0.05) {
				continue;
			}

			prevAngle = null;
			/*
			 * the new angle is too different; draw to previous good point,
			 * begin calculating angle from there.
			 */
			final Point tesselationPrevPoint = tesselationPrev.toPoint();
			final Point lastPoint = last.toPoint();
			// g.drawLine(lastPoint.x, lastPoint.y, lastPoint.x + 1,
			// lastPoint.y);
			g.drawLine(lastPoint.x, lastPoint.y, tesselationPrevPoint.x, tesselationPrevPoint.y);
			last = tesselationPrev;
		}

		if (last != tesselation) {
			final Point lastPoint = last.toPoint();
			final Point tesselationPoint = tesselation.toPoint();
			g.drawLine(lastPoint.x, lastPoint.y, tesselationPoint.x, tesselationPoint.y);
		}
	}

	protected final HighResolutionPoint viewtransform(final int coordX, final double coordY) {
		double x, y;
		x = coordX;
		x /= (double) xmax / getWidth();

		if (logScale) {
			y = logscale(coordY);
			y -= logscale(ymin);
			y /= (logscale(ymax) - logscale(ymin)) / getHeight();
		} else {
			y = coordY;
			y -= ymin;
			y /= (ymax - ymin) / getHeight();
		}

		return new HighResolutionPoint(x, getHeight() - y);
	}

	/* these constants seem fixed for all 6581s... */
	private static final double sid_caps = 470e-12;
	private static final float dac_kinkiness = 0.96f;

	protected static final double evaluate_filter(final FilterConfig filter, final int fc) {
		if (filter.getK() == 0f) {
			final double ik = SID.kinkedDac(fc, dac_kinkiness, 11);
			final double dynamic = filter.getMinimumfetresistance() + filter.getOffset() / Math.pow(filter.getSteepness(), ik);
			final double R = filter.getBaseresistance() * dynamic / (filter.getBaseresistance() + dynamic);
			return 1 / (2 * Math.PI * sid_caps * R);
		} else {
			return filter.getK() * fc + filter.getB();
		}
	}

	protected final double logscale(final double x) {
		if (x < 1) {
			return 0;
		}
		return Math.log(x) / Math.log(10);
	}

	public void propertyChange(final PropertyChangeEvent e) {
		if (!JSIDPlay2.PROP_FILTER.equals(e.getPropertyName())) {
			return;
		}

		final String filterName = (String) e.getNewValue();

		if ("".equals(filterName)) {
			filter = null;
			return;
		}

		filter = sidplay2.cp.m_iniCfg.getFilter(filterName);
		logScale = filter.getK() == 0;
		repaint();
	}
}