package applet.settings;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;
import resid_builder.FilterConfig;
import resid_builder.resid.ISIDDefs.ChipModel;
import resid_builder.resid.SID;
import applet.JSIDPlay2;
import applet.TuneTab;
import applet.ui.DKnob2;

@SuppressWarnings("serial")
public class SettingsPanel extends TuneTab implements ChangeListener {

	private DKnob2 leftVolume;
	private DKnob2 rightVolume;
	private DKnob2 filterResonance;

	// distortion properties
	private DKnob2 distortionRate;
	private DKnob2 distortionNonLinearity;

	float fGainMP3;

	protected JSIDPlay2 fSidplay2;
	protected JComboBox fFilter;

	protected JCheckBox fForceStereoTune;

	protected JTextField fBaseAddress;

	// private JComboBox fAudioMode;
	private ActionListener fFilterListener;
	private JComboBox sid1Model;
	private JComboBox sid2Model;
	private JCheckBox fBoosted8580;
	
	public SettingsPanel(final JSIDPlay2 sidplay) {
		this.fSidplay2 = sidplay;
		sidplay.support.addPropertyChangeListener(this);
	}

	@Override
	public void createContents() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		final JPanel volumeSpecial = new JPanel();
		volumeSpecial.setLayout(new GridLayout(1, 3));
		add(volumeSpecial);

		final JPanel volumeOptions = new JPanel();
		volumeOptions.setBorder(new TitledBorder("Volume Control"));
		volumeOptions.setLayout(new GridLayout(1, 2));
		volumeSpecial.add(volumeOptions);

		leftVolume = new DKnob2("Main SID level", "dB");
		leftVolume.setInterval(-60, 60);
		leftVolume.setDivisor(10);
		leftVolume.addChangeListener(this);
		volumeOptions.add(leftVolume);

		rightVolume = new DKnob2("Second SID level", "dB");
		rightVolume.setInterval(-60, 60);
		rightVolume.setDivisor(10);
		rightVolume.addChangeListener(this);
		volumeOptions.add(rightVolume);

		final JPanel controlOptions = new JPanel();
		controlOptions.setBorder(new TitledBorder("SID Control"));
		controlOptions.setLayout(new FlowLayout(FlowLayout.LEFT)); // FIXME looks like shit
		add(controlOptions);

		controlOptions.add(new JLabel("1st SID:"));

		sid1Model = new JComboBox(new Object[] { "Auto", ChipModel.MOS6581, ChipModel.MOS8580 });
		if (fSidplay2.getLockModel() != null) {
			sid1Model.setSelectedItem(fSidplay2.getLockModel());
		} else {
			sid1Model.setSelectedIndex(0);
		}
		sid1Model.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (sid1Model.getSelectedIndex() == 0) {
					fSidplay2.setLockModel(null);
					ChipModel chipModel = fSidplay2.cp.determineSIDModel();
					fSidplay2.setChipModel(chipModel);
				} else {
					ChipModel m = (ChipModel) sid1Model.getSelectedItem();
					fSidplay2.setLockModel(m);
					fSidplay2.setChipModel(m);
				}
			}
		});
		controlOptions.add(sid1Model);

		controlOptions.add(new JLabel("2nd SID:"));
		
		sid2Model = new JComboBox(new Object[] { "1st SID", ChipModel.MOS6581, ChipModel.MOS8580 });
		if (fSidplay2.getForceStereoTune()) {
			ChipModel stereoSidModel = fSidplay2.getStereoSidModel();
			if(stereoSidModel != null) {
				sid2Model.setSelectedItem(stereoSidModel);
			} else {
				// default: same as 1st SID model
				sid2Model.setSelectedIndex(sid1Model.getSelectedIndex());
			}
		} else {
			sid2Model.setSelectedIndex(0);
		}
		sid2Model.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (sid2Model.getSelectedIndex() == 0) {
					fSidplay2.setStereoSidModel(null);
				} else {
					ChipModel stereoSidModel = (ChipModel) sid2Model.getSelectedItem();
					fSidplay2.setStereoSidModel(stereoSidModel);
				}
			}
		});
		controlOptions.add(sid2Model);
		
		controlOptions.add(new JLabel("2nd address:"));

		fBaseAddress = new JTextField(fSidplay2.getDualSidBase());
		fBaseAddress.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setDualSidBase(fBaseAddress.getText());
				fSidplay2.restartC64();
			}
		});
		controlOptions.add(fBaseAddress);

		fForceStereoTune = new JCheckBox("Force Stereo", fSidplay2.getForceStereoTune());
		fForceStereoTune.setSelected(fSidplay2.getForceStereoTune());
		controlOptions.add(fForceStereoTune);
		fForceStereoTune.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent arg0) {
				fSidplay2.setForceStereoSid(fForceStereoTune.isSelected());
			}
		});

		fBoosted8580 = new JCheckBox("Digiboosted", fSidplay2.getDigiBoost8580());
		controlOptions.add(fBoosted8580);
		fBoosted8580.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent arg0) {
				fSidplay2.setDigiBoost(fBoosted8580.isSelected());
			}
		});


		final JPanel filterOptions = new JPanel();
		filterOptions.setBorder(new TitledBorder("Filter"));
		filterOptions.setLayout(new BorderLayout());
		add(filterOptions);

		final JPanel filterControls = new JPanel();
		filterControls.setLayout(new FlowLayout(FlowLayout.LEFT));
		filterOptions.add(BorderLayout.NORTH, filterControls);

		filterControls.add(new JLabel("Filter definition:"));

		fFilter = new JComboBox();
		fFilterListener = new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				final String filterName = (String) fFilter.getSelectedItem();
				fSidplay2.setFilter(filterName);
			}

		};
		fFilter.addActionListener(fFilterListener);
		filterControls.add(fFilter);

		final JPanel filterSidIcon = new JPanel();
		filterSidIcon.setLayout(new GridBagLayout());
		filterOptions.add(BorderLayout.CENTER, filterSidIcon);
		
		final JPanel filterPanel = new JPanel();
		filterPanel.setBorder(new TitledBorder("Filter Curve"));
		filterPanel.setLayout(new BorderLayout());
		final GridBagConstraints c = new GridBagConstraints();
		c.weightx = 4;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		filterSidIcon.add(filterPanel, c);

		final CurveFit filterCurve = new CurveFit(fSidplay2, this);
		filterPanel.add(filterCurve, BorderLayout.CENTER);

		final JPanel sidIconPanel = new JPanel();
		sidIconPanel.setBorder(new TitledBorder("SID Chip"));
		sidIconPanel.setLayout(new BorderLayout());
		c.weightx = 1;
		filterSidIcon.add(sidIconPanel, c);

		final SIDIcon sidIcon = new SIDIcon(fSidplay2);
		sidIconPanel.add(sidIcon, BorderLayout.CENTER);

		final JPanel distortionOptions = new JPanel();
		distortionOptions.setBorder(new TitledBorder("Distortion"));
		distortionOptions.setLayout(new GridLayout(1, 2));
		add(distortionOptions);

		distortionRate = new DKnob2("Filter attenuation", "");
		distortionRate.setInterval(0, 100);
		distortionRate.setDivisor(1e2f);
		distortionRate.addChangeListener(this);
		distortionOptions.add(distortionRate);

		distortionNonLinearity = new DKnob2("Nonlinearity", "");
		distortionNonLinearity.setInterval(20, 40);
		distortionNonLinearity.setDivisor(1e-5f);
		distortionNonLinearity.addChangeListener(this);
		distortionOptions.add(distortionNonLinearity);

		filterResonance = new DKnob2("Resonance scaling", "");
		filterResonance.setInterval(50, 200);
		filterResonance.setDivisor(1e2f);
		filterResonance.addChangeListener(this);
		distortionOptions.add(filterResonance);

		final FilterConfig f6581 = fSidplay2.cp.m_iniCfg.getFilter(fSidplay2.cp.m_iniCfg.emulation().getFilter6581());
		distortionRate.setIntValue(Math.round(f6581.getAttenuation() * distortionRate.getDivisor()));
		distortionNonLinearity.setIntValue(Math.round(f6581.getNonlinearity() * distortionNonLinearity.getDivisor()));
		filterResonance.setIntValue(Math.round(f6581.getResonanceFactor() * filterResonance.getDivisor()));

	}

	@Override
	public void setTune(final Player sidplay2, final SidTune sidTuneMod) {
		/* find currently active SID model */
		SID sid = fSidplay2.getSID(0);
		if (sid != null) {
			final ChipModel model = sid.getChipModel();
			setDistortionEnabled(model == ChipModel.MOS6581);
			setFilter(model);
		}
		leftVolume.setIntValue((int) (factor2dB(sidplay2.getConfig().leftVolume) * leftVolume.getDivisor()));
		rightVolume.setIntValue((int) (factor2dB(sidplay2.getConfig().rightVolume) * rightVolume.getDivisor()));
		setDistortion();
		fSidplay2.setDigiBoost(fBoosted8580.isSelected());
	}

	private void setDistortion() {
		final float[] dist = fSidplay2.getDistortion();
		if (dist == null) {
			return;
		}
		distortionRate.setIntValue(Math.round(dist[0] * distortionRate.getDivisor()));
		distortionNonLinearity.setIntValue(Math.round(dist[1] * distortionNonLinearity.getDivisor()));
		filterResonance.setIntValue(Math.round(dist[2] * filterResonance.getDivisor()));
	}

	private void setFilter(final ChipModel model) {
		final boolean enable = fSidplay2.cp.m_iniCfg.emulation().isFilter();
		String item = null;
		if (enable) {
			if (model == ChipModel.MOS6581) {
				item = fSidplay2.cp.m_iniCfg.emulation().getFilter6581();
			} else {
				item = fSidplay2.cp.m_iniCfg.emulation().getFilter8580();
			}
		}
		final Object[] items = fSidplay2.cp.m_iniCfg.getFilterList(model);
		fFilter.removeActionListener(fFilterListener);
		fFilter.removeAllItems();
		fFilter.addItem("");
		for (final Object filterName : items) {
			fFilter.addItem(filterName);
		}
		fFilter.addActionListener(fFilterListener);
		if (enable) {
			fFilter.setSelectedItem(item);
		} else {
			fFilter.setSelectedIndex(0);
		}
	}

	private void setDistortionEnabled(final boolean antti) {
		distortionNonLinearity.setEnabled(antti);
		distortionRate.setEnabled(antti);
	}

	/**
	 * If some settings of the GUI changes, some settings of this panel must be set accordingly.
	 */
	public void propertyChange(final PropertyChangeEvent event) {
		if (JSIDPlay2.PROP_CHIP_MODEL.equals(event.getPropertyName())) {
			final boolean model8580 = (Boolean) event.getNewValue();
			setDistortionEnabled(!model8580);
			setFilter(model8580 ? ChipModel.MOS8580 : ChipModel.MOS6581);
		} else if (JSIDPlay2.PROP_FILTER.equals(event.getPropertyName())) {
			setDistortion();
		}
	}

	private float factor2dB(final float factor) {
		return (float) (Math.log(factor) / Math.log(10) * 20);
	}

	private float dB2Factor(final float dB) {
		return (float) Math.pow(10, dB / 20);
	}

	public void stateChanged(final ChangeEvent event) {
		final Object src = event.getSource();
		if (src == leftVolume) {
			fSidplay2.cp.m_engine.getConfig().leftVolume = dB2Factor(leftVolume.getIntValue() / leftVolume.getDivisor());
		} else if (src == rightVolume) {
			fSidplay2.cp.m_engine.getConfig().rightVolume = dB2Factor(rightVolume.getIntValue() / leftVolume.getDivisor());
		} else if (src == distortionRate || src == distortionNonLinearity || src == filterResonance) {
			filterChanged();
		}
	}

	private void filterChanged() {
		fSidplay2.setDistortion((String) fFilter.getSelectedItem(), (distortionRate.getIntValue() / distortionRate.getDivisor()), (Math.round(distortionNonLinearity.getIntValue() / distortionNonLinearity.getDivisor() * distortionNonLinearity.getDivisor()) / distortionNonLinearity.getDivisor()), (filterResonance.getIntValue() / filterResonance.getDivisor()));
	}
}
