package applet.siddump;

import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.io.File;
import java.util.ArrayList;

import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import libsidplay.sidtune.SidTune;
import libsidutils.IniReader;
import libsidutils.SIDDump;
import libsidutils.SIDDump.Player;
import sidplay.IniConfig;
import applet.JSIDPlay2;
import applet.TuneTab;
import applet.events.IMuteVoice;
import applet.events.ITuneStateChanged;

@SuppressWarnings("serial")
public class SIDDumpView extends TuneTab {

	protected SIDDumpModel fDumpTableModel;
	protected JSIDPlay2 fSidplay2;

	protected SIDDump siddump;

	protected File fLastDir;
	protected JTextField fFirstFrameTxt;
	protected JTextField fPatternSpacingTxt;
	protected JTextField fNoteSpacingTxt;
	protected JCheckBox fEnableTimeSecondsCbx;
	protected JCheckBox fLowResCbx;

	protected JTextField fSIDDumpLengthTxt;
	protected JCheckBox fEnableSIDDumpCbx;
	protected JTable fDumpTable;

	public JTable getTable() {
		return fDumpTable;
	}

	protected long fFirstFrame = 0;
	protected int fSeconds = 0;
	protected float fOldNoteFactor = 1.f;
	protected int fBaseFreq = 0;
	protected int fBaseNote = 0xb0;
	protected int fNoteSpacing = 0;
	protected int fPatternSpacing = 0;
	protected int fFontSize = 8;

	protected int fLoadAddress;
	protected int fInitAddress;
	protected int fPlayerAddress;
	protected int fSubTune;
	protected JComboBox fRegOrderCBX;
	protected JButton fReplayDump;
	protected JButton fReplaySelection;
	protected JCheckBox fVoice1Button;
	protected JCheckBox fVoice2Button;
	protected JCheckBox fVoice3Button;
	protected JTextField fCallsPerFrameAddress;

	public SIDDumpView(final JSIDPlay2 play2) {
		fSidplay2 = play2;
		fSidplay2.support.addPropertyChangeListener(this);
		siddump = new SIDDump();
		siddump.open();
	}

	@Override
	public void createContents() {
		Font deriveFont;
		GridBagConstraints c;

		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		final JLabel label = new JLabel(
		"Uncheck the \"Playback\" option on the \"Favorites\" tab!");
		label.setFont(label.getFont().deriveFont(Font.ITALIC));
		label.setLayout(new FlowLayout(FlowLayout.LEFT));
		label.setForeground(Color.BLUE);
		add(label);

		final JPanel sidDump = new JPanel();
		sidDump.setBorder(new TitledBorder("SIDDump"));
		sidDump.setLayout(new GridBagLayout());

		final JPanel buttonsPanel = new JPanel();
		buttonsPanel.setLayout(new BoxLayout(buttonsPanel, BoxLayout.X_AXIS));
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth =4;
		sidDump.add(buttonsPanel, c);

		final JButton loadDump = new JButton("Load Dump...", new ImageIcon(JSIDPlay2.class
				.getResource("icons/table_load.png")));
		loadDump.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				final JFileChooser fileDialog = new JFileChooser(fLastDir);
				final int rc = fileDialog.showOpenDialog(fSidplay2);
				if (rc == JFileChooser.APPROVE_OPTION
						&& fileDialog.getSelectedFile() != null) {
					fLastDir = fileDialog.getSelectedFile().getParentFile();
					final String name = fileDialog.getSelectedFile()
					.getAbsolutePath();
					fDumpTableModel.load(name);
					fNoteSpacingTxt.setText(String.valueOf(fDumpTableModel
							.getNoteSpacing()));
					fNoteSpacing = fDumpTableModel.getNoteSpacing();
					fPatternSpacingTxt.setText(String.valueOf(fDumpTableModel
							.getPatternSpacing()));
					fPatternSpacing = fDumpTableModel.getPatternSpacing();
					fFirstFrameTxt.setText(String.valueOf(fDumpTableModel
							.getFirstFrame()));
					fFirstFrame = fDumpTableModel.getFirstFrame();
					fLowResCbx.setSelected(fDumpTableModel.getLowRes());
					fLoadAddress = fDumpTableModel.getLoadAddress();
					fInitAddress = fDumpTableModel.getInitAddress();
					fPlayerAddress = fDumpTableModel.getPlayerAddress();
					fSubTune = fDumpTableModel.getCurrentSong();
					fEnableTimeSecondsCbx.setSelected(fDumpTableModel
							.getTimeInSeconds());
				}
			}

		});
		buttonsPanel.add(loadDump);

		final JButton saveDump = new JButton("Save Dump...", new ImageIcon(JSIDPlay2.class
				.getResource("icons/table_save.png")));
		saveDump.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				final JFileChooser fileDialog = new JFileChooser(fLastDir);
				final int rc = fileDialog.showSaveDialog(fSidplay2);
				if (rc == JFileChooser.APPROVE_OPTION
						&& fileDialog.getSelectedFile() != null) {
					fLastDir = fileDialog.getSelectedFile().getParentFile();
					final String name = fileDialog.getSelectedFile()
					.getAbsolutePath();
					fDumpTableModel.save(name);
				}
			}

		});
		buttonsPanel.add(saveDump);

		fReplayDump = new JButton("Replay All", new ImageIcon(
				JSIDPlay2.class.getResource("icons/table_replay.png")));
		fReplayDump.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fDumpTableModel.setAborted();
				final int from = 0;
				final int to = fDumpTable.getRowCount() - 1;
				replayTune(from, to);
			}

		});
		buttonsPanel.add(fReplayDump);

		fReplaySelection = new JButton("Replay Selected", new ImageIcon(
				JSIDPlay2.class.getResource("icons/table_replay.png")));
		fReplaySelection.setEnabled(false);
		fReplaySelection.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fDumpTableModel.setAborted();
				final int[] rows = fDumpTable.getSelectedRows();
				replayTune(rows[0], rows[rows.length - 1]);
			}

		});
		buttonsPanel.add(fReplaySelection);

		fEnableSIDDumpCbx = new JCheckBox("Start/Stop Recording");
		fEnableSIDDumpCbx.setToolTipText("On/off SID dump");
		deriveFont = fEnableSIDDumpCbx.getFont().deriveFont(Font.BOLD);
		fEnableSIDDumpCbx.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 1;
		c.anchor = GridBagConstraints.WEST;
		fEnableSIDDumpCbx.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				if (fEnableSIDDumpCbx.isSelected()) {
					fSidplay2.restartC64();
				} else {
					fDumpTableModel.stop();
					fSidplay2.reset(SIDDumpView.this);
				}
			}

		});
		sidDump.add(fEnableSIDDumpCbx, c);

		final JLabel note = new JLabel("Player will be restarted!");
		note.setForeground(Color.BLUE);
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 1;
		c.anchor = GridBagConstraints.WEST;
		sidDump.add(note, c);

		fEnableTimeSecondsCbx = new JCheckBox("Time in Seconds");
		fEnableTimeSecondsCbx
		.setToolTipText("Display time in seconds instead of frames");
		deriveFont = fEnableTimeSecondsCbx.getFont().deriveFont(Font.BOLD);
		fEnableTimeSecondsCbx.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 1;
		c.anchor = GridBagConstraints.WEST;
		sidDump.add(fEnableTimeSecondsCbx, c);

		fLowResCbx = new JCheckBox("Low-resolution mode");
		fLowResCbx
		.setToolTipText("Low-resolution mode (only display 1 row per note)");
		deriveFont = fLowResCbx.getFont().deriveFont(Font.BOLD);
		fLowResCbx.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 3;
		c.gridy = 1;
		c.anchor = GridBagConstraints.WEST;
		sidDump.add(fLowResCbx, c);

		final JPanel recordOptions = new JPanel();
		recordOptions.setBorder(new TitledBorder("Record Options"));
		recordOptions.setLayout(new GridBagLayout());
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 2;
		c.gridwidth=4;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.HORIZONTAL;
		sidDump.add(recordOptions, c);

		final JLabel firstFrameLbl = new JLabel("First frame:");
		deriveFont = firstFrameLbl.getFont().deriveFont(Font.BOLD);
		firstFrameLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(firstFrameLbl, c);

		fFirstFrameTxt = new JTextField();
		fFirstFrameTxt.setText(String.valueOf(fFirstFrame));
		fFirstFrameTxt.setToolTipText("First frame to display, default 0");
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.insets = new Insets(5, 5, 5, 5);
		c.fill = GridBagConstraints.BOTH;
		fFirstFrameTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = fFirstFrameTxt.getText();
				try {
					fFirstFrame = Long.parseLong(first);
					if (fFirstFrame >= 0) {
						fFirstFrameTxt
						.setToolTipText("First frame to display, default 0");
						fFirstFrameTxt.setBackground(Color.white);
					} else {
						fFirstFrameTxt
						.setToolTipText("First frame must be integer starting at 0!");
						fFirstFrameTxt.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					fFirstFrameTxt
					.setToolTipText("First frame must be integer starting at 0!");
					fFirstFrameTxt.setBackground(Color.RED);
				}
			}

		});
		recordOptions.add(fFirstFrameTxt, c);

		final JLabel noteSpacingLbl = new JLabel("Note Spacing:");
		deriveFont = noteSpacingLbl.getFont().deriveFont(Font.BOLD);
		noteSpacingLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(noteSpacingLbl, c);

		fNoteSpacingTxt = new JTextField();
		fNoteSpacingTxt.setText(String.valueOf(fNoteSpacing));
		fNoteSpacingTxt.setToolTipText("Note spacing, default 0 (none)");
		c = new GridBagConstraints();
		c.gridx = 3;
		c.gridy = 0;
		c.weightx = 1;
		c.weighty = 1;
		c.insets = new Insets(5, 5, 5, 5);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		fNoteSpacingTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = fNoteSpacingTxt.getText();
				try {
					fNoteSpacing = Integer.parseInt(first);
					if (fNoteSpacing >= 0) {
						fNoteSpacingTxt
						.setToolTipText("Note spacing, default 0 (none)");
						fNoteSpacingTxt.setBackground(Color.white);
					} else {
						fNoteSpacingTxt
						.setToolTipText("Note spacing must be integer starting at 0!");
						fNoteSpacingTxt.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					fNoteSpacingTxt
					.setToolTipText("Note spacing must be integer starting at 0!");
					fNoteSpacingTxt.setBackground(Color.RED);
				}
			}

		});
		recordOptions.add(fNoteSpacingTxt, c);

		final JLabel dumpLbl = new JLabel("Max. record length (mm:ss):");
		deriveFont = dumpLbl.getFont().deriveFont(Font.BOLD);
		dumpLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(dumpLbl, c);

		fSIDDumpLengthTxt = new JTextField();
		fSIDDumpLengthTxt.setText(String.format("%02d:%02d",
				(fSeconds / 60 % 100), (fSeconds % 60)));
		fSIDDumpLengthTxt
		.setToolTipText("Maximum record length (00:00 - autodetection)");
		fSIDDumpLengthTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String time = fSIDDumpLengthTxt.getText();
				fSeconds = IniReader.parseTime(time);
				if (fSeconds != -1) {
					fSIDDumpLengthTxt
					.setToolTipText("Maximum record length (00:00 - autodetection)");
					fSIDDumpLengthTxt.setBackground(Color.white);
				} else {
					fSIDDumpLengthTxt
					.setToolTipText("The time format is mm:ss");
					fSIDDumpLengthTxt.setBackground(Color.RED);
				}
			}

		});
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.insets = new Insets(5, 5, 5, 5);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(fSIDDumpLengthTxt, c);

		final JLabel patternSpacingLbl = new JLabel("Pattern Spacing:");
		deriveFont = patternSpacingLbl.getFont().deriveFont(Font.BOLD);
		patternSpacingLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(patternSpacingLbl, c);

		fPatternSpacingTxt = new JTextField();
		fPatternSpacingTxt.setText(String.valueOf(fPatternSpacing));
		fPatternSpacingTxt.setToolTipText("Pattern spacing, default 0 (none)");
		c = new GridBagConstraints();
		c.gridx = 3;
		c.gridy = 1;
		c.weightx = 1;
		c.weighty = 1;
		c.insets = new Insets(5, 5, 5, 5);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		fPatternSpacingTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = fPatternSpacingTxt.getText();
				try {
					fPatternSpacing = Integer.parseInt(first);
					if (fPatternSpacing >= 0) {
						fPatternSpacingTxt
						.setToolTipText("Pattern spacing, default 0 (none)");
						fPatternSpacingTxt.setBackground(Color.white);
					} else {
						fPatternSpacingTxt
						.setToolTipText("Pattern spacing must be integer starting at 0!");
						fPatternSpacingTxt.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					fPatternSpacingTxt
					.setToolTipText("Pattern spacing must be integer starting at 0!");
					fPatternSpacingTxt.setBackground(Color.RED);
				}
			}

		});
		recordOptions.add(fPatternSpacingTxt, c);

		final JLabel oldNoteFactorLbl = new JLabel("Old note factor:");
		deriveFont = oldNoteFactorLbl.getFont().deriveFont(Font.BOLD);
		oldNoteFactorLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(oldNoteFactorLbl, c);

		final JTextField fOldNoteFactorTxt = new JTextField();
		fOldNoteFactorTxt.setText(String.valueOf(fOldNoteFactor));
		fOldNoteFactorTxt
		.setToolTipText("\"Oldnote-sticky\" factor. Default 1, increase for better vibrato display "
				+ "(when increased, requires well calibrated frequencies)");
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.insets = new Insets(5, 5, 5, 5);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		fOldNoteFactorTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = fOldNoteFactorTxt.getText();
				try {
					fOldNoteFactor = Float.parseFloat(first);
					if (fOldNoteFactor >= 1) {
						fOldNoteFactorTxt
						.setToolTipText("\"Oldnote-sticky\" factor. Default 1, increase for better vibrato display "
								+ "(when increased, requires well calibrated frequencies)");
						fOldNoteFactorTxt.setBackground(Color.white);
					} else {
						fOldNoteFactorTxt
						.setToolTipText("Old note factor must be a float starting at 1!");
						fOldNoteFactorTxt.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					fOldNoteFactorTxt
					.setToolTipText("Old note factor must be a float starting at 1!");
					fOldNoteFactorTxt.setBackground(Color.RED);
				}
			}

		});
		recordOptions.add(fOldNoteFactorTxt, c);

		final JLabel baseFreqLbl = new JLabel("Base Freq:");
		deriveFont = baseFreqLbl.getFont().deriveFont(Font.BOLD);
		baseFreqLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 3;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(baseFreqLbl, c);

		final JTextField baseFreqTxt = new JTextField();
		baseFreqTxt.setText(String.format("0x%02x", fBaseFreq));
		baseFreqTxt
		.setToolTipText("Frequency recalibration. Give note frequency");
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 3;
		c.weightx = 1;
		c.weighty = 1;
		c.insets = new Insets(5, 5, 5, 5);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		baseFreqTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = baseFreqTxt.getText();
				try {
					fBaseFreq = IniConfig.toInt2(first, 2);
					if (fBaseFreq != -1 && fBaseFreq >= 0) {
						baseFreqTxt
						.setToolTipText("Frequency recalibration. Give note frequency");
						baseFreqTxt.setBackground(Color.white);
					} else {
						baseFreqTxt
						.setToolTipText("Base frequence must be a hex value starting at 0x00!");
						baseFreqTxt.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					baseFreqTxt
					.setToolTipText("Base frequence must be a hex value starting at 0x00!");
					baseFreqTxt.setBackground(Color.RED);
				}
			}

		});
		recordOptions.add(baseFreqTxt, c);

		final JLabel fontSizeLbl = new JLabel("Table Font Size:");
		deriveFont = fontSizeLbl.getFont().deriveFont(Font.BOLD);
		fontSizeLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(fontSizeLbl, c);

		final JTextField fontSizeTxt = new JTextField();
		fontSizeTxt.setText("" + fFontSize);
		fontSizeTxt.setToolTipText("Table font size");
		c = new GridBagConstraints();
		c.gridx = 3;
		c.gridy = 2;
		c.weightx = 1;
		c.weighty = 1;
		c.insets = new Insets(5, 5, 5, 5);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		fontSizeTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = fontSizeTxt.getText();
				try {
					fFontSize = Integer.parseInt(first);
					if (fFontSize > 0) {
						fontSizeTxt.setToolTipText("Table font size");
						fontSizeTxt.setBackground(Color.white);
						fDumpTable.setFont(new Font("Courier", fDumpTable
								.getFont().getStyle(), fFontSize));
						fDumpTable.getTableHeader().setFont(
								new Font("Courier", fDumpTable.getTableHeader()
										.getFont().getStyle(), fFontSize));
					} else {
						fontSizeTxt
						.setToolTipText("Table font size must be an integer bigger than 0!");
						fontSizeTxt.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					fontSizeTxt
					.setToolTipText("Table font size must be an integer bigger than 0!");
					fontSizeTxt.setBackground(Color.RED);
				}
			}

		});
		recordOptions.add(fontSizeTxt, c);

		final JLabel baseNoteLbl = new JLabel("Base note:");
		deriveFont = baseNoteLbl.getFont().deriveFont(Font.BOLD);
		baseNoteLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 3;
		c.weightx = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		recordOptions.add(baseNoteLbl, c);

		final JTextField fBaseNoteTxt = new JTextField();
		fBaseNoteTxt.setText(String.format("0x%02x", fBaseNote));
		fBaseNoteTxt
		.setToolTipText("Select calibration note (abs.notation 0x80-0xdf). Default middle-C (0xb0)");
		c = new GridBagConstraints();
		c.gridx = 3;
		c.gridy = 3;
		c.weightx = 1;
		c.weighty = 1;
		c.insets = new Insets(5, 5, 5, 5);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.BOTH;
		fBaseNoteTxt.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = fBaseNoteTxt.getText();
				try {
					fBaseNote = IniConfig.toInt2(first, 2);
					if (fBaseNote != -1 && fBaseNote >= 128 && fBaseNote <= 223) {
						fBaseNoteTxt
						.setToolTipText("Select calibration note (abs.notation 0x80-0xdf). Default middle-C (0xb0)");
						fBaseNoteTxt.setBackground(Color.white);
					} else {
						fBaseNoteTxt
						.setToolTipText("Base note must be a hex value 0x80 <= X <= 0xdf!");
						fBaseNoteTxt.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					fBaseNoteTxt
					.setToolTipText("Base note must be a hex value 0x80 <= X <= 0xdf!");
					fBaseNoteTxt.setBackground(Color.RED);
				}
			}

		});
		recordOptions.add(fBaseNoteTxt, c);

		final JPanel replayOptions = new JPanel();
		replayOptions.setBorder(new TitledBorder("Replay Options"));
		replayOptions.setLayout(new BoxLayout(replayOptions, BoxLayout.X_AXIS));
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 3;
		c.gridwidth=GridBagConstraints.REMAINDER;
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.HORIZONTAL;
		sidDump.add(replayOptions, c);

		final JPanel options = new JPanel();
		final GridBagLayout mgr = new GridBagLayout();

		options.setLayout(mgr);
		replayOptions.add(options);

		final JLabel outputLbl = new JLabel("Emulate Player:");
		deriveFont = outputLbl.getFont().deriveFont(Font.BOLD);
		outputLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.insets = new Insets(2, 2, 2, 2);
		options.add(outputLbl, c);

		fRegOrderCBX = new JComboBox();
		final ArrayList<Player> players = siddump.getPlayers();
		for (final Player player : players) {
			fRegOrderCBX.addItem(player.getName());
		}
		fRegOrderCBX.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				if (fRegOrderCBX.getSelectedIndex() == -1 ||
						fRegOrderCBX.getSelectedIndex() >= players.size()) {
					return;
				}
				fDumpTableModel.setRegOrder(players.get(fRegOrderCBX.getSelectedIndex()).getBytes());
			}

		});
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 0;
		c.insets = new Insets(2, 2, 2, 2);
		options.add(fRegOrderCBX, c);

		final JLabel callsLbl = new JLabel("Calls per Frame:");
		deriveFont = callsLbl.getFont().deriveFont(Font.BOLD);
		callsLbl.setFont(deriveFont);
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 0;
		c.insets = new Insets(2, 2, 2, 2);
		options.add(callsLbl, c);

		fCallsPerFrameAddress = new JTextField("1", 3);
		fCallsPerFrameAddress
		.setToolTipText("Calls per frame, default 1 (50 Hz)");
		fCallsPerFrameAddress.addKeyListener(new KeyAdapter() {

			@Override
			public void keyReleased(final KeyEvent e) {
				final String first = fCallsPerFrameAddress.getText();
				try {
					final int speed = Integer.parseInt(first);
					if (speed >= 1) {
						fCallsPerFrameAddress
						.setToolTipText("Calls per frame, default 1 (50 Hz)");
						fCallsPerFrameAddress.setBackground(Color.white);
						fDumpTableModel.setReplayFrequency(speed*50);
					} else {
						fCallsPerFrameAddress
						.setToolTipText("Calls per frame must be integer starting at 1!");
						fCallsPerFrameAddress.setBackground(Color.RED);
					}
				} catch (final NumberFormatException e2) {
					fCallsPerFrameAddress
					.setToolTipText("Calls per frame must be integer starting at 1!");
					fCallsPerFrameAddress.setBackground(Color.RED);
				}

			}

		});
		c = new GridBagConstraints();
		c.gridx = 3;
		c.gridy = 0;
		c.insets = new Insets(2, 2, 2, 2);
		options.add(fCallsPerFrameAddress, c);

		fVoice1Button = new JCheckBox("Mute 1");
		fVoice1Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setMute(0, 0, fVoice1Button.isSelected());
			}

		});
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 2;
		c.insets = new Insets(2, 2, 2, 2);
		options.add(fVoice1Button, c);
		fVoice2Button = new JCheckBox("Mute 2");
		fVoice2Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setMute(0, 1, fVoice2Button.isSelected());
			}

		});
		c = new GridBagConstraints();
		c.gridx = 1;
		c.gridy = 2;
		c.insets = new Insets(2, 2, 2, 2);
		options.add(fVoice2Button, c);
		fVoice3Button = new JCheckBox("Mute 3");
		fVoice3Button.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setMute(0, 2, fVoice3Button.isSelected());
			}

		});
		c = new GridBagConstraints();
		c.gridx = 2;
		c.gridy = 2;
		c.insets = new Insets(2, 2, 2, 2);
		options.add(fVoice3Button, c);

		final JLabel label2 = new JLabel(
		"Press Stop and Replay, if changing the player!");
		label2.setFont(label2.getFont().deriveFont(Font.ITALIC));
		label2.setLayout(new FlowLayout(FlowLayout.LEFT));
		label2.setForeground(Color.BLUE);
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 5;
		c.gridwidth = 4;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		add(label2, c);

		fDumpTableModel = new SIDDumpModel();
		fDumpTable = new SIDDumpTable(fDumpTableModel);
		fDumpTable.getSelectionModel().addListSelectionListener(new ListSelectionListener() {

			public void valueChanged(final ListSelectionEvent e) {
				fReplaySelection.setEnabled(fDumpTable.getSelectedRowCount()>0);
			}

		});
		c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 6;
		c.gridwidth = 4;
		c.weightx = 1;
		c.weighty = 1;
		c.fill = GridBagConstraints.BOTH;
		sidDump.add(new JScrollPane(fDumpTable), c);

		add(sidDump);
		if (players.size()>0) {
			// initialize with first item
			fDumpTableModel.setRegOrder(players.get(0).getBytes());
		}
	}

	@Override
	public void setTune(final libsidplay.Player m_engine, final SidTune m_tune) {
		fDumpTableModel.setPlayer(m_engine);
		if (m_tune == null) {
			return;
		}

		fLoadAddress = m_tune.getInfo().loadAddr;
		fDumpTableModel.setLoadAddress(fLoadAddress);
		fInitAddress = m_tune.getInfo().initAddr;
		fDumpTableModel.setInitAddress(fInitAddress);
		fPlayerAddress = m_tune.getInfo().playAddr;
		fDumpTableModel.setPayerAddress(fPlayerAddress);
		fSubTune = m_tune.getInfo().currentSong;
		fDumpTableModel.setCurrentSong(fSubTune);
		fDumpTableModel.setFirstFrame(fFirstFrame);
		if (fSeconds == 0) {
			int songLength = fSidplay2.getSongLength(m_tune);
			if (songLength <= 0) {
				songLength = fSidplay2.getDefaultPlayLength();
				if (songLength == 0) {
					// default
					songLength = 60;
				}
			}
			fSIDDumpLengthTxt.setText(String.format("%02d:%02d",
					(songLength / 60 % 100), (songLength % 60)));
			fDumpTableModel.setRecordLength(songLength);
		} else {
			fDumpTableModel.setRecordLength(fSeconds);
		}
		fDumpTableModel.setTimeInSeconds(fEnableTimeSecondsCbx
				.isSelected());
		fDumpTableModel.setOldNoteFactor(fOldNoteFactor);
		fDumpTableModel.setBaseFreq(fBaseFreq);
		fDumpTableModel.setBaseNote(fBaseNote);
		fDumpTableModel.setPatternSpacing(fPatternSpacing);
		fDumpTableModel.setNoteSpacing(fNoteSpacing);
		fDumpTableModel.setLowRes(fLowResCbx.isSelected());

		fEnableSIDDumpCbx.setEnabled(m_tune.getInfo().playAddr != 0);
		if (m_tune.getInfo().playAddr == 0) {
			fEnableSIDDumpCbx.setSelected(false);
			fEnableSIDDumpCbx
			.setToolTipText("Player address is 0, SIDDump not available!");
		} else {
			fEnableSIDDumpCbx.setToolTipText(null);
		}

		fDumpTableModel.init();
		final boolean enable = fEnableSIDDumpCbx.isSelected();
		if (enable) {
			m_engine.getC64().setSidWriteListener(0, fDumpTableModel);
			m_engine.getC64().setPlayRoutineObserver(fDumpTableModel);
		} else {
			m_engine.getC64().setSidWriteListener(0, null);
			m_engine.getC64().setPlayRoutineObserver(null);
		}

		setVoiceMute(0, 0, fVoice1Button.isSelected());
		setVoiceMute(0, 1, fVoice2Button.isSelected());
		setVoiceMute(0, 2, fVoice3Button.isSelected());

	}

	private void setVoiceMute(final int sidNum, final int voiceNo, final boolean mute) {
		fSidplay2.setMute(sidNum, voiceNo, mute);
	}

	public void propertyChange(final PropertyChangeEvent evt) {
		if (IMuteVoice.class.getName().equals(evt.getPropertyName())) {
			final IMuteVoice muteVoice = (IMuteVoice) evt.getNewValue();
			final int sidNum = muteVoice.getSidNum();
			final int voice = muteVoice.getVoice();
			if (sidNum==0) {
				if (voice == 0 && !evt.getSource().equals(fVoice1Button)) {
					fVoice1Button.setSelected(muteVoice.isMute());
				} else if (voice == 1 && !evt.getSource().equals(fVoice2Button)) {
					fVoice2Button.setSelected(muteVoice.isMute());
				} else if (voice == 2 && !evt.getSource().equals(fVoice3Button)) {
					fVoice3Button.setSelected(muteVoice.isMute());
				}
			}
		} else if (ITuneStateChanged.class.getName().equals(
				evt.getPropertyName())) {
			fReplaySelection.setEnabled(fDumpTable.getSelectedRowCount()>0);
			fReplayDump.setEnabled(true);
			fDumpTableModel.stop();
		}
	}

	protected void replayTune(final int from, final int to) {
		fDumpTableModel.setReplayRange(from, to);

		new Thread(new Runnable() {

			public void run() {
				try {
					fReplaySelection.setEnabled(false);
					fReplayDump.setEnabled(false);
					fDumpTableModel.replay();
				} catch (final Exception e) {
					e.printStackTrace();
				} finally {
					fReplaySelection.setEnabled(fDumpTable
							.getSelectedRowCount() > 0);
					fReplayDump.setEnabled(true);
				}
			}

		}).start();
	}

}
