package applet.sidreg;

import java.util.Vector;

import javax.swing.table.DefaultTableModel;

import libsidplay.common.IReSIDExtension;


@SuppressWarnings("serial")
public class SIDRegModel extends DefaultTableModel implements IReSIDExtension {
	public static final int ABS_CYCLES = 0;
	public static final int COL_CYCLES = 1;
	public static final int COL_DESCRIPTION = 2;
	public static final int COL_CHIP_NUM = 3;
	public static final int COL_VALUE = 4;

	private static final int MAX_ROW_COUNT = 10000;
	private static final int UPDATE_ON_ROW = 2000;

	/**
	 * Recorded row number
	 */
	int fFetchedRow;

	private final String description[] = new String[] {
			"$D400  Voice 1: Frequency Control - Low-Byte",
			"$D401  Voice 1: Frequency Control - High-Byte",
			"$D402  Voice 1: Pulse Waveform Width - Low-Byte",
			"$D403  Voice 1: Pulse Waveform Width - High-Nibble",
			"$D404  Voice 1: Control Register",
			"$D405  Voice 1: Attack / Decay Cycle Control",
			"$D406  Voice 1: Sustain / Release Cycle Control",
			"$D407  Voice 2: Frequency Control - Low-Byte",
			"$D408  Voice 2: Frequency Control - High-Byte",
			"$D409  Voice 2: Pulse Waveform Width - Low-Byte",
			"$D40A  Voice 2: Pulse Waveform Width - High-Nibble",
			"$D40B  Voice 2: Control Register",
			"$D40C  Voice 2: Attack / Decay Cycle Control",
			"$D40D  Voice 2: Sustain / Release Cycle Control",
			"$D40E  Voice 3: Frequency Control - Low-Byte",
			"$D40F  Voice 3: Frequency Control - High-Byte",
			"$D410  Voice 3: Pulse Waveform Width - Low-Byte",
			"$D411  Voice 3: Pulse Waveform Width - High-Nibble",
			"$D412  Voice 3: Control Register",
			"$D413  Voice 3: Attack / Decay Cycle Control",
			"$D414  Voice 3: Sustain / Release Cycle Control",
			"$D415  Filter Cutoff Frequency: Low-Nibble",
			"$D416  Filter Cutoff Frequency: High-Byte",
			"$D417  Filter Resonance Control / Voice Input Control",
			"$D418  Select Filter Mode and Volume",
			"$D419  Analog/Digital Converter: Game Paddle 1",
			"$D41A  Analog/Digital Converter: Game Paddle 2",
			"$D41B  Oscillator 3 Output", "$D41C  Envelope Generator 3 Output" };
	private long fTime;

	public SIDRegModel() {
	}

	@Override
	public int getColumnCount() {
		// time/description[num]/value
		return 5;
	}

	@Override
	public int getRowCount() {
		return dataVector.size();
	}

	@Override
	public Class<String> getColumnClass(final int columnIndex) {
		return String.class;
	}

	@Override
	public String getColumnName(final int column) {
		switch (column) {
		case ABS_CYCLES:

			return "Absolute Time (Cycles)";

		case COL_CYCLES:

			return "Relative Time (Cycles)";

		case COL_CHIP_NUM:

			return "Chip No.";

		case COL_DESCRIPTION:

			return "Description";

		case COL_VALUE:

			return "Value";

		default:
			return "";
		}
	}

	@Override
	public boolean isCellEditable(final int row, final int column) {
		return false;
	}

	public void write(final long time, final int chipNum, final int addr,
			final byte data) {
		if (fTime == 0) {
			fTime = time;
		}
		final long relTime = time - fTime;
		final Vector<String> row = new Vector<String>();
		row.setSize(getColumnCount());
		row.add(ABS_CYCLES, String.valueOf(time));
		row.add(COL_CYCLES, String.valueOf(relTime));
		row.add(COL_DESCRIPTION, addr < description.length ? description[addr] : "?");
		row.add(COL_CHIP_NUM, String.valueOf(chipNum));
		row.add(COL_VALUE, String.format("$%02X", data & 0xff));

		putInRow(row);

		fTime = time;
	}

	/**
	 * Put the row into the table
	 * 
	 * @param output
	 *            the row to add
	 */
	@SuppressWarnings("unchecked")
	private void putInRow(final Vector<String> output) {
		if (fFetchedRow == MAX_ROW_COUNT) {
			return;
		}
		dataVector.insertElementAt(output, fFetchedRow++);
		if (fFetchedRow % UPDATE_ON_ROW == 0) {
			fireTableDataChanged();
		}
	}

	public void init() {
		clearTable();
		fTime = 0;
		fFetchedRow = 0;
	}

	/**
	 * Clear recorded rows
	 */
	private void clearTable() {
		dataVector.removeAllElements();
		fireTableDataChanged();
	}

	/**
	 * Stop recording
	 */
	public void stop() {
		// set total recorded frames update table
		fireTableDataChanged();
	}

}
