package applet.sidreg;

import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.RowFilter;
import javax.swing.border.TitledBorder;
import javax.swing.table.TableRowSorter;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;
import applet.JSIDPlay2;
import applet.TuneTab;
import applet.events.ITuneStateChanged;

@SuppressWarnings("serial")
public class SIDRegView extends TuneTab {

	private TableRowSorter<SIDRegModel> sorter;
	protected JCheckBox fEnableRecording;
	protected SIDRegModel fRegModel;
	protected JPanel fFilterPanel;

	private final JSIDPlay2 sidplay2;

	public SIDRegView(final JSIDPlay2 sidplay2) {
		this.sidplay2 = sidplay2;
		sidplay2.support.addPropertyChangeListener(this);
	}

	/**
	 * Create GUI elements here
	 */
	@Override
	public void createContents() {
		Font deriveFont;
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		final JLabel label = new JLabel(
		"Uncheck the \"Playback\" option on the \"Favorites\" tab!");
		label.setFont(label.getFont().deriveFont(Font.ITALIC));
		label.setLayout(new FlowLayout(FlowLayout.LEFT));
		label.setForeground(Color.BLUE);
		add(label);

		final JPanel sidReg = new JPanel();
		sidReg.setBorder(new TitledBorder("SID Registers"));
		sidReg.setLayout(new BoxLayout(sidReg, BoxLayout.Y_AXIS));
		add(sidReg);

		fEnableRecording = new JCheckBox("Start/Stop Recording");
		fEnableRecording.setToolTipText("On/off SID register dump");
		deriveFont = fEnableRecording.getFont().deriveFont(Font.BOLD);
		fEnableRecording.setFont(deriveFont);
		fEnableRecording.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				toggleDataCollection(fEnableRecording.isSelected());
			}

		});
		sidReg.add(fEnableRecording);

		fFilterPanel = new JPanel();
		fFilterPanel.setBorder(new TitledBorder("Viewer Filter"));
		fFilterPanel.setLayout(new GridLayout(5, 5));
		sidReg.add(fFilterPanel);

		for (int i = 1; i < 4; i++) {
			final JCheckBox fFreqButton = new JCheckBox("Freq " + i);
			fFilterPanel.add(fFreqButton);
			fFreqButton.addActionListener(new ActionListener() {

				public void actionPerformed(final ActionEvent e) {
					updateFilter();
				}

			});
			fFreqButton.setSelected(true);

			final JCheckBox fPulseButton = new JCheckBox("Pulse " + i);
			fFilterPanel.add(fPulseButton);
			fPulseButton.addActionListener(new ActionListener() {

				public void actionPerformed(final ActionEvent e) {
					updateFilter();
				}

			});
			fPulseButton.setSelected(true);

			final JCheckBox fControlButton = new JCheckBox("Ctrl " + i);
			fFilterPanel.add(fControlButton);
			fControlButton.addActionListener(new ActionListener() {

				public void actionPerformed(final ActionEvent e) {
					updateFilter();
				}

			});
			fControlButton.setSelected(true);

			final JCheckBox fADButton = new JCheckBox("A/D " + i);
			fFilterPanel.add(fADButton);
			fADButton.addActionListener(new ActionListener() {

				public void actionPerformed(final ActionEvent e) {
					updateFilter();
				}

			});
			fADButton.setSelected(true);

			final JCheckBox fSRButton = new JCheckBox("S/R " + i);
			fFilterPanel.add(fSRButton);
			fSRButton.addActionListener(new ActionListener() {

				public void actionPerformed(final ActionEvent e) {
					updateFilter();
				}

			});
			fSRButton.setSelected(true);
		}
		final JCheckBox fFilterButton = new JCheckBox("Filter");
		fFilterPanel.add(fFilterButton);
		fFilterButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				updateFilter();
			}

		});
		fFilterButton.setSelected(true);

		final JCheckBox fVolumeButton = new JCheckBox("Vol");
		fFilterPanel.add(fVolumeButton);
		fVolumeButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				updateFilter();
			}

		});
		fVolumeButton.setSelected(true);

		final JCheckBox fADConvButton = new JCheckBox("Paddles");
		fFilterPanel.add(fADConvButton);
		fADConvButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				updateFilter();
			}

		});
		fADConvButton.setSelected(true);

		final JCheckBox fOscButton = new JCheckBox("Osc. 3");
		fFilterPanel.add(fOscButton);
		fOscButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				updateFilter();
			}

		});
		fOscButton.setSelected(true);

		final JCheckBox fEnvButton = new JCheckBox("Env. 3");
		fFilterPanel.add(fEnvButton);
		fEnvButton.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				updateFilter();
			}

		});
		fEnvButton.setSelected(true);

		final JButton selectAll = new JButton("Select All");
		selectAll.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_selectall.png")));
		fFilterPanel.add(selectAll);
		selectAll.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				for (int i = 0; i < fFilterPanel.getComponentCount(); i++) {
					if (!(fFilterPanel.getComponent(i) instanceof JCheckBox)) {
						continue;
					}
					final JCheckBox box = (JCheckBox) fFilterPanel.getComponent(i);
					box.setSelected(true);
				}
				updateFilter();
			}

		});

		final JButton deselectAll = new JButton("Deselect All");
		deselectAll.setIcon(new ImageIcon(JSIDPlay2.class
				.getResource("icons/fav_deselectall.png")));
		fFilterPanel.add(deselectAll);
		deselectAll.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				for (int i = 0; i < fFilterPanel.getComponentCount(); i++) {
					if (!(fFilterPanel.getComponent(i) instanceof JCheckBox)) {
						continue;
					}
					final JCheckBox box = (JCheckBox) fFilterPanel.getComponent(i);
					box.setSelected(false);
				}
				updateFilter();
			}

		});

		fRegModel = new SIDRegModel();
		final SIDRegTable sidRegTable = new SIDRegTable(fRegModel);
		sorter = new TableRowSorter<SIDRegModel>(fRegModel);
		sidRegTable.setRowSorter(sorter);
		sidReg.add(new JScrollPane(sidRegTable));
	}

	protected void updateFilter() {

		final List<RowFilter<Object, Object>> filters = new ArrayList<RowFilter<Object, Object>>(
				fFilterPanel.getComponentCount());
		for (int i = 0; i < fFilterPanel.getComponentCount(); i++) {
			if (!(fFilterPanel.getComponent(i) instanceof JCheckBox)) {
				continue;
			}
			final JCheckBox box = (JCheckBox) fFilterPanel.getComponent(i);
			RowFilter<Object, Object> regexFilter;
			int voice;
			switch (i) {
			case 0:
			case 5:
			case 10:
				voice = i / 5 + 1;
				regexFilter = RowFilter.regexFilter("Voice " + voice
						+ ": Frequency Control", 2);
				break;

			case 1:
			case 6:
			case 11:
				voice = (i - 1) / 5 + 1;
				regexFilter = RowFilter.regexFilter("Voice " + voice
						+ ": Pulse Waveform Width", 2);
				break;

			case 2:
			case 7:
			case 12:
				voice = (i - 2) / 5 + 1;
				regexFilter = RowFilter.regexFilter("Voice " + voice
						+ ": Control Register", 2);
				break;

			case 3:
			case 8:
			case 13:
				voice = (i - 3) / 5 + 1;
				regexFilter = RowFilter.regexFilter("Voice " + voice
						+ ": Attack / Decay", 2);
				break;

			case 4:
			case 9:
			case 14:
				voice = (i - 4) / 5 + 1;
				regexFilter = RowFilter.regexFilter("Voice " + voice
						+ ": Sustain / Release", 2);
				break;

			case 15:
				regexFilter = RowFilter.regexFilter("  Filter", 2);
				break;

			case 16:
				regexFilter = RowFilter.regexFilter("Select Filter Mode", 2);
				break;

			case 17:
				regexFilter = RowFilter.regexFilter("Analog/Digital Converter", 2);
				break;

			case 18:
				regexFilter = RowFilter.regexFilter("Oscillator 3", 2);
				break;

			case 19:
				regexFilter = RowFilter.regexFilter("Envelope Generator 3", 2);
				break;

			default:
				regexFilter = null;
				break;
			}
			if (regexFilter == null) {
				continue;
			}
			if (box.isSelected()) {
				filters.add(regexFilter);
			}
		}
		final RowFilter<Object, Object> sidRegFilter = RowFilter.orFilter(filters);
		sorter.setRowFilter(sidRegFilter);

	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
		toggleDataCollection(fEnableRecording.isSelected());
	}

	private void toggleDataCollection(final boolean enable) {
		if (enable) {
			fRegModel.init();
			sidplay2.cp.m_engine.getC64().setSidWriteListener(0, fRegModel);
		} else {
			fRegModel.stop();
			sidplay2.cp.m_engine.getC64().setSidWriteListener(0, null);
		}
	}

	public void propertyChange(final PropertyChangeEvent evt) {
		if (ITuneStateChanged.class.getName().equals(evt.getPropertyName())) {
			toggleDataCollection(false);
		}
	}
}