package applet.stil;

import java.awt.GridLayout;
import java.beans.PropertyChangeEvent;

import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextPane;
import javax.swing.JTree;
import javax.swing.border.TitledBorder;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import libsidplay.Player;
import libsidplay.sidtune.SidTune;
import libsidutils.STIL.Info;
import libsidutils.STIL.STILEntry;
import libsidutils.STIL.TuneEntry;
import applet.JSIDPlay2;
import applet.TuneTab;

@SuppressWarnings("serial")
public class STILView extends TuneTab implements TreeSelectionListener {

	public static final String NAME = "Name:    ";
	public static final String AUTHOR = "Author:  ";
	public static final String TITLE = "Title:   ";
	public static final String ARTIST = "Artist:  ";
	public static final String COMMENT = "Comment: ";

	public static final int IDX_GLB_COMMENT = 0;

	public static final int IDX_NAME = 1;

	public static final int IDX_AUTHOR = 2;

	public static final int IDX_TITLE = 3;

	public static final int IDX_ARTIST = 4;

	public static final int IDX_COMMENT = 5;

	public static final String INFOS[] = new String[] { "STIL glb. Comment", "STIL Name", "STIL Author", "STIL Title", "STIL Artist", "STIL Comment" };

	private static final String STYLE_NORMAL = "STYLE_NORMAL";

	private static final String STYLE_FILENAME = "STYLE_FILENAME";

	private static final String STYLE_COMMENT = "STYLE_COMMENT";

	private static final String STYLE_SUBTUNE = "STYLE_SUBTUNE";

	private static final String STYLE_NAME = "STYLE_NAME";

	private static final String STYLE_AUTHOR = "STYLE_AUTHOR";

	private static final String STYLE_TITLE = "STYLE_TITLE";

	private static final String STYLE_ARTIST = "STYLE_ARTIST";

	private static final String STYLE_INFO = "STYLE_INFO";

	private final TreeModel fTreeModel = new TreeModel() {

		public void removeTreeModelListener(final TreeModelListener arg0) {
		}

		public void valueForPathChanged(final TreePath arg0, final Object arg1) {
		}

		public void addTreeModelListener(final TreeModelListener arg0) {

		}

		public Object getChild(final Object obj, final int i) {
			if (obj instanceof STILEntry) {
				final STILEntry entry = (STILEntry) obj;
				return entry.subtunes.get(i);
			}
			if (obj instanceof TuneEntry) {
				final TuneEntry tuneEntry = (TuneEntry) obj;
				return tuneEntry.infos.get(i);
			}
			return null;
		}

		public int getChildCount(final Object obj) {
			if (obj == null) {
				return 0;
			}
			if (obj instanceof STILEntry) {
				final STILEntry entry = (STILEntry) obj;
				return entry.subtunes.size();
			}
			if (obj instanceof TuneEntry) {
				final TuneEntry tuneEntry = (TuneEntry) obj;
				return tuneEntry.infos.size();
			}
			if (obj instanceof Info) {
				final Info info = (Info) obj;
				int count = 0;
				if (info.name != null) {
					count++;
				}
				if (info.author != null) {
					count++;
				}
				if (info.title != null) {
					count++;
				}
				if (info.artist != null) {
					count++;
				}
				if (info.comment != null) {
					count++;
				}
				return count;
			}
			return 0;
		}

		public int getIndexOfChild(final Object parent, final Object child) {
			if (parent instanceof STILEntry) {
				final STILEntry entry = (STILEntry) parent;
				for (int i = 0; i < entry.subtunes.size(); i++) {
					if (entry.subtunes.get(i) == child) {
						return i;
					}
				}
			}
			if (parent instanceof TuneEntry) {
				final TuneEntry tuneEntry = (TuneEntry) parent;
				for (int i = 0; i < tuneEntry.infos.size(); i++) {
					if (tuneEntry.infos.get(i) == child) {
						return i;
					}
				}
			}
			return -1;
		}

		public Object getRoot() {
			return fEntry;
		}

		public boolean isLeaf(final Object obj) {
			return obj instanceof Info/*Entry*/;
		}

	};

	private final JSIDPlay2 sidplay2;

	private final JTree fSTILTree = new JTree(new DefaultMutableTreeNode("<Please select HVSC>"));

	protected STILEntry fEntry;

	private JTextPane fSTILArea;

	private StyledDocument fDocument;

	private JScrollPane fScrollPane;

	public STILView(final JSIDPlay2 sidplay) {
		this.sidplay2 = sidplay;
		sidplay.support.addPropertyChangeListener(this);
	}

	@Override
	public void createContents() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		final JPanel stilPanel = new JPanel();
		stilPanel.setBorder(new TitledBorder("SID Tune Information List"));
		stilPanel.setLayout(new GridLayout(1, 2));
		add(stilPanel);

		// Where the tree is initialized:
		fSTILTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

		// Listen for when the selection changes.
		fSTILTree.addTreeSelectionListener(this);

		fSTILArea = new JTextPane();
		fSTILArea.setEditable(false);
		fDocument = fSTILArea.getStyledDocument();

		// Initialize some styles.
		final Style def = StyleContext.getDefaultStyleContext().getStyle(StyleContext.DEFAULT_STYLE);

		final Style regular = fDocument.addStyle("regular", def);
		StyleConstants.setFontFamily(def, "Arial");
		StyleConstants.setFontSize(def, 12);

		Style s = fDocument.addStyle(STYLE_NORMAL, regular);

		s = fDocument.addStyle(STYLE_FILENAME, regular);
		StyleConstants.setBold(s, true);

		s = fDocument.addStyle(STYLE_COMMENT, regular);
		StyleConstants.setItalic(s, true);

		s = fDocument.addStyle(STYLE_NAME, regular);
		StyleConstants.setBold(s, true);

		s = fDocument.addStyle(STYLE_TITLE, regular);
		StyleConstants.setBold(s, true);

		s = fDocument.addStyle(STYLE_AUTHOR, regular);
		StyleConstants.setBold(s, true);

		s = fDocument.addStyle(STYLE_ARTIST, regular);
		StyleConstants.setBold(s, true);

		s = fDocument.addStyle(STYLE_SUBTUNE, regular);
		StyleConstants.setBold(s, true);
		StyleConstants.setItalic(s, true);

		s = fDocument.addStyle(STYLE_INFO, regular);

		try {
			fDocument.remove(0, fDocument.getLength());
			fDocument.insertString(0, "no STIL available of the currently played tune." + "\nPlease select a tune from the HVSC tab decorated with a STIL icon", fDocument.getStyle(STYLE_NORMAL));
		} catch (final BadLocationException e) {
		}

		fScrollPane = new JScrollPane(fSTILArea);
		final JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, new JScrollPane(fSTILTree), fScrollPane);
		splitPane.setOneTouchExpandable(true);
		splitPane.setDividerLocation(100);
		stilPanel.add(splitPane);
	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
		if (m_tune == null) {
			return;
		}
		// sid2_info_t info = m_engine.info();
		final String filename = m_tune.getInfo().filename;
		fEntry = filename != null ? sidplay2.getSTIL(filename) : null;
		displayTune();
		if (fEntry != null) {
			fSTILTree.setSelectionPath(new TreePath(fEntry));
		} else {
			try {
				fDocument.remove(0, fDocument.getLength());
				fDocument.insertString(0, "no STIL available of the currently played tune." + "\nPlease select a tune from the HVSC tab decorated with a STIL icon", fDocument.getStyle(STYLE_NORMAL));
			} catch (final BadLocationException e) {
			}
		}
	}

	/**
	 * Show STIL if tree selection changes
	 */
	public void valueChanged(final TreeSelectionEvent event) {
		if (event.getNewLeadSelectionPath() == null) {
			return;
		}
		final Object comp = event.getNewLeadSelectionPath().getLastPathComponent();
		try {
			setTextAreaFromTree(comp);
		} catch (final BadLocationException e) {
			e.printStackTrace();
		}
	}

	private void setTextAreaFromTree(final Object comp) throws BadLocationException {
		if (!(comp instanceof DefaultMutableTreeNode)) {
			fSTILArea.setText("");
			writeSTIL(comp);
			fSTILArea.setCaretPosition(0);
		}
	}

	private void writeSTIL(final Object comp) throws BadLocationException {
		if (comp instanceof STILEntry) {
			final STILEntry entry = (STILEntry) comp;
			writeEntry(entry);
			for (int i = 0; i < entry.infos.size(); i++) {
				writeSTIL(entry.infos.get(i));
			}
			for (int i = 0; i < entry.subtunes.size(); i++) {
				writeSTIL(entry.subtunes.get(i));
			}
		} else if (comp instanceof TuneEntry) {
			final TuneEntry tuneEntry = (TuneEntry) comp;
			writeSubTune(tuneEntry);
			for (int i = 0; i < tuneEntry.infos.size(); i++) {
				writeSTIL(tuneEntry.infos.get(i));
			}
		} else if (comp instanceof Info) {
			final Info info = (Info) comp;
			writeInfo(info);
		}
	}

	private void writeEntry(final STILEntry entry) throws BadLocationException {
		fDocument.insertString(fDocument.getLength(), "Filename: ", fDocument.getStyle(STYLE_FILENAME));
		fDocument.insertString(fDocument.getLength(), entry.filename, fDocument.getStyle(STYLE_NORMAL));
		fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		if (entry.globalComment != null) {
			fDocument.insertString(fDocument.getLength(), entry.globalComment.trim(), fDocument.getStyle(STYLE_NORMAL));
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		}
		fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
	}

	private void writeSubTune(final TuneEntry tuneEntry) throws BadLocationException {
		fDocument.insertString(fDocument.getLength(), "Sub-tune #", fDocument.getStyle(STYLE_SUBTUNE));
		fDocument.insertString(fDocument.getLength(), tuneEntry.tuneNo + ":", fDocument.getStyle(STYLE_NORMAL));
		if (tuneEntry.globalComment != null) {
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
			fDocument.insertString(fDocument.getLength(), tuneEntry.globalComment.trim(), fDocument.getStyle(STYLE_COMMENT));
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		}
		fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
	}

	private void writeInfo(final Info info) throws BadLocationException {
		if (info.comment != null) {
			fDocument.insertString(fDocument.getLength(), info.comment.trim(), fDocument.getStyle(STYLE_NORMAL));
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		}
		if (info.name != null) {
			fDocument.insertString(fDocument.getLength(), "  Name: ", fDocument.getStyle(STYLE_NAME));
			fDocument.insertString(fDocument.getLength(), info.name, fDocument.getStyle(STYLE_NORMAL));
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		}
		if (info.author != null) {
			fDocument.insertString(fDocument.getLength(), "  Author: ", fDocument.getStyle(STYLE_AUTHOR));
			fDocument.insertString(fDocument.getLength(), info.author, fDocument.getStyle(STYLE_NORMAL));
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		}
		if (info.title != null) {
			fDocument.insertString(fDocument.getLength(), "  Title: ", fDocument.getStyle(STYLE_TITLE));
			fDocument.insertString(fDocument.getLength(), info.title, fDocument.getStyle(STYLE_NORMAL));
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		}
		if (info.artist != null) {
			fDocument.insertString(fDocument.getLength(), "  Artist: ", fDocument.getStyle(STYLE_ARTIST));
			fDocument.insertString(fDocument.getLength(), info.artist, fDocument.getStyle(STYLE_NORMAL));
			fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
		}
		fDocument.insertString(fDocument.getLength(), "\n", fDocument.getStyle(STYLE_NORMAL));
	}

	private void displayTune() {
		if (fSTILTree.getModel() != fSTILTree) {
			fSTILTree.setModel(fTreeModel);
		}
		fSTILTree.updateUI();
	}

	public void propertyChange(final PropertyChangeEvent event) {
		if (JSIDPlay2.PROP_HVSC.equals(event.getPropertyName())) {
			displayTune();
		}
	}
}
