package applet.video;

import java.awt.AWTKeyStroke;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.KeyboardFocusManager;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.beans.PropertyChangeEvent;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.swing.ImageIcon;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import libsidplay.Player;
import libsidplay.components.c1530.Datasette;
import libsidplay.components.keyboard.KeyTableEntry;
import libsidplay.components.keyboard.Keyboard;
import libsidplay.components.mos656x.MOS656X;
import libsidplay.sidtune.SidTune;
import applet.JSIDPlay2;
import applet.TuneTab;
import applet.ui.DKnob2;
import applet.ui.JNiceButton;

/**
 * Java AWT to C64 keyboard mapper.
 * 
 * Operation is based on hybrid of symbolic and positional mapping.
 * Some unobvious mappings have been chosen:
 * 
 * Esc = STOP, shift+Esc = RUN
 * F11 = NMI
 * § = arrow left
 * | = arrow up
 * ¦ = pi
 */
public class VideoView extends TuneTab implements KeyListener, MouseListener {

	private C64Canvas fC64Canvas;
	protected JCheckBox driveSoundBox;
	protected JCheckBox driveOnOffBox;
	protected JSIDPlay2 fSidplay2;
	private Keyboard keyboard;

	public VideoView(final JSIDPlay2 play2) {
		this.fSidplay2 = play2;
		play2.support.addPropertyChangeListener(this);
		addKeyListener(this);
		addMouseListener(this);
		setFocusable(true);
	}

	@Override
	public void createContents() {
		setLayout(new GridBagLayout());
		
		final GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.BOTH;
		
		final JPanel fDiskPanel = new JPanel();
		fDiskPanel.setBorder(new TitledBorder("Floppy Options"));
		fDiskPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		add(fDiskPanel, c);

		final JPanel fTapePanel = new JPanel();
		fTapePanel.setBorder(new TitledBorder("Datasette Options"));
		fTapePanel.setLayout(new FlowLayout(FlowLayout.LEFT));
		c.gridx = 1;
		add(fTapePanel, c);

		driveOnOffBox = new JCheckBox("Enable drive 8");
		driveOnOffBox.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.cp.m_engine.turnDriveOnOff(driveOnOffBox.isSelected());
			}
		});
		fDiskPanel.add(driveOnOffBox);

		driveSoundBox = new JCheckBox("Drive sound");
		driveSoundBox.addActionListener(new ActionListener() {

			public void actionPerformed(final ActionEvent e) {
				fSidplay2.setDriveSound(driveSoundBox.isSelected());
			}
		});
		fDiskPanel.add(driveSoundBox);

		final Datasette datasette = fSidplay2.cp.m_engine.getDatasette();
		
		JNiceButton fRecordButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/tape_record.png")));
		fRecordButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				// press record on tape
				datasette.control(Datasette.Control.RECORD);
			}

		});
		fRecordButton.setToolTipText("Record");
		fTapePanel.add(fRecordButton);

		JNiceButton fPlayButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/tape_play.png")));
		fPlayButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				// press play on tape
				datasette.control(Datasette.Control.START);
			}

		});
		fPlayButton.setToolTipText("Play");
		fTapePanel.add(fPlayButton);

		JNiceButton fRewindButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/tape_rewind.png")));
		fRewindButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				// press rewind on tape
				datasette.control(Datasette.Control.REWIND);
			}

		});
		fRewindButton.setToolTipText("Rewind");
		fTapePanel.add(fRewindButton);

		JNiceButton fForwardButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/tape_forward.png")));
		fForwardButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				// press forward on tape
				datasette.control(Datasette.Control.FORWARD);
			}

		});
		fForwardButton.setToolTipText("Forward");
		fTapePanel.add(fForwardButton);

		JNiceButton fStopButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/tape_stop.png")));
		fStopButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				// press stop on tape
				datasette.control(Datasette.Control.STOP);
			}

		});
		fStopButton.setToolTipText("Stop");
		fTapePanel.add(fStopButton);

		JNiceButton fResetCounterButton = new JNiceButton(new ImageIcon(JSIDPlay2.class.getResource("icons/mini/tape_reset.png")));
		fResetCounterButton.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(final MouseEvent e) {
				// reset tape counter
				datasette.control(Datasette.Control.RESET_COUNTER);
			}

		});
		fResetCounterButton.setToolTipText("Reset Counter");
		fTapePanel.add(fResetCounterButton);

		final JPanel fKeyScreen = new JPanel();
		fKeyScreen.setLayout(new GridBagLayout());
		fKeyScreen.setBorder(new TitledBorder("Screen"));
		c.gridx = 0;
		c.gridy ++;
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weighty = 1;
		c.weightx = 1;
		add(fKeyScreen, c);

		fC64Canvas = new C64Canvas();
		c.fill = GridBagConstraints.NONE;
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = GridBagConstraints.REMAINDER;
		fKeyScreen.add(fC64Canvas, c);
		fC64Canvas.addKeyListener(this);

		c.gridx = 0;
		c.gridy = 1;
		c.weighty = 0;
		c.gridwidth = 1;
		final DKnob2 brightness = new DKnob2("Brightness");
		brightness.setInterval(-100, 100);
		brightness.setDivisor(100);
		fKeyScreen.add(brightness, c);

		c.gridx = 1;
		final DKnob2 contrast = new DKnob2("Contrast");
		contrast.setInterval(0, 200);
		contrast.setDivisor(100);
		fKeyScreen.add(contrast, c);

		c.gridx = 2;
		final DKnob2 gamma = new DKnob2("Gamma");
		gamma.setInterval(180, 250);
		gamma.setDivisor(100);
		fKeyScreen.add(gamma, c);

		c.gridx = 3;
		final DKnob2 evenSaturation = new DKnob2("Saturation");
		evenSaturation.setInterval(0, 100);
		evenSaturation.setDivisor(100);
		fKeyScreen.add(evenSaturation, c);

		c.gridx = 4;
		final DKnob2 phaseShift = new DKnob2("Phase Shift", "deg");
		phaseShift.setInterval(-45, 45);
		phaseShift.setDivisor(1);
		fKeyScreen.add(phaseShift, c);

		c.gridx = 5;
		final DKnob2 offset = new DKnob2("Offset");
		offset.setInterval(10, 200);
		offset.setDivisor(100);
		fKeyScreen.add(offset, c);

		c.gridx = 7;
		final DKnob2 tint = new DKnob2("Tint");
		tint.setInterval(-10, 10);
		tint.setDivisor(100);
		fKeyScreen.add(tint, c);

		c.gridx = 8;
		final DKnob2 blur = new DKnob2("Blur");
		blur.setInterval(50, 100);
		blur.setDivisor(100);
		fKeyScreen.add(blur, c);

		c.gridx = 9;
		final DKnob2 bleed = new DKnob2("Bleed");
		bleed.setInterval(0, 10);
		bleed.setDivisor(10);
		fKeyScreen.add(bleed, c);

		{
			final MOS656X vic = fSidplay2.cp.m_engine.getC64().getVIC();
			brightness.setValue(vic.getPalette().getBrightness());
			contrast.setValue(vic.getPalette().getContrast());
			gamma.setValue(vic.getPalette().getGamma());
			evenSaturation.setValue(vic.getPalette().getSaturation());
			offset.setValue(vic.getPalette().getOffset());
			phaseShift.setValue(vic.getPalette().getPhaseShift());
			tint.setValue(vic.getPalette().getTint());
			blur.setValue(vic.getPalette().getLuminanceC());
			bleed.setValue(vic.getPalette().getDotCreep());
		}

		for (final DKnob2 knob : new DKnob2[] {
				brightness, contrast, gamma, evenSaturation, offset, phaseShift, tint, blur, bleed
		}) {
			knob.addChangeListener(new ChangeListener() {
				public void stateChanged(final ChangeEvent e) {
					final MOS656X vic = fSidplay2.cp.m_engine.getC64().getVIC();
					vic.getPalette().setBrightness(brightness.getValue());
					vic.getPalette().setContrast(contrast.getValue());
					vic.getPalette().setGamma(gamma.getValue());
					vic.getPalette().setSaturation(evenSaturation.getValue());
					vic.getPalette().setOffset(offset.getValue());
					vic.getPalette().setPhaseShift(phaseShift.getValue());
					vic.getPalette().setTint(tint.getValue());
					vic.getPalette().setLuminanceC(blur.getValue());
					vic.getPalette().setDotCreep(bleed.getValue());
					vic.updatePalette();
				}
			});

		}
		setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS, new HashSet<AWTKeyStroke>());
	}

	@Override
	public void setTune(final Player m_engine, final SidTune m_tune) {
		fC64Canvas.setC64(m_engine.getC64().getVIC());
		keyboard = m_engine.getKeyboard();
	}

	// maps key codes from key events to keys of the C64
	private final static Map<Integer, String> keycodeKeyMap = new HashMap<Integer, String>();

	static {
		keycodeKeyMap.put(KeyEvent.VK_BACK_SPACE, "BACKSPACE");
		keycodeKeyMap.put(KeyEvent.VK_ESCAPE, "ESCAPE");
		keycodeKeyMap.put(KeyEvent.VK_ENTER, "ENTER");
		keycodeKeyMap.put(KeyEvent.VK_TAB, "TAB");
		keycodeKeyMap.put(KeyEvent.VK_HOME, "HOME");
		keycodeKeyMap.put(KeyEvent.VK_INSERT, "INSERT");
		keycodeKeyMap.put(KeyEvent.VK_F11, "F11");

		keycodeKeyMap.put(KeyEvent.VK_F1, "F1");
		keycodeKeyMap.put(KeyEvent.VK_F2, "F2");
		keycodeKeyMap.put(KeyEvent.VK_F3, "F3");
		keycodeKeyMap.put(KeyEvent.VK_F4, "F4");
		keycodeKeyMap.put(KeyEvent.VK_F5, "F5");
		keycodeKeyMap.put(KeyEvent.VK_F6, "F6");
		keycodeKeyMap.put(KeyEvent.VK_F7, "F7");
		keycodeKeyMap.put(KeyEvent.VK_F8, "F8");

		keycodeKeyMap.put(KeyEvent.VK_LEFT, "LEFT");
		keycodeKeyMap.put(KeyEvent.VK_RIGHT, "RIGHT");
		keycodeKeyMap.put(KeyEvent.VK_UP, "UP");
		keycodeKeyMap.put(KeyEvent.VK_DOWN, "DOWN");

		keycodeKeyMap.put(KeyEvent.VK_A, "a");
		keycodeKeyMap.put(KeyEvent.VK_B, "b");
		keycodeKeyMap.put(KeyEvent.VK_C, "c");
		keycodeKeyMap.put(KeyEvent.VK_D, "d");
		keycodeKeyMap.put(KeyEvent.VK_E, "e");
		keycodeKeyMap.put(KeyEvent.VK_F, "f");
		keycodeKeyMap.put(KeyEvent.VK_G, "g");
		keycodeKeyMap.put(KeyEvent.VK_H, "h");
		keycodeKeyMap.put(KeyEvent.VK_I, "i");
		keycodeKeyMap.put(KeyEvent.VK_J, "j");
		keycodeKeyMap.put(KeyEvent.VK_K, "k");
		keycodeKeyMap.put(KeyEvent.VK_L, "l");
		keycodeKeyMap.put(KeyEvent.VK_M, "m");
		keycodeKeyMap.put(KeyEvent.VK_N, "n");
		keycodeKeyMap.put(KeyEvent.VK_O, "o");
		keycodeKeyMap.put(KeyEvent.VK_P, "p");
		keycodeKeyMap.put(KeyEvent.VK_Q, "q");
		keycodeKeyMap.put(KeyEvent.VK_R, "r");
		keycodeKeyMap.put(KeyEvent.VK_S, "s");
		keycodeKeyMap.put(KeyEvent.VK_T, "t");
		keycodeKeyMap.put(KeyEvent.VK_U, "u");
		keycodeKeyMap.put(KeyEvent.VK_V, "v");
		keycodeKeyMap.put(KeyEvent.VK_W, "w");
		keycodeKeyMap.put(KeyEvent.VK_X, "x");
		keycodeKeyMap.put(KeyEvent.VK_Y, "y");
		keycodeKeyMap.put(KeyEvent.VK_Z, "z");
	}

	/**
	 * Maps keys to KeyTableEntries
	 */
	private final static Map<String, KeyTableEntry> keyMap = new HashMap<String, KeyTableEntry>();

	private static KeyTableEntry shift(final KeyTableEntry in) {
		return new KeyTableEntry(in, true);
	}
	private static KeyTableEntry unshift(final KeyTableEntry in) {
		return new KeyTableEntry(in, false);
	}

	static {
		/* FIXME what does Awt call this key? Should be in keycodeKeyMap
		 * No guarantees that § and ½ are on same key separated by shift. */
		keyMap.put("§", KeyTableEntry.ARROW_LEFT);

		// FIXME where to put the arrow-up key, and by extension pi?
		// alt-gr + < works for me and is vertical,
		// but might fail for non-fi layouts.
		keyMap.put("|", KeyTableEntry.ARROW_UP);
		keyMap.put("¦", shift(KeyTableEntry.ARROW_UP));

		/* top row */
		keyMap.put("0", unshift(KeyTableEntry.ZERO));
		keyMap.put("1", unshift(KeyTableEntry.ONE));
		keyMap.put("!", shift(KeyTableEntry.ONE));
		keyMap.put("2", unshift(KeyTableEntry.TWO));
		keyMap.put("\"", shift(KeyTableEntry.TWO));
		keyMap.put("3", unshift(KeyTableEntry.THREE));
		keyMap.put("#", shift(KeyTableEntry.THREE));
		keyMap.put("4", unshift(KeyTableEntry.FOUR));
		keyMap.put("$", shift(KeyTableEntry.FOUR));
		keyMap.put("5", unshift(KeyTableEntry.FIVE));
		keyMap.put("%", shift(KeyTableEntry.FIVE));
		keyMap.put("6", unshift(KeyTableEntry.SIX));
		keyMap.put("&", shift(KeyTableEntry.SIX));
		keyMap.put("7", unshift(KeyTableEntry.SEVEN));
		keyMap.put("'", shift(KeyTableEntry.SEVEN));
		keyMap.put("8", unshift(KeyTableEntry.EIGHT));
		keyMap.put("(", shift(KeyTableEntry.EIGHT));
		keyMap.put("9", unshift(KeyTableEntry.NINE));
		keyMap.put(")", shift(KeyTableEntry.NINE));
		keyMap.put("+", unshift(KeyTableEntry.PLUS));
		keyMap.put("-", unshift(KeyTableEntry.MINUS));
		keyMap.put("£", unshift(KeyTableEntry.POUND));
		keyMap.put("HOME", KeyTableEntry.CLEAR_HOME);
		keyMap.put("BACKSPACE", KeyTableEntry.INS_DEL);
		keyMap.put("INSERT", shift(KeyTableEntry.INS_DEL));

		/* qwerty row */
		keyMap.put("TAB", KeyTableEntry.CTRL); // FIXME: can
		keyMap.put("q", KeyTableEntry.Q);
		keyMap.put("w", KeyTableEntry.W);
		keyMap.put("e", KeyTableEntry.E);
		keyMap.put("r", KeyTableEntry.R);
		keyMap.put("t", KeyTableEntry.T);
		keyMap.put("y", KeyTableEntry.Y);
		keyMap.put("u", KeyTableEntry.U);
		keyMap.put("i", KeyTableEntry.I);
		keyMap.put("o", KeyTableEntry.O);
		keyMap.put("p", KeyTableEntry.P);
		keyMap.put("@", unshift(KeyTableEntry.AT));
		keyMap.put("*", unshift(KeyTableEntry.STAR));

		/* asdf */
		keyMap.put("ESCAPE", KeyTableEntry.RUN_STOP);   // ctrl-R special map
		keyMap.put("a", KeyTableEntry.A);
		keyMap.put("s", KeyTableEntry.S);
		keyMap.put("d", KeyTableEntry.D);
		keyMap.put("f", KeyTableEntry.F);
		keyMap.put("g", KeyTableEntry.G);
		keyMap.put("h", KeyTableEntry.H);
		keyMap.put("j", KeyTableEntry.J);
		keyMap.put("k", KeyTableEntry.K);
		keyMap.put("l", KeyTableEntry.L);
		keyMap.put(":", unshift(KeyTableEntry.COLON));
		keyMap.put("[", shift(KeyTableEntry.COLON));
		keyMap.put(";", unshift(KeyTableEntry.SEMI_COLON));
		keyMap.put("]", shift(KeyTableEntry.SEMI_COLON));
		keyMap.put("=", unshift(KeyTableEntry.EQUALS));
		keyMap.put("ENTER", KeyTableEntry.RETURN);

		keyMap.put("z", KeyTableEntry.Z);
		keyMap.put("x", KeyTableEntry.X);
		keyMap.put("c", KeyTableEntry.C);
		keyMap.put("v", KeyTableEntry.V);
		keyMap.put("b", KeyTableEntry.B);
		keyMap.put("n", KeyTableEntry.N);
		keyMap.put("m", KeyTableEntry.M);
		keyMap.put(",", unshift(KeyTableEntry.COMMA));
		keyMap.put(".", unshift(KeyTableEntry.DOT));
		keyMap.put("<", shift(KeyTableEntry.COMMA));
		keyMap.put(">", shift(KeyTableEntry.DOT));
		keyMap.put("/", unshift(KeyTableEntry.SLASH));
		keyMap.put("?", shift(KeyTableEntry.SLASH));
		keyMap.put("DOWN", unshift(KeyTableEntry.CURSOR_UP_DOWN));
		keyMap.put("UP", shift(KeyTableEntry.CURSOR_UP_DOWN));
		keyMap.put("RIGHT", unshift(KeyTableEntry.CURSOR_LEFT_RIGHT));
		keyMap.put("LEFT", shift(KeyTableEntry.CURSOR_LEFT_RIGHT));

		/* space row */
		keyMap.put(" ", KeyTableEntry.SPACE);

		// Function keys: F1-F8
		keyMap.put("F1", KeyTableEntry.F1);
		keyMap.put("F2", shift(KeyTableEntry.F1));
		keyMap.put("F3", KeyTableEntry.F3);
		keyMap.put("F4", shift(KeyTableEntry.F3));
		keyMap.put("F5", KeyTableEntry.F5);
		keyMap.put("F6", shift(KeyTableEntry.F5));
		keyMap.put("F7", KeyTableEntry.F7);
		keyMap.put("F8", shift(KeyTableEntry.F7));
	}

	// implementation of the KeyListener interface
	/**
	 * Convert key event to key selection suitable for the Keyboard class and
	 * pass it to that class
	 */
	private final Map<Integer, KeyTableEntry> prevailingModifiers = new HashMap<Integer, KeyTableEntry>();
	public void keyPressed(final KeyEvent event) {
		final int keyCode = event.getKeyCode();

		if (keyCode == KeyEvent.VK_CONTROL) {
			keyboard.cbm(true);
			return;
		}

		/* Maintain the illusion that ctrl == cbm. Windows sends
		 * CTRL and ALT when Alt Gr is used, and thus we cancel the earlier CTRL key
		 * whenever Alt Gr is used. This allows typing $ in Finnish layout.
		 * Unfortunately, there appears to be no way to reliably detect if the control
		 * key is *really* down.
		 * 
		 * However, user can still press alt first and ctrl next, and
		 * that way achieve some combos like cbm+@.
		 */
		if (keyCode == KeyEvent.VK_ALT) {
			keyboard.cbm(false);
			return;
		}
		
		if (keyCode == KeyEvent.VK_SHIFT) {
			if (event.getKeyLocation() == KeyEvent.KEY_LOCATION_LEFT) {
				keyboard.leftShift(true);
			}
			if (event.getKeyLocation() == KeyEvent.KEY_LOCATION_RIGHT) {
				keyboard.rightShift(true);
			}
			return;
		}

		// any of our special keycodes which have no name
		final String key;
		if (keycodeKeyMap.containsKey(keyCode)) {
			key = keycodeKeyMap.get(keyCode);
		} else {
			key = "" + event.getKeyChar();
		}

		/* RESTORE key is hooked straight into CPU's NMI line. The
		 * stop + restore invocation is thus "Escape + F11". The NMI
		 * is probably asserted only for a very short time, because it
		 * does not seem to mask simultaneous CIA interrupts despite
		 * being held down. */
		if ("F11".equals(key)) {
			fSidplay2.cp.m_engine.getKeyboard().restore();
			return;
		}

		/* When key is released, java doesn't report the modifiers.
		 * So we track all pressed keys and their modifiers here to ensure
		 * that when the key is released, we hand the right kind of event to
		 * Keyboard.
		 */
		final KeyTableEntry ke = keyMap.get(key);
		if (ke != null) {
			/* this key is already down? Well, let's not report it TWICE. */
			if (! prevailingModifiers.containsKey(event.getKeyCode())) {
				keyboard.keyPressed(ke);
				prevailingModifiers.put(event.getKeyCode(), ke);
			}
		}
	}

	/**
	 * Convert key event to key selection suitable for the Keyboard class and
	 * pass it to that class
	 */
	public void keyReleased(final KeyEvent event) {
		final int keyCode = event.getKeyCode();

		/* earlier pressed key was released. */
		if (prevailingModifiers.containsKey(keyCode)) {
			keyboard.keyReleased(prevailingModifiers.remove(event.getKeyCode()));
			return;
		}

		/* release NMI */
		if (keycodeKeyMap.containsKey(keyCode)) {
			final String key = keycodeKeyMap.get(keyCode);
			if ("F11".equals(key)) {
				return;
			}
		}

		if (keyCode == KeyEvent.VK_CONTROL) {
			keyboard.cbm(false);
		}

		if (keyCode == KeyEvent.VK_SHIFT) {
			if (event.getKeyLocation() == KeyEvent.KEY_LOCATION_LEFT) {
				keyboard.leftShift(false);
			}
			if (event.getKeyLocation() == KeyEvent.KEY_LOCATION_RIGHT) {
				keyboard.rightShift(false);
			}
		}

	}

	public void keyTyped(final KeyEvent event) {
	}

	public void mouseClicked(final MouseEvent e) {
		requestFocusInWindow();
	}

	public void mouseEntered(final MouseEvent e) {
	}

	public void mouseExited(final MouseEvent e) {
	}

	public void mousePressed(final MouseEvent e) {
	}

	public void mouseReleased(final MouseEvent e) {
	}

	public void propertyChange(PropertyChangeEvent evt) {
	}
}
