/**
 *                           Sid Builder Classes
 *                           -------------------
 *  begin                : Sat May 6 2001
 *  copyright            : (C) 2001 by Simon White
 *  email                : s_a_white@email.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 * @author Ken Hndel
 *
 */
package libsidplay.common;

import resid_builder.FilterConfig;
import resid_builder.resid.ISIDDefs.SamplingMethod;

public abstract class SIDEmu {

	private final SIDBuilder builder;

	protected final EventScheduler context;

	protected int m_bufferpos;
	protected float m_buffer[];
	private long lastTime;

	private final byte[] sidMem = new byte[32];

	public SIDEmu(final SIDBuilder builder, EventScheduler context) {
		this.builder = builder;
		this.context = context;
	}

	//
	// Standard component functions
	//

	public final void reset() {
		lastTime = context.getTime(Event.Phase.PHI2);
		reset((byte) 0);
	}

	public abstract void reset(byte volume);

	public abstract byte read(int addr);

	public byte readInternalRegister(final int addr) {
		return sidMem[addr];
	}

	public void write(final int addr, final byte data) {
		sidMem[addr] = data;
	}

	public abstract void clock();
	
	public final int clocksSinceLastAccess() {
		final long now = context.getTime(Event.Phase.PHI2);
		int diff = (int) (now - lastTime);
		lastTime = now;
		return diff;
	}

	//
	// Standard SID functions
	//

	public abstract void voice(int num, boolean mute);

	final public SIDBuilder builder() {
		return builder;
	}

	public final int bufferpos() { return m_bufferpos; }
	public final void bufferpos(final int pos) { m_bufferpos = pos; }
	public final float[] buffer() { return m_buffer; }

	/**
	 * Configure audio sampling rate
	 * 
	 * @param systemfreq System main clock 
	 * @param outputfreq Output sampling frequency
	 * @param method Sampling method to use for converstion.
	 */
	public void sampling(final double systemfreq, final float outputfreq, final SamplingMethod method) {
	}

	public abstract void filter(boolean enable);
	public abstract void filter(FilterConfig filter6581, FilterConfig filter8580);
}
