/*
 * via1d15xx.c - VIA1 emulation in the 1541, 1541II, 1570 and 1571 disk drive.
 *
 * Written by
 *  Andreas Boose <viceteam@t-online.de>
 *  Andre' Fachat <fachat@physik.tu-chemnitz.de>
 *  Daniel Sladic <sladic@eecg.toronto.edu>
 *  Ettore Perazzoli <ettore@comm2000.it>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */
package libsidplay.components.c1541;

import libsidplay.common.EventScheduler;

/**
 * Handling of the Bus Controller.
 * 
 * @author Ken Hndel
 * 
 */
public final class VIA6522BC extends VIACore {
	/**
	 * IEC Bus.
	 */
	private final IECBus iecbus;

	/**
	 * Creates a new instance of VIA6522BC.
	 * 
	 * @param c1541
	 *            the C1541 we are attached to
	 * @param bus
	 *            the IEC Bus
	 * @param ctx
	 *            event context
	 */
	public VIA6522BC(final C1541 c1541, final IECBus bus,
			final EventScheduler ctx) {
		super(ctx, c1541, "1541Drive" + c1541.getID() + "Via1");
		this.iecbus = bus;
	}

	@Override
	public void setCa2(final int state) {
	}

	@Override
	public void setCb2(final int state) {
	}

	@Override
	public void storePra(final byte value) {
	}

	@Override
	public final void storePrb(final byte byt) {
		if (byt != oldpb) {
			final byte[] driveBus = iecbus.drvBus;
			final byte[] driveData = iecbus.drvData;

			driveData[drive.getID() + 8] = (byte) ~byt;
			driveBus[drive.getID() + 8] = (byte) ((driveData[drive.getID() + 8] & 0xff) << 3 & 0x40 | (driveData[drive
			                                                                                                     .getID() + 8] & 0xff) << 6
			                                                                                                     & (~driveData[drive.getID() + 8] ^ iecbus.cpuBus) << 3
			                                                                                                     & 0x80);

			iecbus.updatePorts();
		}
	}

	@Override
	public void storeAcr(final byte value) {
	}

	@Override
	public void storeSr(final byte value) {
	}

	@Override
	public void storeT2l(final byte value) {
	}

	@Override
	public final byte readPra() {
		// TODO parallel cable support
		return (byte) (via[VIACore.VIA_PRA] & via[VIACore.VIA_DDRA] | 0xff & ~via[VIACore.VIA_DDRA]);
	}

	@Override
	public final byte readPrb() {
		final int number = drive.getID();
		/* 0 for drive0, 0x20 for drive 1 */
		byte orval = (byte) (number << 5);

		return (byte) (((via[VIACore.VIA_PRB] & 0x1a) | iecbus.drvPort) ^ 0x85 | orval);
	}
}
