package libsidplay.sidtune;

public class T64 extends Prg {

	protected static SidTune load(final String fileName, final byte[] dataBuf)
			throws SidTuneError {
		if (fileName != null) {
			final int lastIndexOf = fileName.lastIndexOf(".");
			final String ext = lastIndexOf != -1 ? fileName
					.substring(lastIndexOf) : "";
			if (!ext.equalsIgnoreCase(".t64")) {
				return null;
			}
		}

		final T64 sidtune = new T64();
		
		// always load first entry:
		if (!sidtune.getEntry(dataBuf, 1)) {
			return null;
		}

		// Automatic settings
		sidtune.info.songs = 1;
		sidtune.info.startSong = 1;
		sidtune.info.numberOfInfoStrings = 0;
		sidtune.program = dataBuf;

		// Create the speed/clock setting table.
		sidtune.convertOldStyleSpeedToTables(~0, sidtune.info.clockSpeed);
		return sidtune;
	}

	/**
	 * Load T64 file entry from buffer.
	 * 
	 * @param dataBuf
	 *            buffer with file data
	 * @param entryNum
	 *            entry number to load
	 * @return true - success
	 */
	private boolean getEntry(final byte[] dataBuf, final int entryNum) {
		int totalEntries = ((dataBuf[35] & 0xff) << 8) + dataBuf[34] & 0xff;
		if ((entryNum < 1) || (entryNum > totalEntries)) {
			// entry not available
			return false;
		}

		// determine T64 entry offset
		int pos = 32 /* header */ + 32 * entryNum;

		// expect 1 (Normal tape file)
		if (dataBuf[pos++] != 1)
			return false;
		// expect 0x82 (PRG)
		if (dataBuf[pos++] != (byte) 0x82)
			return false;

		// Get start address (or Load address)
		info.loadAddr = (dataBuf[pos++] & 0xff)
				+ ((dataBuf[pos++] & 0xff) << 8);
		
		// Get end address (actual end address in memory, if the file was loaded
		// into a C64).
		info.c64dataLen = (dataBuf[pos++] & 0xff)
				+ ((dataBuf[pos++] & 0xff) << 8);
		info.c64dataLen -= info.loadAddr;
		
		// skip unused
		pos += 2;

		// determine offset of program data
		fileOffset = 0;
		for (int power = 0; power <= 3; power++) {
			fileOffset += (dataBuf[pos++] & 0xff) << 8 * power;
		}

		// skip unused
		pos += 4;
		
		// Get program name
		info.filename = new String(dataBuf, pos, 16).trim();
		return true;
	}

}
