package netsiddev;

import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.Image;
import java.awt.Panel;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.net.URL;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JRootPane;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.border.Border;

public class AboutBox implements ActionListener {
	private final static String credits = "JSIDDevice v1.0 - SID Network Device\n\n" +
										  "JSIDDevice uses parts of JSidplay2 which are\ncopyrighted to:\n\n" +
										  "Network Interface:\n  Copyright  2007-2010 Ken Hndel,\n" +
										  "  Antti S. Lankila and Wilfred Bos\n" +
										  "Distortion Simulation and 6581/8580 emulation:\n  Copyright  2005-2010 Antti S. Lankila\n" +
										  "ReSID engine and 6581/8580 emulation:\n  Copyright  1999-2004 Dag Lem\n\n" +
										  "Source code can be found at:\n" +
										  "  http://sourceforge.net/projects/jsidplay2/";

	private Button okButton;

	private Frame frame = new Frame();
	private JDialog dialog = new EscapeDialog(frame , "About JSIDDevice", true);

	AboutBox() {
		dialog.setLayout(new BorderLayout());
		
        URL url = this.getClass().getResource("jsidplay2.gif"); 
        Image image = Toolkit.getDefaultToolkit().getImage(url);
        dialog.setIconImage(image);

		dialog.add(createTextAreaCredits());
		dialog.add("South", createOKPanel());

		centerFrame();
		dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
		dialog.pack();
	}

	private Panel createTextAreaCredits() {
		Panel p = new Panel();
		JTextArea textField = new JTextArea(credits);
		textField.setEditable(false);
		textField.setLineWrap(false);
		textField.setHighlighter(null);
		Border outsideBorder = BorderFactory.createLineBorder(Color.BLACK, 1);
		Border insideBorder = BorderFactory.createEmptyBorder(4, 4, 4, 4);
		textField.setBorder(BorderFactory.createCompoundBorder(outsideBorder, insideBorder));
		p.add(textField);
		return p;
	}

	private Panel createOKPanel() {
		Panel p = new Panel();
		p.setLayout(new FlowLayout());
		
		okButton = new Button("OK");
		p.add(okButton);
		okButton.addActionListener(this);
		return p;
	}
	
	private void centerFrame() {
		Dimension d = dialog.getToolkit().getScreenSize();
		dialog.setLocation(d.width / 3, d.height / 3);
	}

	public void show() {
		dialog.setVisible(true);
	}

	public void actionPerformed(ActionEvent ae) {
		if (ae.getSource() == okButton) {
			dialog.setVisible(false);
		}
	}
	
	private class EscapeDialog extends JDialog {
		private static final long serialVersionUID = 7957643410825068191L;

		public EscapeDialog(Frame owner, String title, boolean modal) {
			super(owner, title, modal);
		}

		protected JRootPane createRootPane() {
			JRootPane rootPane = new JRootPane();
			KeyStroke stroke = KeyStroke.getKeyStroke("ESCAPE");
			Action actionListener = new AbstractAction() {
				private static final long serialVersionUID = 9016737277049160164L;

				public void actionPerformed(ActionEvent actionEvent) {
					setVisible(false);
				}
			};
			InputMap inputMap = rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
			inputMap.put(stroke, "ESCAPE");
			rootPane.getActionMap().put("ESCAPE", actionListener);

			return rootPane;
		}

	}
	
	public void dispose() {
		frame.dispose();
	}
}
