/**
 *             ReSID builder class for creating/controlling resids
 *             ---------------------------------------------------
 *  begin                : Wed Sep 5 2001
 *  copyright            : (C) 2001 by Simon White
 *  email                : s_a_white@email.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 * @author Ken Hndel
 *
 */
package resid_builder;

import java.util.ArrayList;
import java.util.List;

import libsidplay.common.EventScheduler;
import libsidplay.common.SIDBuilder;
import libsidplay.common.SIDEmu;
import resid_builder.resid.ISIDDefs.ChipModel;

/*******************************************************************************
 * Since ReSID is not part of this project we are actually creating a wrapper
 * instead of implementing a SID emulation
 * *****************************************************************************
 * ReSID Builder Class
 * *****************************************************************************
 * Create the SID builder object
 */
public class ReSIDBuilder extends SIDBuilder {
	// Error String(s)

	private final List<ReSID> sidobjs = new ArrayList<ReSID>();

	/**
	 * Find a free SID of the required specs
	 */
	@Override
	public SIDEmu lock(final EventScheduler env, final ChipModel model) {
		final ReSID sid = new ReSID(this, env);
		sid.model(model);
		sidobjs.add(sid);
		return sid;
	}

	/**
	 * Allow something to use this SID
	 */
	@Override
	public void unlock(final SIDEmu device) {
		sidobjs.remove(device);
	}

	// Settings that affect all SIDS

	@Override
	public void filter(final boolean enable) {
		final int size = sidobjs.size();
		for (int i = 0; i < size; i++) {
			final ReSID sid = sidobjs.get(i);
			sid.filter(enable);
		}
	}

	public void filter(final FilterConfig filter6581, final FilterConfig filter8580) {
		final int size = sidobjs.size();
		for (int i = 0; i < size; i++) {
			final ReSID sid = sidobjs.get(i);
			sid.filter(filter6581, filter8580);
		}
	}
}
