/**
 *                         Sidplay2 config file reader.
 *                         ----------------------------
 *  begin                : Sun Mar 25 2001
 *  copyright            : (C) 2000 by Simon White
 *  email                : s_a_white@email.com
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 * @author Ken Hndel
 *
 */
package sidplay;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.AccessControlException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.StringTokenizer;

import libsidplay.common.ISID2Types.Clock;
import libsidutils.IniReader;
import resid_builder.FilterConfig;
import resid_builder.resid.ISIDDefs.ChipModel;
import resid_builder.resid.ISIDDefs.SamplingMethod;

public class IniConfig {
	/* bump this each time you want to invalidate sidplay2.ini files on disk */
	private static final int REQUIRED_CONFIG_VERSION = 14;

	/**
	 * INI configuration filename or null (use internal configuration)
	 */
	final String iniPath;

	public class sidplay2_section {
		private int version;
		public int getVersion() {
			return version;
		}
		private boolean enableDatabase;
		public boolean isEnableDatabase() {
			return enableDatabase;
		}
		public void setEnableDatabase(boolean enableDatabase) {
			ini.makeDirty(this.enableDatabase, enableDatabase);
			this.enableDatabase = enableDatabase;
		}
		private int playLength;
		public int getPlayLength() {
			return playLength;
		}
		public void setPlayLength(int playLength) {
			ini.makeDirty(this.playLength, playLength);
			this.playLength = playLength;
		}
		private int recordLength;
		public int getRecordLength() {
			return recordLength;
		}
		private String cgsc;
		public String getCgsc() {
			return cgsc;
		}
		public void setCgsc(String cgsc) {
			ini.makeDirty(this.cgsc, cgsc);
			this.cgsc = cgsc;
		}
		private String hvsc;
		public String getHvsc() {
			return hvsc;
		}
		public void setHvsc(String hvsc) {
			ini.makeDirty(this.hvsc, hvsc);
			this.hvsc = hvsc;
		}
		private boolean single;
		public boolean isSingle() {
			return single;
		}
		public void setSingle(boolean single) {
			ini.makeDirty(this.single, single);
			this.single = single;
		}
		private String soasc6581R2;
		public String getSoasc6581R2() {
			return soasc6581R2;
		}
		public void setSoasc6581R2(String soasc6581r2) {
			ini.makeDirty(this.soasc6581R2, soasc6581r2);
			soasc6581R2 = soasc6581r2;
		}
		private String soasc6581R4;
		public String getSoasc6581R4() {
			return soasc6581R4;
		}
		public void setSoasc6581R4(String soasc6581r4) {
			ini.makeDirty(this.soasc6581R4, soasc6581r4);
			soasc6581R4 = soasc6581r4;
		}
		private String soasc8580R5;
		public String getSoasc8580R5() {
			return soasc8580R5;
		}
		public void setSoasc8580R5(String soasc8580r5) {
			ini.makeDirty(this.soasc8580R5, soasc8580r5);
			soasc8580R5 = soasc8580r5;
		}
		private boolean enableProxy;
		public boolean isEnableProxy() {
			return enableProxy;
		}
		public void setEnableProxy(boolean enableProxy) {
			ini.makeDirty(this.enableProxy, enableProxy);
			this.enableProxy = enableProxy;
		}
		private String proxyHostname;
		public String getProxyHostname() {
			return proxyHostname;
		}
		public void setProxyHostname(String proxyHostname) {
			ini.makeDirty(this.proxyHostname, proxyHostname);
			this.proxyHostname = proxyHostname;
		}
		private int proxyPort;
		public int getProxyPort() {
			return proxyPort;
		}
		public void setProxyPort(int proxyPort) {
			ini.makeDirty(this.proxyPort, proxyPort);
			this.proxyPort = proxyPort;
		}

		private String lastDirectory;
		public String getLastDirectory() {
			return lastDirectory;
		}
		public void setLastDirectory(String lastDirectory) {
			ini.makeDirty(this.lastDirectory, lastDirectory);
			this.lastDirectory = lastDirectory;
		}
		
		// application window metrics
		private int frameX;
		public int getFrameX() {
			return frameX;
		}
		public void setFrameX(int frameX) {
			ini.makeDirty(this.frameX, frameX);
			this.frameX = frameX;
		}
		private int frameY;
		public int getFrameY() {
			return frameY;
		}
		public void setFrameY(int frameY) {
			ini.makeDirty(this.frameY, frameY);
			this.frameY = frameY;
		}
		private int frameWidth;
		public int getFrameWidth() {
			return frameWidth;
		}
		public void setFrameWidth(int frameWidth) {
			ini.makeDirty(this.frameWidth, frameWidth);
			this.frameWidth = frameWidth;
		}
		private int frameHeight;
		public int getFrameHeight() {
			return frameHeight;
		}
		public void setFrameHeight(int frameHeight) {
			ini.makeDirty(this.frameHeight, frameHeight);
			this.frameHeight = frameHeight;
		}
	}

	public class console_section {
		// INI Section - [Console]
		private boolean ansi;
		public boolean isAnsi() {
			return ansi;
		}
		private char topLeft;
		public char getTopLeft() {
			return topLeft;
		}
		private char topRight;
		public char getTopRight() {
			return topRight;
		}
		private char bottomLeft;
		public char getBottomLeft() {
			return bottomLeft;
		}
		private char bottomRight;
		public char getBottomRight() {
			return bottomRight;
		}
		private char vertical;
		public char getVertical() {
			return vertical;
		}
		private char horizontal;
		public char getHorizontal() {
			return horizontal;
		}
		private char junctionLeft;
		public char getJunctionLeft() {
			return junctionLeft;
		}
		private char junctionRight;
		public char getJunctionRight() {
			return junctionRight;
		}
	}

	public class audio_section {
		// INI Section - [Audio]
		private int frequency;
		public int getFrequency() {
			return frequency;
		}
		public void setFrequency(int frequency) {
			ini.makeDirty(this.frequency, frequency);
			this.frequency = frequency;
		}
		private SamplingMethod sampling;
		public SamplingMethod getSampling() {
			return sampling;
		}
		public void setSampling(SamplingMethod sampling) {
			ini.makeDirty(this.sampling, sampling);
			this.sampling = sampling;
		}
		private int channels;
		public int getChannels() {
			return channels;
		}
		private boolean playOriginal;
		public boolean isPlayOriginal() {
			return playOriginal;
		}
		public void setPlayOriginal(boolean playOriginal) {
			this.playOriginal = playOriginal;
		}
		private String mp3File;
		public String getMp3File() {
			return mp3File;
		}
		public void setMp3File(String mp3File) {
			ini.makeDirty(this.mp3File, mp3File);
			this.mp3File = mp3File;
		}
	}

	public class emulation_section {
		// INI Section - [Emulation]
		private Clock defaultClockSpeed;
		public Clock getDefaultClockSpeed() {
			return defaultClockSpeed;
		}
		private Clock userClockSpeed;
		public Clock getUserClockSpeed() {
			return userClockSpeed;
		}
		private ChipModel defaultSidModel;
		public ChipModel getDefaultSidModel() {
			return defaultSidModel;
		}
		private ChipModel userSidModel;
		public ChipModel getUserSidModel() {
			return userSidModel;
		}
		public void setUserSidModel(ChipModel userSidModel) {
			ini.makeDirty(this.userSidModel, userSidModel);
			this.userSidModel = userSidModel;
		}
		private int hardsid6581;
		public int getHardsid6581() {
			return hardsid6581;
		}
		public void setHardsid6581(int hardsid6581) {
			ini.makeDirty(this.hardsid6581, hardsid6581);
			this.hardsid6581 = hardsid6581;
		}
		private int hardsid8580;
		public int getHardsid8580() {
			return hardsid8580;
		}
		public void setHardsid8580(int hardsid8580) {
			ini.makeDirty(this.hardsid8580, hardsid8580);
			this.hardsid8580 = hardsid8580;
		}
		private boolean filter;
		public boolean isFilter() {
			return filter;
		}
		public void setFilter(boolean filter) {
			ini.makeDirty(this.filter, filter);
			this.filter = filter;
		}
		private String filter6581;
		public String getFilter6581() {
			return filter6581;
		}
		public void setFilter6581(String filter6581) {
			ini.makeDirty(this.filter6581, filter6581);
			this.filter6581 = filter6581;
		}
		private String filter8580;
		public String getFilter8580() {
			return filter8580;
		}
		public void setFilter8580(String filter8580) {
			ini.makeDirty(this.filter8580, filter8580);
			this.filter8580 = filter8580;
		}
		private boolean digiBoosted8580;
		public boolean isDigiBoosted8580() {
			return digiBoosted8580;
		}
		public void setDigiBoosted8580(boolean digiBoosted8580) {
			ini.makeDirty(this.digiBoosted8580, digiBoosted8580);
			this.digiBoosted8580 = digiBoosted8580;
		}
		private int dualSidBase;
		public int getDualSidBase() {
			return dualSidBase;
		}
		public void setDualSidBase(int dualSidBase) {
			ini.makeDirty(this.dualSidBase, dualSidBase);
			this.dualSidBase = dualSidBase;
		}
		private boolean forceStereoTune;
		public boolean isForceStereoTune() {
			return forceStereoTune;
		}
		public void setForceStereoTune(boolean forceStereoTune) {
			ini.makeDirty(this.forceStereoTune, forceStereoTune);
			this.forceStereoTune = forceStereoTune;
		}

		private ChipModel stereoSidModel;
		public ChipModel getStereoSidModel() {
			return stereoSidModel;
		}
		public void setStereoSidModel(ChipModel stereoSidModel) {
			ini.makeDirty(this.stereoSidModel, stereoSidModel);
			this.stereoSidModel = stereoSidModel;
		}
	}

	public class favorites_section {
		// favorites
		private String[] favorites_titles;
		public String[] getFavorites_titles() {
			return favorites_titles;
		}
		public void setFavorites_titles(String[] favoritesTitles) {
			ini.makeDirty(this.favorites_titles, favoritesTitles);
			favorites_titles = favoritesTitles;
		}
		private String[] favorites_filenames;
		public String[] getFavorites_filenames() {
			return favorites_filenames;
		}
		public void setFavorites_filenames(String[] favoritesFilenames) {
			ini.makeDirty(this.favorites_filenames, favoritesFilenames);
			favorites_filenames = favoritesFilenames;
		}
		private String favorites_current;
		public String getFavorites_current() {
			return favorites_current;
		}
		public void setFavorites_current(String favoritesCurrent) {
			ini.makeDirty(this.favorites_current, favoritesCurrent);
			favorites_current = favoritesCurrent;
		}
	}

	protected static final String FILE_NAME = "sidplay2.ini";

	protected boolean status;
	protected sidplay2_section sidplay2_s = new sidplay2_section();
	protected console_section console_s = new console_section();
	protected audio_section audio_s = new audio_section();
	protected emulation_section emulation_s = new emulation_section();
	protected favorites_section favorites_s = new favorites_section();
	protected FilterConfig currentFilter6581;
	protected FilterConfig currentFilter8580;

	private IniReader ini;

	protected void clear() {
		status = true;

		sidplay2_s.version = 0; // INI File Version; set from .ini or internal loading only
		sidplay2_s.hvsc = null;
		sidplay2_s.cgsc = null;
		sidplay2_s.enableDatabase = true;
		sidplay2_s.single = false;
		sidplay2_s.playLength = 0; // INFINITE
		sidplay2_s.recordLength = 3 * 60 + 30; // 3.5 minutes
		sidplay2_s.soasc6581R2 = null;
		sidplay2_s.soasc6581R4 = null;
		sidplay2_s.soasc8580R5 = null;
		sidplay2_s.enableProxy = true;
		sidplay2_s.proxyHostname = null;
		sidplay2_s.proxyPort = 80;
		sidplay2_s.frameX = -1; // uninitialized
		sidplay2_s.frameY = -1; // uninitialized
		sidplay2_s.frameWidth = -1; // uninitialized
		sidplay2_s.frameHeight = -1; // uninitialized
		sidplay2_s.lastDirectory = null;
		
		console_s.ansi = false;
		console_s.topLeft = '+';
		console_s.topRight = '+';
		console_s.bottomLeft = '+';
		console_s.bottomRight = '+';
		console_s.vertical = '|';
		console_s.horizontal = '-';
		console_s.junctionLeft = '+';
		console_s.junctionRight = '+';

		audio_s.frequency = 48000;
		audio_s.sampling = SamplingMethod.DECIMATE;
		audio_s.channels = 1;
		audio_s.mp3File = null;
		audio_s.playOriginal = false;

		emulation_s.defaultClockSpeed = Clock.PAL;
		emulation_s.userClockSpeed = null;
		emulation_s.defaultSidModel = ChipModel.MOS6581;
		emulation_s.userSidModel = null;
		emulation_s.hardsid6581 = 1;
		emulation_s.hardsid8580 = 1;
		emulation_s.filter = true;
		emulation_s.dualSidBase = 0xd420;
		emulation_s.forceStereoTune = false;
		emulation_s.stereoSidModel = null;

		emulation_s.filter6581 = null;
		emulation_s.filter8580 = null;
		emulation_s.digiBoosted8580 = false;

		favorites().favorites_titles = new String[0];
		favorites().favorites_filenames = new String[0];
	}

	protected boolean readSidplay2() {
		sidplay2_s.version = ini.getPropertyInt("SIDPlay2", "Version", 0);
		sidplay2_s.hvsc = ini.getPropertyString("SIDPlay2", "HVSC Dir", sidplay2_s.hvsc);
		sidplay2_s.cgsc = ini.getPropertyString("SIDPlay2", "CGSC Dir", sidplay2_s.cgsc);
		sidplay2_s.soasc6581R2 = ini.getPropertyString("SIDPlay2", "SOASC_6581R2", sidplay2_s.soasc6581R2);
		sidplay2_s.soasc6581R4 = ini.getPropertyString("SIDPlay2", "SOASC_6581R4", sidplay2_s.soasc6581R4);
		sidplay2_s.soasc8580R5 = ini.getPropertyString("SIDPlay2", "SOASC_8580R5", sidplay2_s.soasc8580R5);
		sidplay2_s.enableProxy = ini.getPropertyInt("SIDPlay2", "EnableProxy", sidplay2_s.enableProxy ? 1 : 0) > 0;
		sidplay2_s.proxyHostname = ini.getPropertyString("SIDPlay2", "ProxyHostname", sidplay2_s.proxyHostname);
		sidplay2_s.proxyPort = ini.getPropertyInt("SIDPlay2", "ProxyPort", sidplay2_s.proxyPort);
		sidplay2_s.enableDatabase = ini.getPropertyInt("SIDPlay2", "EnableDatabase", sidplay2_s.enableDatabase ? 1 : 0) > 0;

		int time = 0;
		if ((time = ini.getPropertyTime("SIDPlay2", "Default Play Length", time)) > 0) {
			sidplay2_s.playLength = time;
		}
		if ((time = ini.getPropertyTime("SIDPlay2", "Default Record Length", time)) > 0) {
			sidplay2_s.recordLength = time;
		}

		sidplay2_s.single = ini.getPropertyInt("SIDPlay2", "SingleTrack", sidplay2_s.single ? 1 : 0) > 0;

		sidplay2_s.frameX = ini.getPropertyInt("SIDPlay2", "Frame X", sidplay2_s.frameX);
		sidplay2_s.frameY = ini.getPropertyInt("SIDPlay2", "Frame Y", sidplay2_s.frameY);
		sidplay2_s.frameWidth = ini.getPropertyInt("SIDPlay2", "Frame Width", sidplay2_s.frameX);
		sidplay2_s.frameHeight = ini.getPropertyInt("SIDPlay2", "Frame Height", sidplay2_s.frameHeight);
		sidplay2_s.lastDirectory = ini.getPropertyString("SIDPlay2", "Last Directory", "");

		return true;
	}

	private boolean writeSidplay2() {
		ini.changeSetting("SIDPlay2", "Version", "" + sidplay2_s.version);
		ini.changeSetting("SIDPlay2", "HVSC Dir", sidplay2_s.hvsc != null ? sidplay2_s.hvsc : "");
		ini.changeSetting("SIDPlay2", "CGSC Dir", sidplay2_s.cgsc != null ? sidplay2_s.cgsc : "");
		ini.changeSetting("SIDPlay2", "SOASC_6581R2", sidplay2_s.soasc6581R2 != null ? sidplay2_s.soasc6581R2 : "");
		ini.changeSetting("SIDPlay2", "SOASC_6581R4", sidplay2_s.soasc6581R4 != null ? sidplay2_s.soasc6581R4 : "");
		ini.changeSetting("SIDPlay2", "SOASC_8580R5", sidplay2_s.soasc8580R5 != null ? sidplay2_s.soasc8580R5 : "");
		ini.changeSetting("SIDPlay2", "EnableProxy", String.valueOf(sidplay2_s.enableProxy ? 1 : 0));
		ini.changeSetting("SIDPlay2", "ProxyHostname", sidplay2_s.proxyHostname != null ? sidplay2_s.proxyHostname : "");
		ini.changeSetting("SIDPlay2", "ProxyPort", String.valueOf(sidplay2_s.proxyPort));
		ini.changeSetting("SIDPlay2", "EnableDatabase", String.valueOf(sidplay2_s.enableDatabase ? 1 : 0));
		ini.changeSetting("SIDPlay2", "Default Play Length", String.format("%02d:%02d", (sidplay2_s.playLength / 60), (sidplay2_s.playLength % 60)));
		ini.changeSetting("SIDPlay2", "Default Record Length", String.format("%02d:%02d", (sidplay2_s.recordLength / 60), (sidplay2_s.recordLength % 60)));
		ini.changeSetting("SIDPlay2", "SingleTrack", String.valueOf(sidplay2_s.single ? 1 : 0));
		ini.changeSetting("SIDPlay2", "Frame X", "" + sidplay2_s.frameX);
		ini.changeSetting("SIDPlay2", "Frame Y", "" + sidplay2_s.frameY);
		ini.changeSetting("SIDPlay2", "Frame Width", "" + sidplay2_s.frameWidth);
		ini.changeSetting("SIDPlay2", "Frame Height", "" + sidplay2_s.frameHeight);
		ini.changeSetting("SIDPlay2", "Last Directory", sidplay2_s.lastDirectory);
		return true;
	}

	protected boolean readConsole() {
		console_s.ansi = ini.getPropertyBool("Console", "Ansi", console_s.ansi);
		console_s.topLeft = getChar(ini.getPropertyString("Console", "Char Top Left", "'" + console_s.topLeft + "'"), console_s.topLeft);
		console_s.topRight = getChar(ini.getPropertyString("Console", "Char Top Right", "'" + console_s.topRight + "'"), console_s.topRight);
		console_s.bottomLeft = getChar(ini.getPropertyString("Console", "Char Bottom Left", "'" + console_s.bottomLeft + "'"), console_s.bottomLeft);
		console_s.bottomRight = getChar(ini.getPropertyString("Console", "Char Bottom Right", "'" + console_s.bottomRight + "'"), console_s.bottomRight);
		console_s.vertical = getChar(ini.getPropertyString("Console", "Char Vertical", "'" + console_s.vertical + "'"), console_s.vertical);
		console_s.horizontal = getChar(ini.getPropertyString("Console", "Char Horizontal", "'" + console_s.horizontal + "'"), console_s.horizontal);
		console_s.junctionLeft = getChar(ini.getPropertyString("Console", "Char Junction Left", "'" + console_s.junctionLeft + "'"), console_s.junctionLeft);
		console_s.junctionRight = getChar(ini.getPropertyString("Console", "Char Junction Right", "'" + console_s.junctionRight + "'"), console_s.junctionRight);
		return true;
	}

	private boolean writeConsole() {
		if (console_s.ansi) {
			ini.changeSetting("Console", "Ansi", String.valueOf(console_s.ansi));
		} else {
			ini.changeSetting("Console", "Ansi", "");
		}

		ini.changeSetting("Console", "Char Top Left", "'" + String.valueOf(console_s.topLeft) + "'");
		ini.changeSetting("Console", "Char Top Right", "'" + String.valueOf(console_s.topRight) + "'");
		ini.changeSetting("Console", "Char Bottom Left", "'" + String.valueOf(console_s.bottomLeft) + "'");
		ini.changeSetting("Console", "Char Bottom Right", "'" + String.valueOf(console_s.bottomRight) + "'");
		ini.changeSetting("Console", "Char Vertical", "'" + String.valueOf(console_s.vertical) + "'");
		ini.changeSetting("Console", "Char Horizontal", "'" + String.valueOf(console_s.horizontal) + "'");
		ini.changeSetting("Console", "Char Junction Left", "'" + String.valueOf(console_s.junctionLeft) + "'");
		ini.changeSetting("Console", "Char Junction Right", "'" + String.valueOf(console_s.junctionRight) + "'");

		return true;
	}

	protected boolean readAudio() {
		boolean ret = true;
		audio_s.frequency = ini.getPropertyInt("Audio", "Frequency", audio_s.frequency);
		audio_s.sampling = SamplingMethod.valueOf(ini.getPropertyString(
				"Audio", "Sampling", "Decimate").toUpperCase(Locale.ENGLISH));
		audio_s.channels = ini.getPropertyInt("Audio", "Channels", 1);
		audio_s.mp3File = ini.getPropertyString("Audio", "MP3File", audio_s.mp3File);
		audio_s.playOriginal = ini.getPropertyBool("Audio", "PlayOriginal", audio_s.playOriginal);
		return ret;
	}

	private boolean writeAudio() {
		ini.changeSetting("Audio", "Frequency", String.valueOf(audio_s.frequency));
		ini.changeSetting("Audio", "Sampling", String.valueOf(audio_s.sampling));
		ini.changeSetting("Audio", "Channels", String.valueOf(audio_s.channels));

		if (audio_s.mp3File != null) {
			ini.changeSetting("Audio", "MP3File", audio_s.mp3File);
		} else {
			ini.changeSetting("Audio", "MP3File", "");
		}

		if (audio_s.playOriginal) {
			ini.changeSetting("Audio", "PlayOriginal", String.valueOf(audio_s.playOriginal));
		} else {
			ini.changeSetting("Audio", "PlayOriginal", "");
		}
		return true;
	}

	protected boolean readEmulation() {
		boolean ret = true;
		{
			final String str = ini.getPropertyString("Emulation", "DefaultClockSpeed", "PAL");
			emulation_s.defaultClockSpeed = Clock.valueOf(str);
		}

		{
			final String str = ini.getPropertyString("Emulation", "UserClockSpeed", "");
			if (! str.equals("")) {
				emulation_s.userClockSpeed = Clock.valueOf(str);
			}			
		}

		{
			final String str = ini.getPropertyString("Emulation", "DefaultSidModel", "MOS6581");
			emulation_s.defaultSidModel = ChipModel.valueOf(str);
		}

		{
			final String str = ini.getPropertyString("Emulation", "UserSidModel", "");
			if (! str.equals("")) {
				emulation_s.userSidModel = ChipModel.valueOf(str);
			}
		}


		
		emulation_s.hardsid6581 = ini.getPropertyInt("Emulation", "HardSID6581", emulation_s.hardsid6581);
		emulation_s.hardsid8580 = ini.getPropertyInt("Emulation", "HardSID8580", emulation_s.hardsid8580);
		emulation_s.filter = ini.getPropertyBool("Emulation", "UseFilter", true);
		emulation_s.digiBoosted8580 = ini.getPropertyBool("Emulation", "DigiBoosted8580", false);
		emulation_s.filter6581 = ini.getPropertyString("Emulation", "Filter6581", emulation_s.filter6581);
		if ("".equals(emulation_s.filter6581)) {
			emulation_s.filter6581 = null;
		}
		emulation_s.filter8580 = ini.getPropertyString("Emulation", "Filter8580", emulation_s.filter8580);
		if ("".equals(emulation_s.filter8580)) {
			emulation_s.filter8580 = null;
		}

		emulation_s.dualSidBase = Integer.decode(ini.getPropertyString("Emulation", "dualSidBase", ""+emulation_s.dualSidBase));
		emulation_s.forceStereoTune = ini.getPropertyBool("Emulation", "forceStereoTune", false);
		{
			final String str = ini.getPropertyString("Emulation", "StereoSidModel", "");
			if (! str.equals("")) {
				emulation_s.stereoSidModel = ChipModel.valueOf(str);
			}
		}

		// These next two change the ini section!
		if (emulation_s.filter6581 != null && emulation_s.filter6581.length() > 0) {
			// Try to load the filter
			currentFilter6581 = FilterConfig.read(ini, emulation_s.filter6581);
		}

		if (emulation_s.filter8580 != null && emulation_s.filter8580.length() > 0) {
			// Try to load the filter
			currentFilter8580 = FilterConfig.read(ini, emulation_s.filter8580);
		}

		return ret;
	}

	public FilterConfig getFilter(final String filterName) {
		return FilterConfig.read(ini, filterName);
	}

	public String[] getFilterList(final ChipModel model) {
		final List<String> filters = new ArrayList<String>();
		for (final String heading : ini.listSections()) {
			if (!heading.matches("Filter.*")) {
				continue;
			}

			if (getFilter(heading).getK() != 0 ^ model == ChipModel.MOS6581) {
				filters.add(heading);
			}
		}
		Collections.sort(filters);

		return filters.toArray(new String[] {});
	}

	private boolean writeEmulation() {
		ini.changeSetting("Emulation", "DefaultClockSpeed", emulation_s.defaultClockSpeed);
		ini.changeSetting("Emulation", "UserClockSpeed", emulation_s.userClockSpeed);
		ini.changeSetting("Emulation", "DefaultSidModel", emulation_s.defaultSidModel);
		ini.changeSetting("Emulation", "UserSidModel", emulation_s.userSidModel);
		ini.changeSetting("Emulation", "HardSID6581", emulation_s.hardsid6581);
		ini.changeSetting("Emulation", "HardSID8580", emulation_s.hardsid8580);
		ini.changeSetting("Emulation", "UseFilter", emulation_s.filter);
		ini.changeSetting("Emulation", "DigiBoosted8580", emulation_s.digiBoosted8580);
		ini.changeSetting("Emulation", "Filter6581", emulation_s.filter6581);
		ini.changeSetting("Emulation", "Filter8580", emulation_s.filter8580);
		ini.changeSetting("Emulation", "dualSidBase", String.format("0x%4x", emulation_s.dualSidBase));
		ini.changeSetting("Emulation", "forceStereoTune", emulation_s.forceStereoTune);
		ini.changeSetting("Emulation", "StereoSidModel", emulation_s.stereoSidModel);
		return true;
	}

	private boolean readFavorites() {
		final String titles = listToString(favorites_s.favorites_titles);
		favorites_s.favorites_titles = stringToList(ini.getPropertyString("Favorites", "Titles", titles));
		final String filenames = listToString(favorites_s.favorites_filenames);
		favorites_s.favorites_filenames = stringToList(ini.getPropertyString("Favorites", "Filenames", filenames));
		favorites_s.favorites_current = ini.getPropertyString("Favorites", "Current", favorites_s.favorites_current);
		return true;
	}

	private boolean writeFavorites() {
		ini.changeSetting("Favorites", "Titles", listToString(favorites_s.favorites_titles));
		ini.changeSetting("Favorites", "Filenames", listToString(favorites_s.favorites_filenames));
		ini.changeSetting("Favorites", "Current", String.valueOf(favorites_s.favorites_current));
		return true;
	}

	public IniConfig() {
		iniPath = getINIPath();

		clear();
	}

	public void read() {
		if (iniPath == null) {
			// Use internal INI file in the applet version
			readInternal();
			return;
		}
		ini = null;

		final File iniFile = new File(iniPath);
		if (iniFile.exists()) {
			clear();
			System.out.println("Reading INI file: " + iniFile.getAbsolutePath());
			try {
				ini = new IniReader(iniFile.getAbsolutePath());
				status &= readSidplay2();
				status &= readConsole();
				status &= readAudio();
				status &= readEmulation();
				status &= readFavorites();
				ini.setDirty(false);
				/* validate loaded configuration */
				if (status && sidplay2_s.version == REQUIRED_CONFIG_VERSION) {
					return;
				}
				createINIBackup(iniFile);
			} catch (final Exception e) {
				System.err.println("INI reading failed: " + e.getMessage());
			}

			System.out.println("INI file old/broken (version=" + sidplay2_s.version + ", status=" + status + "). Using default settings.");
		}

		readInternal();
	}

	/**
	 * Determine INI filename.
	 * <OL>
	 * <LI>If INI file exists in the user directory, then use it, else
	 * <LI>use INI file in the current working directory
	 * </OL>
	 * 
	 * @return the absolute path name of the INI file to use
	 */
	private String getINIPath() {
		try {
			File configPlace = null;
			for (final String s : new String[] { System.getProperty("user.dir"), System.getProperty("user.home"), }) {
				if (s == null) {
					continue;
				}

				configPlace = new File(s, FILE_NAME);
				if (configPlace.exists()) {
					return configPlace.getAbsolutePath();
				}
			}
			/* no file? create one in user home (last option we tried) */
			assert configPlace != null;
			return configPlace.getAbsolutePath();
		} catch (final AccessControlException e) {
			// No external config file in the applet version
			return null;
		}

	}

	/**
	 * Create backup of old INI file
	 * 
	 * @param iniFile
	 *            the INI file to backup
	 */
	private void createINIBackup(final File iniFile) {
		InputStream is = null;
		OutputStream os = null;
		try {
			is = new FileInputStream(iniFile);
			os = new BufferedOutputStream(new FileOutputStream(iniFile.getAbsolutePath() + ".bak"));
			while (is.available() > 0) {
				os.write(is.read());
			}
		} catch (final IOException e) {
			e.printStackTrace();
		} finally {
			if (os != null) {
				try {
					os.close();
				} catch (final IOException e1) {
					e1.printStackTrace();
				}
			}
			if (is != null) {
				try {
					is.close();
				} catch (final IOException e1) {
					e1.printStackTrace();
				}
			}
		}
	}

	public void readInternal() {
		final InputStream is = getClass().getClassLoader().getResourceAsStream("sidplay/ini/" + FILE_NAME);
		System.out.println("Use internal INI file: " + "sidplay/ini/" + FILE_NAME);

		clear();

		try {
			ini = new IniReader(is);
			is.close();
		} catch (final IOException e) {
			status = false;
			System.err.println(e.getMessage());
			return;
		}

		status &= readSidplay2();
		status &= readConsole();
		status &= readAudio();
		status &= readEmulation();
		status &= readFavorites();

		sidplay2_s.version = REQUIRED_CONFIG_VERSION;
		ini.setDirty(false);
	}

	public void write() {
		if (!ini.isDirty()) {
			return;
		}
		status &= writeSidplay2();
		status &= writeConsole();
		status &= writeAudio();
		status &= writeEmulation();
		status &= writeFavorites();

		try {
			final File iniFile = new File(iniPath);
			ini.save(iniFile.getAbsolutePath());
			System.out.println("INI file saved: " + iniFile.getAbsolutePath());
			ini.setDirty(false);
		} catch (final IOException e) {
			System.err.println(e.getMessage());
			status = false;
		}
	}

	public boolean bool() {
		return status;
	}

	// Sidplay2 Specific Section
	public final sidplay2_section sidplay2() {
		return sidplay2_s;
	}

	public final console_section console() {
		return console_s;
	}

	public final audio_section audio() {
		return audio_s;
	}

	public final emulation_section emulation() {
		return emulation_s;
	}

	public final favorites_section favorites() {
		return favorites_s;
	}

	public final FilterConfig filter(final ChipModel model) {
		if (model == ChipModel.MOS8580) {
			return currentFilter8580;
		} else {
			return currentFilter6581;
		}
	}

	public File getIniFile() {
		assert iniPath != null;

		return new File(iniPath);
	}

	private String[] stringToList(final String str) {
		if (str == null) {
			return new String[0];
		}
		final ArrayList<String> result = new ArrayList<String>();
		final StringTokenizer tok = new StringTokenizer(str, ",", false);
		while (tok.hasMoreElements()) {
			final String name = (String) tok.nextElement();
			if ("null".equals(name)) {
				result.add(null);
			} else {
				result.add(name);
			}
		}
		return result.toArray(new String[result.size()]);
	}

	private String listToString(final String[] list) {
		if (list == null) {
			return "";
		}
		final StringBuffer result = new StringBuffer();
		for (int i = 0; i < list.length; i++) {
			if (i != 0) {
				result.append(",");
			}
			result.append(list[i]);
		}
		return result.toString();
	}

	public static int toInt(final String base) {
		return toInt2(base, 4);
	}

	public static int toInt2(final String base, final int length) {
		if (base.length() == 2 + length && base.startsWith("0x")) {
			return Integer.parseInt(base.substring(2), 16);
		}
		return -1;
	}

	private char getChar(final String str, final char defaultChar) {
		char c = 0;
		if (str.length() == 0) {
			return defaultChar;
		}
		// Check if we have an actual Character
		if (str.charAt(0) == '\'') {
			if (str.charAt(2) != '\'') {
				status = false;
			} else {
				c = str.charAt(1);
			}
		} // Nope is number
		else {
			c = (char) Integer.parseInt(str);
		}

		// Clip off special characters
		if (c >= 32) {
			return c;
		}
		return defaultChar;
	}

}
