package sidplay.audio;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;

public class CompareDriver extends AudioDriver {
	private final JavaSound audioDriver;
	private AudioInputStream din;

	private boolean playOriginal;
	private AudioConfig audioConfig;

	private String mp3File;
	private byte[] mp3Data;

	public CompareDriver() {
		audioDriver = new JavaSound();
	}

	private void openMp3File() throws UnsupportedAudioFileException, IOException {
		final AudioInputStream in = AudioSystem.getAudioInputStream(new File(mp3File));
		final AudioFormat baseFormat = in.getFormat();
		final AudioFormat decodedFormat = new AudioFormat(
				AudioFormat.Encoding.PCM_SIGNED, // Encoding to use
				baseFormat.getSampleRate(), // sample rate (same as base format)
				16, // sample size in bits (thx to Javazoom)
				audioConfig.channels, // # of Channels
				audioConfig.channels * 2, // Frame Size
				baseFormat.getSampleRate(), // Frame Rate
				false // little-endian
		);
		din = AudioSystem.getAudioInputStream(decodedFormat, in);
		audioConfig.frequency = (int) decodedFormat.getSampleRate();
	}

	@Override
	public void open(final AudioConfig _cfg, final String name)
	throws LineUnavailableException, UnsupportedAudioFileException, IOException {
		audioConfig = _cfg;

		openMp3File();
		audioDriver.open(audioConfig, name);

		mp3Data = new byte[audioDriver.buffer().capacity()];
	}

	@Override
	public void pause() {
		audioDriver.pause();
	}

	@Override
	public void reset() {
		audioDriver.reset();
	}

	@Override
	public void write() {
		try {
			int pos = 0;
			while (pos < mp3Data.length) {
				int readSize = -1;
				try {
					readSize = din.read(mp3Data, pos, mp3Data.length - pos);
				}
				catch (ArrayIndexOutOfBoundsException aio) {
				}
				if (readSize == -1) {
					openMp3File();
				}
				pos += readSize;
			}
		}
		catch (final Exception e) {
			throw new RuntimeException(e);
		}

		/* which one do we play? */
		if (playOriginal) {
			System.arraycopy(mp3Data, 0, audioDriver.buffer().array(), 0, mp3Data.length);
		}

		audioDriver.write();
	}

	@Override
	public void close() {
		audioDriver.close();
		// Stop
		try {
			if (din != null) {
				din.close();
			}
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	public void setPlayOriginal(final boolean playOriginal) {
		this.playOriginal = playOriginal;
	}
	
	@Override
	public ByteBuffer buffer() {
		return audioDriver.buffer();
	}

	public void setMp3File(String mp3File) {
		this.mp3File = mp3File;
	}
}
