package sidplay.audio;

import java.io.IOException;
import java.nio.ByteBuffer;

import javax.sound.sampled.LineUnavailableException;

public class LiveRecorderWav extends AudioDriver {
	private final JavaSound audioDriver;
	private final WavFile wavFile;

	public LiveRecorderWav() {
		audioDriver = new JavaSound();
		wavFile = new WavFile();
	}

	@Override
	public String extension() {
		return wavFile.extension();
	}

	@Override
	public void open(final AudioConfig cfg, final String name) throws LineUnavailableException, IOException {
		audioDriver.open(cfg, name);
		wavFile.open(cfg, name);
	}

	@Override
	public void pause() {
		audioDriver.pause();
		wavFile.pause();
	}

	@Override
	public void reset() {
		audioDriver.reset();
		wavFile.reset();
	}

	@Override
	public void write() {
		System.arraycopy(audioDriver.buffer().array(), 0,
				wavFile.buffer().array(), 0, wavFile.buffer().capacity());
		audioDriver.write();
		wavFile.write();
	}

	@Override
	public void close() {
		audioDriver.close();
		wavFile.close();
	}
	
	@Override
	public ByteBuffer buffer() {
		return audioDriver.buffer();
	}
}
